// Original author: Anders Logg. Added to Kraftwerk 2001.
// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#ifndef __PARAMETER_LIST_HH
#define __PARAMETER_LIST_HH

#include <stdarg.h>
#include "Parameter.hh"
#include "kw_constants.h"

#define KW_FILE_FORMAT_DX    1
#define KW_FILE_FORMAT_INP   2

///
class ParameterList{
public:
  
  /// Constructor
  ParameterList();
  /// Destructor
  virtual ~ParameterList() { delete parameters; };

  /// Initialize and check all parameters
  virtual void Initialize() = 0;
  
  /// Add a parameter
  void Add(const char *identifier, Type type, ...);
  
  /// Set the value of a parameter
  void SetByArgumentList(const char *identifier, va_list aptr);
  void Set(const char *identifier, ...);

  /// Get the value of a parameter
  void GetByArgumentList(const char *identifier, va_list aptr);
  void Get(const char *identifier, ...);

  /// Save all parameters to the default file
  void Save();

  /// Save all parameters to the given file
  void Save(const char *filename);

  /// Load all parameters from the default file
  void Load();

  /// Load all parameters from the given file
  void Load(const char *filename);

  /// Check if the parameter has been changed
  bool Changed(const char *identifier);
  
private:

  int  ParameterIndex(const char *identifier);
  void Realloc();

  // A list of all the parameters
  Parameter *parameters;

  // Size of list
  int alloc_size;

  // Position for next item
  int current;
  
};

#endif
