// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.
//
// Original author: Anders Logg. Added to Kraftwerk 2002-01

#ifndef __EQUATION_HH
#define __EQUATION_HH

#include <kw_constants.h>
#include <mvvtp.h>
#include <mvmtp.h>
#include "LocalField.hh"
#include "GlobalField.hh"
#include "ShapeFunction.hh"

class GlobalField;
class FiniteElement;

///
class Equation {

public:
  
  Equation(int nsd);
  ~Equation();

  virtual real IntegrateLHS(ShapeFunction &u, ShapeFunction &v) = 0;
  virtual real IntegrateRHS(ShapeFunction &v) = 0;

  void Update(int cell);

  void AttachField(int i, GlobalField *globalfield, int component = 0);

  void SetTime     (real t);
  void SetTimeStep (real dt);
  
  /// Returns the number of equations 
  int GetNoEq();
    
  void InitCoeff(int no_coeff);
  void SetCoeff(GlobalField *gf, int n);

  int GetStartVectorComponent();

  friend class Discretiser;
  
protected:
  
  void AllocateFields(int no_fields);

  void AttachFiniteElement(FiniteElement *element);
  
  LocalField **field;
  int no_fields;

  GlobalField **coeff;
  int no_coeff;
  
  bool variational_multiscale_eddy_viscosity;
  bool smagorinsky_eddy_viscosity;

  int start_vector_component;

  int nsd;   // number of space dimensions
  int no_eq; // number of equations
  int noBF;  // ???

  real dt;
  real t;
  
  real pi;

  real h;

private:

  void Init(int nsd);
  
};

#endif
