// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.
//
// Original author: Hoffman/Logg. Added to Kraftwerk 2002-01

#ifndef __FUNCTION_SPACE_HH
#define __FUNCTION_SPACE_HH

#include <kw_constants.h>

class FiniteElement;
class ShapeFunction;

class FunctionSpace{

public:

  FunctionSpace(FiniteElement *element, int nvc);
  ~FunctionSpace();

  int GetDim();
  int GetSpaceDim();
  int GetNoComponents();
  int GetCellNo();

  real GetCoord(int node, int dim);

  int  GetNoCells();

  virtual real GetGradient(int i, int j) = 0;

  virtual void Update() = 0;
  
  real GetCircumRadius();

  friend class FiniteElement;

  FiniteElement* GetFiniteElement();
  ShapeFunction* GetShapeFunction(int dof);

protected:

  FiniteElement *element;
  ShapeFunction **shapefunction;

  int dim; // Scalar dimension 
  int nsd; // Number of space dimensions
  int nvc; // Number of vector components
  
};

#endif
