// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#ifndef __GLOBALFIELD_HH
#define __GLOBALFIELD_HH

#include <kw_constants.h>

class Grid;
class Vector;

enum Representation { NONE, NODAL, CELL, CONSTANT };

enum nodalbasis { NO_BASIS, LINE_BASIS, TRI_BASIS, QUAD_BASIS, TET_BASIS, HEX_BASIS, PYR_BASIS, 
		  PRISM_BASIS, SIMPLEX_BASIS, HYPERCUBE_BASIS };
 
class GlobalField{
  
public:

  GlobalField();
  GlobalField(Grid *grid);
  GlobalField(Grid *grid, real constant);
  GlobalField(Grid *grid, Vector *x, int nvd = 1);
  
  ~GlobalField();

  void AttachGrid( Grid* g );
  
  int GetNoDof();
  int GetVectorDim();

  void InitByNodes(Vector *nodal_values, int nvd);
  void InitByDof(Vector *dof_values, int nvd);
  void InitByCells(Vector *cell_values, int nvd);
  void InitByConstant(real c);
  
  int GetDim(int cell);

  void GetLocalDof(int cell, real *local_dofs, int component = 0);

  Representation GetRep() { return representation; }  

  real GetCellValue(int cell, int component);
  real GetCellValue(int cell);  
  real GetNodalValue(int nod, int component);
  real GetNodalValue(int nod);
  real GetDofValue(int nod, int component);
  real GetDofValue(int nod);
  real GetConstantValue(); 
  
private:
  
  Grid* grid;
  
  int no_dof;
  int nvd;

  Representation representation;

  Vector *dof_values;
  Vector *cell_values;
  real constant;    

};

#endif





