// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#ifndef __LOCALFIELD_HH
#define __LOCALFIELD_HH

#include <kw_constants.h>

class FunctionSpace;
class ShapeFunction;
class GlobalField;
class FiniteElement;

class LocalField {
  
public:

  LocalField();
  LocalField(FiniteElement *element);

  ~LocalField();

  void AttachFiniteElement(FiniteElement *element);
  void AttachGlobalField(GlobalField *globalfield, int component);
  
  void operator = (LocalField &lf);

  void Update(int cell);
  
  void Refill(int cellnumber, real *nvals);
  void Refill(int cellnumber, int component, GlobalField *field);

  void Add(LocalField &lf);
  void Mult(real a);
  
  void EvalGradient(real* gradient); 
  void ComputeGradient(); 
  real EvalGradient(int component); 
  bool GradientComputed(); 

  real dx();
  real dy();
  real dz();
  real d(int i);

  int GetDim() const;

  int  GetCellNo() const;
  real GetCoord(int node, int dim) const;
  real GetDofValue(int i) const;
  real GetMeanValue() const;

  FiniteElement* GetFiniteElement() const;
  FunctionSpace* GetFunctionSpace();
  const ShapeFunction* GetShapeFunction(int i) const; 

  // Operators

  real operator* (const LocalField &v) const;
  real operator* (const ShapeFunction &v) const;
  real operator* (real a) const;

  friend real operator* (real a, LocalField &v);
  friend real operator* (const ShapeFunction &v, const LocalField &w);
  
protected:

  void Resize(int dim);

  real *dof;
  real *gradient;
  
  int dim;
  int component;
  
  bool gradient_computed;

  GlobalField *globalfield;
  FunctionSpace *functionspace;

};

#endif
