// Original author: Johan Hoffman. Added to Kraftwerk 2002-02-05.
// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#include "ShapeFunction.hh"
#include "FunctionSpace.hh"
#include <Settings.hh>
#include <Display.hh>

extern Settings *settings;


//-----------------------------------------------------------------------------
ShapeFunction::ShapeFunction(FunctionSpace *functionspace, int dof)
{
  this->functionspace = functionspace;
  this->dof = dof;
  active = true;
}
//-----------------------------------------------------------------------------
ShapeFunction::~ShapeFunction()
{
}
//-----------------------------------------------------------------------------
void ShapeFunction::SetDof(int dof)
{
  this->dof = dof;
}
//-----------------------------------------------------------------------------
void ShapeFunction::Active(bool active)
{
  this->active = active;
}
//-----------------------------------------------------------------------------
bool ShapeFunction::Active() const
{
  return active;
}
//-----------------------------------------------------------------------------
int ShapeFunction::GetDof() const
{
  return dof;
}
//-----------------------------------------------------------------------------
int ShapeFunction::GetDim()
{
  return functionspace->GetDim();
}
//-----------------------------------------------------------------------------
int ShapeFunction::GetCellNo()
{
  return functionspace->GetCellNo();
}
//-----------------------------------------------------------------------------
int ShapeFunction::GetNoCells()
{
  return functionspace->GetNoCells();
}
//-----------------------------------------------------------------------------
real ShapeFunction::GetCircumRadius()
{
  return functionspace->GetCircumRadius();
}
//-----------------------------------------------------------------------------
real ShapeFunction::GetCoord(int node, int dim)
{
  return functionspace->GetCoord(node,dim);
}
//-----------------------------------------------------------------------------
FiniteElement* ShapeFunction::GetFiniteElement()
{
  return functionspace->GetFiniteElement();
}
//-----------------------------------------------------------------------------
real operator* (real a, ShapeFunction &v)
{
  // This makes sure that a * ShapeFunction is commutative

  return ( v*a );
}
//-----------------------------------------------------------------------------
