// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.
//
// Original author: Johan Hoffman. Added to Kraftwerk 2002-02-05.

#ifndef __SHAPEFUNCTION_HH
#define __SHAPEFUNCTION_HH

#include <kw_constants.h>

class FunctionSpace;
class FiniteElement;

class ShapeFunction{
  
public:

  ShapeFunction(FunctionSpace *functionspace, int dof);
  ~ShapeFunction();

  void SetDof(int dof);
  int  GetDof() const; 
  int  GetCellNo();

  int  GetDim();
  int  GetNoCells();
  
  real GetCircumRadius();

  real GetCoord(int node, int dim);

  virtual void barycentric(real *point, real *bcoord) = 0;
  
  virtual real dx() = 0;
  virtual real dy() = 0;
  virtual real dz() = 0;
  virtual real d(int i) = 0;
  
  friend class FunctionSpace;
  
  void Active(bool active);
  bool Active() const;

  FiniteElement* GetFiniteElement();

  // Operators
  
  virtual real operator* (const ShapeFunction &v) const = 0;
  virtual real operator* (real a) const = 0;
  
  friend real operator* (real a, ShapeFunction &v);
  
  friend class TetLinSpace;
  friend class TetLinFunction;
  friend class TriLinFunction;
  
protected:
  
  FunctionSpace *functionspace;

  int dof;
  bool active;
  
private: 

};

#endif
