// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.
//
// Original author: Johan Hoffman. Added to Kraftwerk 2002-01-14.

#include "TetLinFunction.hh"
#include "TetLinSpace.hh"
#include "ShapeFunction.hh"
#include "FiniteElement.hh"
#include "LocalField.hh"
#include <Display.hh> 

//-----------------------------------------------------------------------------
TetLinFunction::TetLinFunction(FunctionSpace *functionspace, int dof) :
  ShapeFunction(functionspace,dof)
{
}
//-----------------------------------------------------------------------------
TetLinFunction::~TetLinFunction()
{
}
//-----------------------------------------------------------------------------
void TetLinFunction::barycentric(real *point, real *bcoord)
{
  point[0]-=functionspace->GetCoord(0,0);
  point[1]-=functionspace->GetCoord(0,1);
  point[2]-=functionspace->GetCoord(0,2);

  bcoord[1] = functionspace->GetGradient(1,0)*point[0] + 
              functionspace->GetGradient(1,1)*point[1] + 
              functionspace->GetGradient(1,2)*point[2]; 
  bcoord[2] = functionspace->GetGradient(2,0)*point[0] + 
              functionspace->GetGradient(2,1)*point[1] + 
              functionspace->GetGradient(2,2)*point[2];
  bcoord[3] = functionspace->GetGradient(3,0)*point[0] + 
              functionspace->GetGradient(3,1)*point[1] + 
              functionspace->GetGradient(3,2)*point[2];    
  bcoord[0] = 1 - bcoord[1] - bcoord[2] - bcoord[3];
}
//-----------------------------------------------------------------------------
real TetLinFunction::dx()
{
  return functionspace->GetGradient(dof,0);
}
//-----------------------------------------------------------------------------
real TetLinFunction::dy()
{
  return functionspace->GetGradient(dof,1);
}
//-----------------------------------------------------------------------------
real TetLinFunction::dz()
{
  return functionspace->GetGradient(dof,2);
}
//-----------------------------------------------------------------------------
real TetLinFunction::d(int i)
{
  return functionspace->GetGradient(dof,i);
}
//-----------------------------------------------------------------------------
real TetLinFunction::operator* (const ShapeFunction &v) const
{
  if ( !active || !v.active ) return 0.0;

  if ( dof == v.dof )
	 return 0.1;
  else
	 return 0.05;
}
//-----------------------------------------------------------------------------
real TetLinFunction::operator* (real a) const
{
  if ( !active ) return 0.0;

  return a*0.25;
}
//-----------------------------------------------------------------------------
