// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#include "TriLinFunction.hh"
#include "TriLinSpace.hh"
#include "ShapeFunction.hh"
#include "FiniteElement.hh"
#include "LocalField.hh"
#include <Display.hh> 

//-----------------------------------------------------------------------------
TriLinFunction::TriLinFunction(FunctionSpace *functionspace, int dof) :
  ShapeFunction(functionspace,dof)
{
}
//-----------------------------------------------------------------------------
TriLinFunction::~TriLinFunction()
{
}
//-----------------------------------------------------------------------------
void TriLinFunction::barycentric(real *point, real *bcoord)
{
  point[0]-=functionspace->GetCoord(0,0);
  point[1]-=functionspace->GetCoord(0,1);
  point[2]-=functionspace->GetCoord(0,2);

  bcoord[1] = functionspace->GetGradient(1,0)*point[0] + 
              functionspace->GetGradient(1,1)*point[1] + 
              functionspace->GetGradient(1,2)*point[2]; 
  bcoord[2] = functionspace->GetGradient(2,0)*point[0] + 
              functionspace->GetGradient(2,1)*point[1] + 
              functionspace->GetGradient(2,2)*point[2];
  bcoord[3] = functionspace->GetGradient(3,0)*point[0] + 
              functionspace->GetGradient(3,1)*point[1] + 
              functionspace->GetGradient(3,2)*point[2];    
  bcoord[0] = 1 - bcoord[1] - bcoord[2] - bcoord[3];
}
//-----------------------------------------------------------------------------
real TriLinFunction::dx()
{
  return functionspace->GetGradient(dof,0);
}
//-----------------------------------------------------------------------------
real TriLinFunction::dy()
{
  return functionspace->GetGradient(dof,1);
}
//-----------------------------------------------------------------------------
real TriLinFunction::dz()
{
  // FIXME: This one shouldn't be here
  display->InternalError("TriLinFunction::dz()","This function should not be used.");
}
//-----------------------------------------------------------------------------
real TriLinFunction::d(int i)
{
  return functionspace->GetGradient(dof,i);
}
//-----------------------------------------------------------------------------
real TriLinFunction::operator* (const ShapeFunction &v) const
{
  if ( !active || !v.active ) return 0.0;

  if ( dof == v.dof )
	 return KW_1_6;
  else
	 return KW_1_12;
}
//-----------------------------------------------------------------------------
real TriLinFunction::operator* (real a) const
{
  if ( !active ) return 0.0;

  return a*KW_1_3;
}
//-----------------------------------------------------------------------------
