// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#ifndef __OPEN_DX_HH
#define __OPEN_DX_HH

#include <stdio.h>
#include <assert.h>
#include <iostream.h>
#include <stdarg.h>

#include <kw_constants.h>
#include <Grid.hh>

class Vector;

class OpenDX{

public:
  
  OpenDX  (const char *cFileName, int iNumberOfData, ...);
  ~OpenDX ();

  /// Sets the label for component [component]
  void SetLabel         (int component, const char *label);
  /// Resets the file, overwriting previously saved frames
  void Reset            ();
  /// Adds a frame to the file, different format
  void AddFrame         (Grid *grid, Vector *u, real t);
  
  /// Returns data (interpolated) at time t
  //void GetData          (Grid *grid, MV_ColMat<real> *data, real t);
  /// Returns data (interpolated) at time t for component [component]
  //void GetDataComponent (Grid *grid, MV_ColMat<real> *data, real t, int component);
  
  /// Writes an .inp file (interpolated) at time t 
  void WriteInp (Grid *grid, real t);
  
  
private:
  
  /// Writes an .inp file from data (interpolated) at time t 
  //void WriteInpFromData (char *file, Grid *grid, MV_ColMat<real>* data, int nsys, int firstcomp);

  void WriteHeader      ();
  void WriteGrid        (Grid *grid);
  //void WriteField       (Grid *grid, MV_ColMat<real> *data);
  void WriteField       (Grid *grid, Vector *u);
  void WriteSeries      ();
  void RemoveSeries     ();
  void SaveTime         (real t);
  
  void ReadGrid         (Grid *grid);
  //void ReadField        (Grid *grid, MV_ColMat<real> *data);
  bool ReadObject       (const char *object_name, int *items, int *dimension);
  //void ReadData         (MV_ColMat<real> *data,
  //								int component, int obj1, int obj2,
  //								float t1, float t2, real t);
  void FindObject       (real t, int *obj1, int *obj2, float *t1, float *t2);
  
  FILE *fp;
  //char cFileName[KW_LINELENGTH];
  char cFileName[100];

  bool bFileCleared;
  bool bWrittenHeader;
  bool bWrittenGrid;

  int iFrame;
  int iObject;

  int datasize;
  int iNumberOfData;
  int *iDataDimensions;

  char **cDataLabels;

  int   iTimeValuesSize;
  real *rTimeValues;
  
};

#endif
