// (c) 2002 Johan Hoffman & Anders Logg, Chalmers Finite Element Center.
// Licensed under the GNU GPL Version 2.

#ifndef __OUTPUT_HH
#define __OUTPUT_HH

#include <stdarg.h>
#include <stdio.h>
#include <kw_constants.h>

class Grid;
class Vector;

enum OutputFileType { matlab, opendx };

class Output{
public:

  Output(const char *filename, int no_data, ...);
  ~Output();

  /// Sets the component label
  void SetLabel(int component, const char *label);
  /// Adds a frame to the file
  void AddFrame(Grid *grid, Vector *u, real t);
  /// Resets the file, overwriting previously saved frames
  void Reset();
  
private:

  void MatlabAddFrame      (FILE *fp, Grid *grid, Vector *u, real t);
  void __MatlabWriteHeader (FILE *fp);
  void __MatlabWriteGrid   (FILE *fp, Grid *grid);
  void __MatlabWriteField  (FILE *fp, Vector *u, real t);

  void UpdateSystemData();
  
  // Filename for saving output
  char filename[KW_LINELENGTH];

  // Data dimensions and labels
  int datasize;
  int no_data;
  int *dimensions;
  char **labels;

  // Indicators for saving
  int current_frame;
  bool done_grid;
  bool done_header;

  // System variables
  char sys_user[KW_LINELENGTH];
  char sys_time[KW_LINELENGTH];
  char sys_host[KW_LINELENGTH];
  char sys_machine[KW_LINELENGTH];
  char sys_sysname[KW_LINELENGTH];
  char sys_version[KW_LINELENGTH];

  // Problem description
  char problem_description[KW_LINELENGTH];
  
  OutputFileType filetype;
  
};

#endif
