#include <Display.hh>
#include "Vector.hh"

//-----------------------------------------------------------------------------
Vector::Vector()
{
  n = 0;
  values = 0;
}
//-----------------------------------------------------------------------------
Vector::Vector(int n)
{
  values = 0;

  Resize(n);
}
//-----------------------------------------------------------------------------
Vector::~Vector()
{
  delete [] values;
}
//-----------------------------------------------------------------------------
void Vector::Resize(int n)
{
  if ( values )
	 delete [] values;

  this->n = n;

  values = new real[n];

  for (int i=0;i<n;i++)
	 values[i] = 0.0;
}
//-----------------------------------------------------------------------------
void Vector::CopyTo(Vector* vec)
{
  if (Size()!=vec->Size()) display->InternalError("Vector::operator = ()","Vectors are not the same length");

  for (int i=0; i<n; i++) vec->Set(i,values[i]);    
}
//-----------------------------------------------------------------------------
void Vector::CopyFrom(Vector* vec)
{
  if (Size()!=vec->Size()) display->InternalError("Vector::operator = ()","Vectors are not the same length");

  for (int i=0; i<n; i++) Set(i,vec->Get(i));    
}
//-----------------------------------------------------------------------------
void Vector::SetToConstant(real val)
{
  for (int i=0; i<n; i++) Set(i,val);    
}
//-----------------------------------------------------------------------------
real Vector::operator()(int i)
{
  if ((i<0)||(i>=n))
	 display->InternalError("Vector::operator()","Illegal vector index: %d",i);

  return values[i];
}
//-----------------------------------------------------------------------------
void Vector::Set(int i, real val)
{
  if ((i<0)||(i>=n)){
    display->InternalError("Vector::Set()","Illegal vector index: %d",i);
  }
  values[i] = val;
}
//-----------------------------------------------------------------------------
real Vector::Get(int i)
{
  if ((i<0)||(i>=n)){
    real *l;
    l[39462] = 1;
    display->InternalError("Vector::Get()","Illegal vector index: %d",i);
  }

  return values[i];
}
//-----------------------------------------------------------------------------
void Vector::Add(int i, real val)
{
  if ((i<0)||(i>=n))
    display->InternalError("Vector::Add()","Illegal vector index: %d",i);

  values[i] += val;
}
//-----------------------------------------------------------------------------
void Vector::Add(real a, Vector *v)
{
  if ( n != v->Size() )
	 display->InternalError("Vector::Add()","Dimensions don't match: %d != %d.",n,v->Size());

  for (int i=0;i<n;i++)
	 values[i] += a*v->values[i];
}
//-----------------------------------------------------------------------------
void Vector::Mult(int i, real val)
{
  if ((i<0)||(i>=n))
	 display->InternalError("Vector::Add()","Illegal vector index");

  values[i] *= val;
}
//-----------------------------------------------------------------------------
real Vector::Dot(Vector *v)
{
  if ( n != v->Size() )
	 display->InternalError("Vector::Dot()","Dimensions don't match: %d != %d.",n,v->Size());
  
  real sum = 0.0;
  for (int i=0;i<n;i++)
	 sum += values[i]*v->values[i];

  return sum;
}
//-----------------------------------------------------------------------------
real Vector::Norm()
{
  real norm = 0.0; 
  for (int i=0;i<n;i++) norm += sqr(values[i]);
  return sqrt(norm);
}
//-----------------------------------------------------------------------------
int Vector::Size()
{
  return n;
}
//-----------------------------------------------------------------------------
void Vector::Display()
{
  printf("v = ");
  for (int i=0;i<n;i++)
	 printf("%f ",values[i]);
  printf("\n");
}
//-----------------------------------------------------------------------------
