#ifndef EquationNS_cG1cG1_h
#define EquationNS_cG1cG1_h

class EquationNS_cG1cG1: 
  public Equation {


private: 

  real theta; // theta=0.5 => cG(1)cG(1) 

  real delta1,delta2,r;
  real C1,C2,K;

  real Re,dt,t;

  real Cs,S_abs,nu,nu_T;

  real umid1_1,umid1_2,umid1_3,umid1_4;
  real umid2_1,umid2_2,umid2_3,umid2_4;
  real umid3_1,umid3_2,umid3_3,umid3_4;
  real umid1,umid2,umid3;

  real u1sum,u2sum,u3sum,psum;

  real u1u1,u1u2,u1u3,u2u1,u2u2,u2u3,u3u1,u3u2,u3u3;
  real f1u1,f1u2,f1u3,f2u1,f2u2,f2u3,f3u1,f3u2,f3u3;

  real uTS1,uTS2,uTS3,uNL1,uNL2,uNL3;
  real uTS1uTS1,uTS1uTS2,uTS1uTS3,uTS2uTS1,uTS2uTS2,uTS2uTS3,uTS3uTS1,uTS3uTS2,uTS3uTS3;
  real uNL1uTS1,uNL1uTS2,uNL1uTS3,uNL2uTS1,uNL2uTS2,uNL2uTS3,uNL3uTS1,uNL3uTS2,uNL3uTS3;

  real normU,divU,divUmid;

  real Ii;
  real Iii;
  real Iij;

  MV_Vector<real> f1;
  MV_Vector<real> f2;
  MV_Vector<real> f3;

  real vx,vy,vz;
  real uv1,uv2,uv3,uv4;
  real D1u1,D2u1,D3u1,D1u2,D2u2,D3u2,D1u3,D2u3,D3u3;
  real D1umid1,D2umid1,D3umid1,D1umid2,D2umid2,D3umid2,D1umid3,D2umid3,D3umid3;
  real uv;
  real MASS1,MASS2,MASS3,NL1,NL2,NL3,LAP1,LAP2,LAP3,LAPp;
  real uDu_uDv_1,uDu_uDv_2,uDu_uDv_3;
  real LSdiv1,LSdiv2,LSdiv3;
  real p_u_Dv_1,p_u_Dv_2,p_u_Dv_3;
  real f_uDv_1,f_uDv_2,f_uDv_3;
  real SD1,SD2,SD3;
  real uprev_TS_v_1,uprev_TS_v_2,uprev_TS_v_3;
  real f_v_1,f_v_2,f_v_3;
  real p_div_v_1,p_div_v_2,p_div_v_3;
  real u_Dv_1,u_Dv_2,u_Dv_3;
  real u_v_1,u_v_2,u_v_3;
  real uDu_Dq,f_Dq;

public:
    
    //-----------------------------------------
    //      User definitions of equations
    //-----------------------------------------
    EquationNS_cG1cG1() : Equation(3,3) { 
      
      pi                = KW_PI;

      theta             = 0.5; 

      C1                = 1.0;
      C2                = 1.0;
      K                 = 20.0;

      Cs                = 0.1;
      
      Ii                = 0.25;
      Iii               = 0.1;
      Iij               = 0.05;

      f1.newsize(4);
      f2.newsize(4);
      f3.newsize(4);

      settings->Get("variational multscale eddy viscosity",&variational_multiscale_eddy_viscosity);
      settings->Get("smagorinsky eddy viscosity",&variational_multiscale_eddy_viscosity);
    }


    ~EquationNS_cG1cG1(){}


    //--------- Equation definition -----------
    inline
    void jacobian( int el, int gpt, int tst_bf, int trial_bf, int comp ) {}
    
    //--------- Load definition --------------
    inline
    void residual( int el, int gpt, int tst_bf ){}
  
  
    //--------- Load definition for exact integration on tetrahedrons --------------
    inline
    void jacobianExactInt( MV_Vector<real>& jac, int& tst_bf, int& trial_bf, 
			   int& j_eq, int& el, Grid* grid ){}



    //--------- Load definition for exact integration on tetrahedrons --------------
    inline
    void residualExactInt( MV_ColMat<real>& bres, int& bf, int& el, Grid* grid ){

      FET4n3D* FE=NULL;
      FE = &(grid->FET4);
      if ( FE->getElmNo() != el ) FE->refill( el );

      if ( bf == 0 ){
	h = 2.0*FE->circumRadius();
	
	x(0) = FE->x1();
	x(1) = FE->x2();
	x(2) = FE->x3();
	x(3) = FE->x4();
	
	y(0) = FE->y1();
	y(1) = FE->y2();
	y(2) = FE->y3();
	y(3) = FE->y4();
	
	z(0) = FE->z1();
	z(1) = FE->z2();
	z(2) = FE->z3();
	z(3) = FE->z4();
	
	uprev_TS1(0)     = COEFF(0) -> GetNodalValue( FE->n1() );    
	uprev_TS1(1)     = COEFF(0) -> GetNodalValue( FE->n2() );
	uprev_TS1(2)     = COEFF(0) -> GetNodalValue( FE->n3() );
	uprev_TS1(3)     = COEFF(0) -> GetNodalValue( FE->n4() );
	FE->evalGradientLoc(uprev_TS1,uprev_TS1Grad);
	uprev_TS2(0)     = COEFF(1) -> GetNodalValue( FE->n1() );
	uprev_TS2(1)     = COEFF(1) -> GetNodalValue( FE->n2() );
	uprev_TS2(2)     = COEFF(1) -> GetNodalValue( FE->n3() );
	uprev_TS2(3)     = COEFF(1) -> GetNodalValue( FE->n4() );
	FE->evalGradientLoc(uprev_TS2,uprev_TS2Grad);
	uprev_TS3(0)     = COEFF(2) -> GetNodalValue( FE->n1() );
	uprev_TS3(1)     = COEFF(2) -> GetNodalValue( FE->n2() );
	uprev_TS3(2)     = COEFF(2) -> GetNodalValue( FE->n3() );
	uprev_TS3(3)     = COEFF(2) -> GetNodalValue( FE->n4() );
	FE->evalGradientLoc(uprev_TS3,uprev_TS3Grad);
	//pprev_TS(0)      = COEFF(3) -> GetNodalValue( FE->n1() );
	//pprev_TS(1)      = COEFF(3) -> GetNodalValue( FE->n2() );
	//pprev_TS(2)      = COEFF(3) -> GetNodalValue( FE->n3() );
	//pprev_TS(3)      = COEFF(3) -> GetNodalValue( FE->n4() );
	//FE->evalGradientLoc(pprev_TS,pprev_TSGrad);
	
	//uprev_NL1(0)     = COEFF(4) -> GetNodalValue( FE->n1() );    
	//uprev_NL1(1)     = COEFF(4) -> GetNodalValue( FE->n2() );
	//uprev_NL1(2)     = COEFF(4) -> GetNodalValue( FE->n3() );
	//uprev_NL1(3)     = COEFF(4) -> GetNodalValue( FE->n4() );
	//FE->evalGradientLoc(uprev_NL1,uprev_NL1Grad);
	//uprev_NL2(0)     = COEFF(5) -> GetNodalValue( FE->n1() );
	//uprev_NL2(1)     = COEFF(5) -> GetNodalValue( FE->n2() );
	//uprev_NL2(2)     = COEFF(5) -> GetNodalValue( FE->n3() );
	//uprev_NL2(3)     = COEFF(5) -> GetNodalValue( FE->n4() );
	//FE->evalGradientLoc(uprev_NL2,uprev_NL2Grad);
	//uprev_NL3(0)     = COEFF(6) -> GetNodalValue( FE->n1() );
	//uprev_NL3(1)     = COEFF(6) -> GetNodalValue( FE->n2() );
	//uprev_NL3(2)     = COEFF(6) -> GetNodalValue( FE->n3() );
	//uprev_NL3(3)     = COEFF(6) -> GetNodalValue( FE->n4() );
	//FE->evalGradientLoc(uprev_NL3,uprev_NL3Grad);
	//pprev_NL(0)      = COEFF(7) -> GetNodalValue( FE->n1() );
	//pprev_NL(1)      = COEFF(7) -> GetNodalValue( FE->n2() );
	//pprev_NL(2)      = COEFF(7) -> GetNodalValue( FE->n3() );
	//pprev_NL(3)      = COEFF(7) -> GetNodalValue( FE->n4() );
	//FE->evalGradientLoc(pprev_NL,pprev_NLGrad);
	
	//u1(0)     = COEFF(8)  -> GetNodalValue( FE->n1() );
	//u1(1)     = COEFF(8)  -> GetNodalValue( FE->n2() );
	//u1(2)     = COEFF(8)  -> GetNodalValue( FE->n3() );
	//u1(3)     = COEFF(8)  -> GetNodalValue( FE->n4() );
	//FE->evalGradientLoc(u1,u1Grad);
	//u2(0)     = COEFF(9) -> GetNodalValue( FE->n1() );
	//u2(1)     = COEFF(9) -> GetNodalValue( FE->n2() );
	//u2(2)     = COEFF(9) -> GetNodalValue( FE->n3() );
	//u2(3)     = COEFF(9) -> GetNodalValue( FE->n4() );
	//FE->evalGradientLoc(u2,u2Grad);
	//u3(0)     = COEFF(10) -> GetNodalValue( FE->n1() );
	//u3(1)     = COEFF(10) -> GetNodalValue( FE->n2() );
	//u3(2)     = COEFF(10) -> GetNodalValue( FE->n3() );
	//u3(3)     = COEFF(10) -> GetNodalValue( FE->n4() );
	//FE->evalGradientLoc(u3,u3Grad);
	//p(0)      = COEFF(11) -> GetNodalValue( FE->n1() );
	//p(1)      = COEFF(11) -> GetNodalValue( FE->n2() );
	//p(2)      = COEFF(11) -> GetNodalValue( FE->n3() );
	//p(3)      = COEFF(11) -> GetNodalValue( FE->n4() );
	//FE->evalGradientLoc(p,pGrad);
	
	dt           = COEFF(12) -> Eval();
	//t            = COEFF(13) -> Eval();
	Re           = COEFF(14) -> Eval();
	
	f1(0) = (1.0/Re) * 32.0 * ( y(0)*(1.0-y(0)) + z(0)*(1.0-z(0)) );
	f1(1) = (1.0/Re) * 32.0 * ( y(1)*(1.0-y(1)) + z(1)*(1.0-z(1)) );
	f1(2) = (1.0/Re) * 32.0 * ( y(2)*(1.0-y(2)) + z(2)*(1.0-z(2)) );
	f1(3) = (1.0/Re) * 32.0 * ( y(3)*(1.0-y(3)) + z(3)*(1.0-z(3)) );
	f1 = 0.0;

	f2(0) = 0.001 * sin(10.0*pi*x(0));
	f2(1) = 0.001 * sin(10.0*pi*x(1));
	f2(2) = 0.001 * sin(10.0*pi*x(2));
	f2(3) = 0.001 * sin(10.0*pi*x(3));
	f2 = 0.0;

	f3(0) = 0.001 * sin(10.0*pi*x(0));
	f3(1) = 0.001 * sin(10.0*pi*x(1));
	f3(2) = 0.001 * sin(10.0*pi*x(2));
	f3(3) = 0.001 * sin(10.0*pi*x(3));
	f3 = 0.0;
	
	uTS1 = uprev_TS1(0) + uprev_TS1(1) + uprev_TS1(2) + uprev_TS1(3);
	uTS2 = uprev_TS2(0) + uprev_TS2(1) + uprev_TS2(2) + uprev_TS2(3);
	uTS3 = uprev_TS3(0) + uprev_TS3(1) + uprev_TS3(2) + uprev_TS3(3);

	//u1sum = u1(0) + u1(1) + u1(2) + u1(3);
	//u2sum = u2(0) + u2(1) + u2(2) + u2(3);
	//u3sum = u3(0) + u3(1) + u3(2) + u3(3);
	//normU = sqrt( sqr(u1sum) + sqr(u2sum) + sqr(u3sum) ) / 4;
	
	//divU = u1Grad(0) + u2Grad(1) + u3Grad(2);
	
	if ( h * Re > 1.0 ){ 
	  //delta1 = 0.5 / sqrt( 1.0/sqr(dt) + sqr(normU/h) );
	  delta1 = C1 * h;
	}
	else{ 
	  delta1 = C1 * sqr(h);
	}
	if ( h * Re > 1.0 ){ 
	  delta2 = C2 * h;
	}
	else{ 
	  delta2 = C2 * sqr(h);
	}
	/*
	if ( divU > K ){ 
	  r = 0.5 * divU;
	}
	else{ 
	  r = 0.0;
	}
	*/
	r = 0.0;
      }
      
      if ( bf == 0 ){ 
	uv1 = Iii;
	uv2 = uv3 = uv4 = Iij;
	vx = FE->dN1x();
	vy = FE->dN1y();
	vz = FE->dN1z();
      }else if ( bf == 1 ){
	uv2 = Iii;
	uv1 = uv3 = uv4 = Iij;
	vx = FE->dN2x();
	vy = FE->dN2y();
	vz = FE->dN2z();
      }else if ( bf == 2 ){
	uv3 = Iii;
	uv1 = uv2 = uv4 = Iij;
	vx = FE->dN3x();
	vy = FE->dN3y();
	vz = FE->dN3z();
      }else if ( bf == 3 ){
	uv4 = Iii;
	uv1 = uv2 = uv3 = Iij;
	vx = FE->dN4x();
	vy = FE->dN4y();
	vz = FE->dN4z();
      }	

      if ( bf == 0 ){
	uTS1uTS1 = uprev_TS1(0)*uprev_TS1(0)*Iii+uprev_TS1(0)*uprev_TS1(1)*Iij+uprev_TS1(0)*uprev_TS1(2)*Iij+uprev_TS1(0)*uprev_TS1(3)*Iij
	         + uprev_TS1(1)*uprev_TS1(0)*Iij+uprev_TS1(1)*uprev_TS1(1)*Iii+uprev_TS1(1)*uprev_TS1(2)*Iij+uprev_TS1(1)*uprev_TS1(3)*Iij
	         + uprev_TS1(2)*uprev_TS1(0)*Iij+uprev_TS1(2)*uprev_TS1(1)*Iii+uprev_TS1(2)*uprev_TS1(2)*Iij+uprev_TS1(2)*uprev_TS1(3)*Iij
  	         + uprev_TS1(3)*uprev_TS1(0)*Iij+uprev_TS1(3)*uprev_TS1(1)*Iii+uprev_TS1(3)*uprev_TS1(2)*Iij+uprev_TS1(3)*uprev_TS1(3)*Iij;
	uTS1uTS2 = uprev_TS1(0)*uprev_TS2(0)*Iii+uprev_TS1(0)*uprev_TS2(1)*Iij+uprev_TS1(0)*uprev_TS2(2)*Iij+uprev_TS1(0)*uprev_TS2(3)*Iij
	         + uprev_TS1(1)*uprev_TS2(0)*Iij+uprev_TS1(1)*uprev_TS2(1)*Iii+uprev_TS1(1)*uprev_TS2(2)*Iij+uprev_TS1(1)*uprev_TS2(3)*Iij
	         + uprev_TS1(2)*uprev_TS2(0)*Iij+uprev_TS1(2)*uprev_TS2(1)*Iii+uprev_TS1(2)*uprev_TS2(2)*Iij+uprev_TS1(2)*uprev_TS2(3)*Iij
  	         + uprev_TS1(3)*uprev_TS2(0)*Iij+uprev_TS1(3)*uprev_TS2(1)*Iii+uprev_TS1(3)*uprev_TS2(2)*Iij+uprev_TS1(3)*uprev_TS2(3)*Iij;
	uTS1uTS3 = uprev_TS1(0)*uprev_TS3(0)*Iii+uprev_TS1(0)*uprev_TS3(1)*Iij+uprev_TS1(0)*uprev_TS3(2)*Iij+uprev_TS1(0)*uprev_TS3(3)*Iij
	         + uprev_TS1(1)*uprev_TS3(0)*Iij+uprev_TS1(1)*uprev_TS3(1)*Iii+uprev_TS1(1)*uprev_TS3(2)*Iij+uprev_TS1(1)*uprev_TS3(3)*Iij
	         + uprev_TS1(2)*uprev_TS3(0)*Iij+uprev_TS1(2)*uprev_TS3(1)*Iii+uprev_TS1(2)*uprev_TS3(2)*Iij+uprev_TS1(2)*uprev_TS3(3)*Iij
  	         + uprev_TS1(3)*uprev_TS3(0)*Iij+uprev_TS1(3)*uprev_TS3(1)*Iii+uprev_TS1(3)*uprev_TS3(2)*Iij+uprev_TS1(3)*uprev_TS3(3)*Iij;
	uTS2uTS1 = uprev_TS2(0)*uprev_TS1(0)*Iii+uprev_TS2(0)*uprev_TS1(1)*Iij+uprev_TS2(0)*uprev_TS1(2)*Iij+uprev_TS2(0)*uprev_TS1(3)*Iij
	         + uprev_TS2(1)*uprev_TS1(0)*Iij+uprev_TS2(1)*uprev_TS1(1)*Iii+uprev_TS2(1)*uprev_TS1(2)*Iij+uprev_TS2(1)*uprev_TS1(3)*Iij
	         + uprev_TS2(2)*uprev_TS1(0)*Iij+uprev_TS2(2)*uprev_TS1(1)*Iii+uprev_TS2(2)*uprev_TS1(2)*Iij+uprev_TS2(2)*uprev_TS1(3)*Iij
  	         + uprev_TS2(3)*uprev_TS1(0)*Iij+uprev_TS2(3)*uprev_TS1(1)*Iii+uprev_TS2(3)*uprev_TS1(2)*Iij+uprev_TS2(3)*uprev_TS1(3)*Iij;
	uTS2uTS2 = uprev_TS2(0)*uprev_TS2(0)*Iii+uprev_TS2(0)*uprev_TS2(1)*Iij+uprev_TS2(0)*uprev_TS2(2)*Iij+uprev_TS2(0)*uprev_TS2(3)*Iij
	         + uprev_TS2(1)*uprev_TS2(0)*Iij+uprev_TS2(1)*uprev_TS2(1)*Iii+uprev_TS2(1)*uprev_TS2(2)*Iij+uprev_TS2(1)*uprev_TS2(3)*Iij
	         + uprev_TS2(2)*uprev_TS2(0)*Iij+uprev_TS2(2)*uprev_TS2(1)*Iii+uprev_TS2(2)*uprev_TS2(2)*Iij+uprev_TS2(2)*uprev_TS2(3)*Iij
  	         + uprev_TS2(3)*uprev_TS2(0)*Iij+uprev_TS2(3)*uprev_TS2(1)*Iii+uprev_TS2(3)*uprev_TS2(2)*Iij+uprev_TS2(3)*uprev_TS2(3)*Iij;
	uTS2uTS3 = uprev_TS2(0)*uprev_TS3(0)*Iii+uprev_TS2(0)*uprev_TS3(1)*Iij+uprev_TS2(0)*uprev_TS3(2)*Iij+uprev_TS2(0)*uprev_TS3(3)*Iij
	         + uprev_TS2(1)*uprev_TS3(0)*Iij+uprev_TS2(1)*uprev_TS3(1)*Iii+uprev_TS2(1)*uprev_TS3(2)*Iij+uprev_TS2(1)*uprev_TS3(3)*Iij
	         + uprev_TS2(2)*uprev_TS3(0)*Iij+uprev_TS2(2)*uprev_TS3(1)*Iii+uprev_TS2(2)*uprev_TS3(2)*Iij+uprev_TS2(2)*uprev_TS3(3)*Iij
  	         + uprev_TS2(3)*uprev_TS3(0)*Iij+uprev_TS2(3)*uprev_TS3(1)*Iii+uprev_TS2(3)*uprev_TS3(2)*Iij+uprev_TS2(3)*uprev_TS3(3)*Iij;
	uTS3uTS1 = uprev_TS3(0)*uprev_TS1(0)*Iii+uprev_TS3(0)*uprev_TS1(1)*Iij+uprev_TS3(0)*uprev_TS1(2)*Iij+uprev_TS3(0)*uprev_TS1(3)*Iij
	         + uprev_TS3(1)*uprev_TS1(0)*Iij+uprev_TS3(1)*uprev_TS1(1)*Iii+uprev_TS3(1)*uprev_TS1(2)*Iij+uprev_TS3(1)*uprev_TS1(3)*Iij
	         + uprev_TS3(2)*uprev_TS1(0)*Iij+uprev_TS3(2)*uprev_TS1(1)*Iii+uprev_TS3(2)*uprev_TS1(2)*Iij+uprev_TS3(2)*uprev_TS1(3)*Iij
  	         + uprev_TS3(3)*uprev_TS1(0)*Iij+uprev_TS3(3)*uprev_TS1(1)*Iii+uprev_TS3(3)*uprev_TS1(2)*Iij+uprev_TS3(3)*uprev_TS1(3)*Iij;
	uTS3uTS2 = uprev_TS3(0)*uprev_TS2(0)*Iii+uprev_TS3(0)*uprev_TS2(1)*Iij+uprev_TS3(0)*uprev_TS2(2)*Iij+uprev_TS3(0)*uprev_TS2(3)*Iij
	         + uprev_TS3(1)*uprev_TS2(0)*Iij+uprev_TS3(1)*uprev_TS2(1)*Iii+uprev_TS3(1)*uprev_TS2(2)*Iij+uprev_TS3(1)*uprev_TS2(3)*Iij
	         + uprev_TS3(2)*uprev_TS2(0)*Iij+uprev_TS3(2)*uprev_TS2(1)*Iii+uprev_TS3(2)*uprev_TS2(2)*Iij+uprev_TS3(2)*uprev_TS2(3)*Iij
  	         + uprev_TS3(3)*uprev_TS2(0)*Iij+uprev_TS3(3)*uprev_TS2(1)*Iii+uprev_TS3(3)*uprev_TS2(2)*Iij+uprev_TS3(3)*uprev_TS2(3)*Iij;
	uTS3uTS3 = uprev_TS3(0)*uprev_TS3(0)*Iii+uprev_TS3(0)*uprev_TS3(1)*Iij+uprev_TS3(0)*uprev_TS3(2)*Iij+uprev_TS3(0)*uprev_TS3(3)*Iij
	         + uprev_TS3(1)*uprev_TS3(0)*Iij+uprev_TS3(1)*uprev_TS3(1)*Iii+uprev_TS3(1)*uprev_TS3(2)*Iij+uprev_TS3(1)*uprev_TS3(3)*Iij
	         + uprev_TS3(2)*uprev_TS3(0)*Iij+uprev_TS3(2)*uprev_TS3(1)*Iii+uprev_TS3(2)*uprev_TS3(2)*Iij+uprev_TS3(2)*uprev_TS3(3)*Iij
  	         + uprev_TS3(3)*uprev_TS3(0)*Iij+uprev_TS3(3)*uprev_TS3(1)*Iii+uprev_TS3(3)*uprev_TS3(2)*Iij+uprev_TS3(3)*uprev_TS3(3)*Iij;

	f1u1 = f1(0)*uprev_TS1(0)*Iii + f1(0)*uprev_TS1(1)*Iij + f1(0)*uprev_TS1(2)*Iij + f1(0)*uprev_TS1(3)*Iij
	     + f1(1)*uprev_TS1(0)*Iii + f1(1)*uprev_TS1(1)*Iij + f1(1)*uprev_TS1(2)*Iij + f1(1)*uprev_TS1(3)*Iij
	     + f1(2)*uprev_TS1(0)*Iii + f1(2)*uprev_TS1(1)*Iij + f1(2)*uprev_TS1(2)*Iij + f1(2)*uprev_TS1(3)*Iij
	     + f1(3)*uprev_TS1(0)*Iii + f1(3)*uprev_TS1(1)*Iij + f1(3)*uprev_TS1(2)*Iij + f1(3)*uprev_TS1(3)*Iij;
	f1u2 = f1(0)*uprev_TS2(0)*Iii + f1(0)*uprev_TS2(1)*Iij + f1(0)*uprev_TS2(2)*Iij + f1(0)*uprev_TS2(3)*Iij
	     + f1(1)*uprev_TS2(0)*Iii + f1(1)*uprev_TS2(1)*Iij + f1(1)*uprev_TS2(2)*Iij + f1(1)*uprev_TS2(3)*Iij
	     + f1(2)*uprev_TS2(0)*Iii + f1(2)*uprev_TS2(1)*Iij + f1(2)*uprev_TS2(2)*Iij + f1(2)*uprev_TS2(3)*Iij
	     + f1(3)*uprev_TS2(0)*Iii + f1(3)*uprev_TS2(1)*Iij + f1(3)*uprev_TS2(2)*Iij + f1(3)*uprev_TS2(3)*Iij;
	f1u3 = f1(0)*uprev_TS3(0)*Iii + f1(0)*uprev_TS3(1)*Iij + f1(0)*uprev_TS3(2)*Iij + f1(0)*uprev_TS3(3)*Iij
	     + f1(1)*uprev_TS3(0)*Iii + f1(1)*uprev_TS3(1)*Iij + f1(1)*uprev_TS3(2)*Iij + f1(1)*uprev_TS3(3)*Iij
	     + f1(2)*uprev_TS3(0)*Iii + f1(2)*uprev_TS3(1)*Iij + f1(2)*uprev_TS3(2)*Iij + f1(2)*uprev_TS3(3)*Iij
	     + f1(3)*uprev_TS3(0)*Iii + f1(3)*uprev_TS3(1)*Iij + f1(3)*uprev_TS3(2)*Iij + f1(3)*uprev_TS3(3)*Iij;
	f2u1 = f2(0)*uprev_TS1(0)*Iii + f2(0)*uprev_TS1(1)*Iij + f2(0)*uprev_TS1(2)*Iij + f2(0)*uprev_TS1(3)*Iij
	     + f2(1)*uprev_TS1(0)*Iii + f2(1)*uprev_TS1(1)*Iij + f2(1)*uprev_TS1(2)*Iij + f2(1)*uprev_TS1(3)*Iij
	     + f2(2)*uprev_TS1(0)*Iii + f2(2)*uprev_TS1(1)*Iij + f2(2)*uprev_TS1(2)*Iij + f2(2)*uprev_TS1(3)*Iij
	     + f2(3)*uprev_TS1(0)*Iii + f2(3)*uprev_TS1(1)*Iij + f2(3)*uprev_TS1(2)*Iij + f2(3)*uprev_TS1(3)*Iij;
	f2u2 = f2(0)*uprev_TS2(0)*Iii + f2(0)*uprev_TS2(1)*Iij + f2(0)*uprev_TS2(2)*Iij + f2(0)*uprev_TS2(3)*Iij
	     + f2(1)*uprev_TS2(0)*Iii + f2(1)*uprev_TS2(1)*Iij + f2(1)*uprev_TS2(2)*Iij + f2(1)*uprev_TS2(3)*Iij
	     + f2(2)*uprev_TS2(0)*Iii + f2(2)*uprev_TS2(1)*Iij + f2(2)*uprev_TS2(2)*Iij + f2(2)*uprev_TS2(3)*Iij
	     + f2(3)*uprev_TS2(0)*Iii + f2(3)*uprev_TS2(1)*Iij + f2(3)*uprev_TS2(2)*Iij + f2(3)*uprev_TS2(3)*Iij;
	f2u3 = f2(0)*uprev_TS3(0)*Iii + f2(0)*uprev_TS3(1)*Iij + f2(0)*uprev_TS3(2)*Iij + f2(0)*uprev_TS3(3)*Iij
	     + f2(1)*uprev_TS3(0)*Iii + f2(1)*uprev_TS3(1)*Iij + f2(1)*uprev_TS3(2)*Iij + f2(1)*uprev_TS3(3)*Iij
	     + f2(2)*uprev_TS3(0)*Iii + f2(2)*uprev_TS3(1)*Iij + f2(2)*uprev_TS3(2)*Iij + f2(2)*uprev_TS3(3)*Iij
	     + f2(3)*uprev_TS3(0)*Iii + f2(3)*uprev_TS3(1)*Iij + f2(3)*uprev_TS3(2)*Iij + f2(3)*uprev_TS3(3)*Iij;
	f3u1 = f3(0)*uprev_TS1(0)*Iii + f3(0)*uprev_TS1(1)*Iij + f3(0)*uprev_TS1(2)*Iij + f3(0)*uprev_TS1(3)*Iij
	     + f3(1)*uprev_TS1(0)*Iii + f3(1)*uprev_TS1(1)*Iij + f3(1)*uprev_TS1(2)*Iij + f3(1)*uprev_TS1(3)*Iij
	     + f3(2)*uprev_TS1(0)*Iii + f3(2)*uprev_TS1(1)*Iij + f3(2)*uprev_TS1(2)*Iij + f3(2)*uprev_TS1(3)*Iij
	     + f3(3)*uprev_TS1(0)*Iii + f3(3)*uprev_TS1(1)*Iij + f3(3)*uprev_TS1(2)*Iij + f3(3)*uprev_TS1(3)*Iij;
	f3u2 = f3(0)*uprev_TS2(0)*Iii + f3(0)*uprev_TS2(1)*Iij + f3(0)*uprev_TS2(2)*Iij + f3(0)*uprev_TS2(3)*Iij
	     + f3(1)*uprev_TS2(0)*Iii + f3(1)*uprev_TS2(1)*Iij + f3(1)*uprev_TS2(2)*Iij + f3(1)*uprev_TS2(3)*Iij
	     + f3(2)*uprev_TS2(0)*Iii + f3(2)*uprev_TS2(1)*Iij + f3(2)*uprev_TS2(2)*Iij + f3(2)*uprev_TS2(3)*Iij
	     + f3(3)*uprev_TS2(0)*Iii + f3(3)*uprev_TS2(1)*Iij + f3(3)*uprev_TS2(2)*Iij + f3(3)*uprev_TS2(3)*Iij;
	f3u3 = f3(0)*uprev_TS3(0)*Iii + f3(0)*uprev_TS3(1)*Iij + f3(0)*uprev_TS3(2)*Iij + f3(0)*uprev_TS3(3)*Iij
	     + f3(1)*uprev_TS3(0)*Iii + f3(1)*uprev_TS3(1)*Iij + f3(1)*uprev_TS3(2)*Iij + f3(1)*uprev_TS3(3)*Iij
	     + f3(2)*uprev_TS3(0)*Iii + f3(2)*uprev_TS3(1)*Iij + f3(2)*uprev_TS3(2)*Iij + f3(2)*uprev_TS3(3)*Iij
	     + f3(3)*uprev_TS3(0)*Iii + f3(3)*uprev_TS3(1)*Iij + f3(3)*uprev_TS3(2)*Iij + f3(3)*uprev_TS3(3)*Iij;

      }

      f_uDv_1 = theta * (vx*f1u1 + vy*f1u2 + vz*f1u3);
      f_uDv_2 = theta * (vx*f2u1 + vy*f2u2 + vz*f2u3);
      f_uDv_3 = theta * (vx*f3u1 + vy*f3u2 + vz*f3u3);

      f_v_1 = f1(0)*uv1+f1(1)*uv2+f1(2)*uv3+f1(3)*uv4;
      f_v_2 = f2(0)*uv1+f2(1)*uv2+f2(2)*uv3+f2(3)*uv4;
      f_v_3 = f3(0)*uv1+f3(1)*uv2+f3(2)*uv3+f3(3)*uv4;

      f_Dq = ( f1(0)+f1(1)+f1(2)+f1(3) )*Ii*vx;
           + ( f2(0)+f2(1)+f2(2)+f2(3) )*Ii*vy;
	   + ( f3(0)+f3(1)+f3(2)+f3(3) )*Ii*vz;


      uDu_Dq = theta*theta*( uTS1*uprev_TS1Grad(0)+uTS2*uprev_TS1Grad(1)+uTS3*uprev_TS1Grad(2) )*Ii*vx
	     + theta*theta*( uTS1*uprev_TS2Grad(0)+uTS2*uprev_TS2Grad(1)+uTS3*uprev_TS2Grad(2) )*Ii*vy
	     + theta*theta*( uTS1*uprev_TS3Grad(0)+uTS2*uprev_TS3Grad(1)+uTS3*uprev_TS3Grad(2) )*Ii*vz;

      uDu_uDv_1 = theta*uprev_TS1Grad(0) * theta*theta*(vx*uTS1uTS1 + vy*uTS1uTS2 + vz*uTS1uTS3)  
	        + theta*uprev_TS1Grad(1) * theta*theta*(vx*uTS2uTS1 + vy*uTS2uTS2 + vz*uTS2uTS3)
	        + theta*uprev_TS1Grad(2) * theta*theta*(vx*uTS3uTS1 + vy*uTS3uTS2 + vz*uTS3uTS3);

      uDu_uDv_2 = theta*uprev_TS2Grad(0) * theta*theta*(vx*uTS1uTS1 + vy*uTS1uTS2 + vz*uTS1uTS3)  
	        + theta*uprev_TS2Grad(1) * theta*theta*(vx*uTS2uTS1 + vy*uTS2uTS2 + vz*uTS2uTS3)
	        + theta*uprev_TS2Grad(2) * theta*theta*(vx*uTS3uTS1 + vy*uTS3uTS2 + vz*uTS3uTS3);

      uDu_uDv_3 = theta*uprev_TS3Grad(0) * theta*theta*(vx*uTS1uTS1 + vy*uTS1uTS2 + vz*uTS1uTS3)  
	        + theta*uprev_TS3Grad(1) * theta*theta*(vx*uTS2uTS1 + vy*uTS2uTS2 + vz*uTS2uTS3)
	        + theta*uprev_TS3Grad(2) * theta*theta*(vx*uTS3uTS1 + vy*uTS3uTS2 + vz*uTS3uTS3);


      NL1   = theta*(uprev_TS1(0)*uv1+uprev_TS1(1)*uv2+uprev_TS1(2)*uv3+uprev_TS1(3)*uv4) * theta*uprev_TS1Grad(0)
	    + theta*(uprev_TS2(0)*uv1+uprev_TS2(1)*uv2+uprev_TS2(2)*uv3+uprev_TS2(3)*uv4) * theta*uprev_TS1Grad(1)
	    + theta*(uprev_TS3(0)*uv1+uprev_TS3(1)*uv2+uprev_TS3(2)*uv3+uprev_TS3(3)*uv4) * theta*uprev_TS1Grad(2);

      NL2   = theta*(uprev_TS1(0)*uv1+uprev_TS1(1)*uv2+uprev_TS1(2)*uv3+uprev_TS1(3)*uv4) * theta*uprev_TS2Grad(0)
	    + theta*(uprev_TS2(0)*uv1+uprev_TS2(1)*uv2+uprev_TS2(2)*uv3+uprev_TS2(3)*uv4) * theta*uprev_TS2Grad(1)
	    + theta*(uprev_TS3(0)*uv1+uprev_TS3(1)*uv2+uprev_TS3(2)*uv3+uprev_TS3(3)*uv4) * theta*uprev_TS2Grad(2);

      NL3   = theta*(uprev_TS1(0)*uv1+uprev_TS1(1)*uv2+uprev_TS1(2)*uv3+uprev_TS1(3)*uv4) * theta*uprev_TS3Grad(0)
	    + theta*(uprev_TS2(0)*uv1+uprev_TS2(1)*uv2+uprev_TS2(2)*uv3+uprev_TS2(3)*uv4) * theta*uprev_TS3Grad(1)
	    + theta*(uprev_TS3(0)*uv1+uprev_TS3(1)*uv2+uprev_TS3(2)*uv3+uprev_TS3(3)*uv4) * theta*uprev_TS3Grad(2);


      if ( smagorinsky_eddy_viscosity ){
	if ( bf == 0 ){
	  S_abs = sqrt( sqr(uprev_TS1Grad(0)) + sqr(uprev_TS1Grad(1)) + sqr(uprev_TS1Grad(2))
		      + sqr(uprev_TS2Grad(0)) + sqr(uprev_TS2Grad(1)) + sqr(uprev_TS2Grad(2))
	              + sqr(uprev_TS3Grad(0)) + sqr(uprev_TS3Grad(1)) + sqr(uprev_TS3Grad(2)) );
	  nu_T  = pow(Cs*h,2.0) * S_abs; 
	  nu = (1.0/Re) + nu_T;
	}
      } else{
	nu = (1.0/Re);
      }
      LAP1 = nu * theta * ( uprev_TS1Grad(0)*vx + uprev_TS1Grad(1)*vy + uprev_TS1Grad(2)*vz );
      LAP2 = nu * theta * ( uprev_TS2Grad(0)*vx + uprev_TS2Grad(1)*vy + uprev_TS2Grad(2)*vz );
      LAP3 = nu * theta * ( uprev_TS3Grad(0)*vx + uprev_TS3Grad(1)*vy + uprev_TS3Grad(2)*vz );

      uprev_TS_v_1 = uprev_TS1(0)*uv1+uprev_TS1(1)*uv2+uprev_TS1(2)*uv3+uprev_TS1(3)*uv4;
      uprev_TS_v_2 = uprev_TS2(0)*uv1+uprev_TS2(1)*uv2+uprev_TS2(2)*uv3+uprev_TS2(3)*uv4;
      uprev_TS_v_3 = uprev_TS3(0)*uv1+uprev_TS3(1)*uv2+uprev_TS3(2)*uv3+uprev_TS3(3)*uv4;
      
      SD1 = uDu_uDv_1 - f_uDv_1;
      SD2 = uDu_uDv_2 - f_uDv_2;
      SD3 = uDu_uDv_3 - f_uDv_3;
      
      if ( bf == 0 ){
	divU = theta * (uprev_TS1Grad(0) + uprev_TS2Grad(1) + uprev_TS3Grad(2));
      }      

      LSdiv1 = divU * vx;
      LSdiv2 = divU * vy;
      LSdiv3 = divU * vz;
            
      bres(0,bf) = uprev_TS_v_1/dt - LAP1 - NL1 + f_v_1 - delta1*SD1 - delta2*LSdiv1;
      bres(1,bf) = uprev_TS_v_2/dt - LAP2 - NL2 + f_v_2 - delta1*SD2 - delta2*LSdiv2;
      bres(2,bf) = uprev_TS_v_3/dt - LAP3 - NL3 + f_v_3 - delta1*SD3 - delta2*LSdiv3;
      bres(3,bf) = delta1*(uDu_Dq-f_Dq) + (divU*Ii); 
    }
  

  inline
  void MatrixVectorProductExactInt( MV_ColMat<real>& bres, int& bf, int& el, FET4n3D& FE ){

      if ( bf == 0 ){
	h = 2.0*FE.circumRadius();
	
	x(0) = FE.x1();
	x(1) = FE.x2();
	x(2) = FE.x3();
	x(3) = FE.x4();
	
	y(0) = FE.y1();
	y(1) = FE.y2();
	y(2) = FE.y3();
	y(3) = FE.y4();
	
	z(0) = FE.z1();
	z(1) = FE.z2();
	z(2) = FE.z3();
	z(3) = FE.z4();
	
	uprev_TS1(0)     = COEFF(0) -> GetNodalValue( FE.n1() );    
	uprev_TS1(1)     = COEFF(0) -> GetNodalValue( FE.n2() );
	uprev_TS1(2)     = COEFF(0) -> GetNodalValue( FE.n3() );
	uprev_TS1(3)     = COEFF(0) -> GetNodalValue( FE.n4() );
	FE.evalGradientLoc(uprev_TS1,uprev_TS1Grad);
	uprev_TS2(0)     = COEFF(1) -> GetNodalValue( FE.n1() );
	uprev_TS2(1)     = COEFF(1) -> GetNodalValue( FE.n2() );
	uprev_TS2(2)     = COEFF(1) -> GetNodalValue( FE.n3() );
	uprev_TS2(3)     = COEFF(1) -> GetNodalValue( FE.n4() );
	FE.evalGradientLoc(uprev_TS2,uprev_TS2Grad);
	uprev_TS3(0)     = COEFF(2) -> GetNodalValue( FE.n1() );
	uprev_TS3(1)     = COEFF(2) -> GetNodalValue( FE.n2() );
	uprev_TS3(2)     = COEFF(2) -> GetNodalValue( FE.n3() );
	uprev_TS3(3)     = COEFF(2) -> GetNodalValue( FE.n4() );
	FE.evalGradientLoc(uprev_TS3,uprev_TS3Grad);
	//pprev_TS(0)      = COEFF(3) -> GetNodalValue( FE.n1() );
	//pprev_TS(1)      = COEFF(3) -> GetNodalValue( FE.n2() );
	//pprev_TS(2)      = COEFF(3) -> GetNodalValue( FE.n3() );
	//pprev_TS(3)      = COEFF(3) -> GetNodalValue( FE.n4() );
	//FE.evalGradientLoc(pprev_TS,pprev_TSGrad);
	
	uprev_NL1(0)     = COEFF(4) -> GetNodalValue( FE.n1() );    
	uprev_NL1(1)     = COEFF(4) -> GetNodalValue( FE.n2() );
	uprev_NL1(2)     = COEFF(4) -> GetNodalValue( FE.n3() );
	uprev_NL1(3)     = COEFF(4) -> GetNodalValue( FE.n4() );
	//FE.evalGradientLoc(uprev_NL1,uprev_NL1Grad);
	uprev_NL2(0)     = COEFF(5) -> GetNodalValue( FE.n1() );
	uprev_NL2(1)     = COEFF(5) -> GetNodalValue( FE.n2() );
	uprev_NL2(2)     = COEFF(5) -> GetNodalValue( FE.n3() );
	uprev_NL2(3)     = COEFF(5) -> GetNodalValue( FE.n4() );
	//FE.evalGradientLoc(uprev_NL2,uprev_NL2Grad);
	uprev_NL3(0)     = COEFF(6) -> GetNodalValue( FE.n1() );
	uprev_NL3(1)     = COEFF(6) -> GetNodalValue( FE.n2() );
	uprev_NL3(2)     = COEFF(6) -> GetNodalValue( FE.n3() );
	uprev_NL3(3)     = COEFF(6) -> GetNodalValue( FE.n4() );
	//FE.evalGradientLoc(uprev_NL3,uprev_NL3Grad);
	//pprev_NL(0)      = COEFF(7) -> GetNodalValue( FE.n1() );
	//pprev_NL(1)      = COEFF(7) -> GetNodalValue( FE.n2() );
	//pprev_NL(2)      = COEFF(7) -> GetNodalValue( FE.n3() );
	//pprev_NL(3)      = COEFF(7) -> GetNodalValue( FE.n4() );
	//FE.evalGradientLoc(pprev_NL,pprev_NLGrad);
	
	u1(0)     = COEFF(8)  -> GetNodalValue( FE.n1() );
	u1(1)     = COEFF(8)  -> GetNodalValue( FE.n2() );
	u1(2)     = COEFF(8)  -> GetNodalValue( FE.n3() );
	u1(3)     = COEFF(8)  -> GetNodalValue( FE.n4() );
	FE.evalGradientLoc(u1,u1Grad);
	u2(0)     = COEFF(9) -> GetNodalValue( FE.n1() );
	u2(1)     = COEFF(9) -> GetNodalValue( FE.n2() );
	u2(2)     = COEFF(9) -> GetNodalValue( FE.n3() );
	u2(3)     = COEFF(9) -> GetNodalValue( FE.n4() );
	FE.evalGradientLoc(u2,u2Grad);
	u3(0)     = COEFF(10) -> GetNodalValue( FE.n1() );
	u3(1)     = COEFF(10) -> GetNodalValue( FE.n2() );
	u3(2)     = COEFF(10) -> GetNodalValue( FE.n3() );
	u3(3)     = COEFF(10) -> GetNodalValue( FE.n4() );
	FE.evalGradientLoc(u3,u3Grad);
	p(0)      = COEFF(11) -> GetNodalValue( FE.n1() );
	p(1)      = COEFF(11) -> GetNodalValue( FE.n2() );
	p(2)      = COEFF(11) -> GetNodalValue( FE.n3() );
	p(3)      = COEFF(11) -> GetNodalValue( FE.n4() );
	FE.evalGradientLoc(p,pGrad);
	
	dt           = COEFF(12) -> Eval();
	//t            = COEFF(13) -> Eval();
	Re           = COEFF(14) -> Eval();
	
	u1sum = u1(0) + u1(1) + u1(2) + u1(3);
	u2sum = u2(0) + u2(1) + u2(2) + u2(3);
	u3sum = u3(0) + u3(1) + u3(2) + u3(3);
	normU = sqrt( sqr(u1sum) + sqr(u2sum) + sqr(u3sum) ) / 4;
	
	divU = u1Grad(0) + u2Grad(1) + u3Grad(2);
	
	if ( h * Re > 1.0 ){ 
	  //delta1 = 0.5 / sqrt( 1.0/sqr(dt) + sqr(normU/h) );
	  delta1 = C1 * h;
	}
	else{ 
	  delta1 = C1 * sqr(h);
	}
	if ( h * Re > 1.0 ){ 
	  delta2 = C2 * h;
	}
	else{ 
	  delta2 = C2 * sqr(h);
	}
	if ( divU > K ){ 
	  r = 0.5 * divU;
	}
	else{ 
	  r = 0.0;
	}
	r = 0.0;

	f1(0) = (1.0/Re) * 32.0 * ( y(0)*(1.0-y(0)) + z(0)*(1.0-z(0)) );
	f1(1) = (1.0/Re) * 32.0 * ( y(1)*(1.0-y(1)) + z(1)*(1.0-z(1)) );
	f1(2) = (1.0/Re) * 32.0 * ( y(2)*(1.0-y(2)) + z(2)*(1.0-z(2)) );
	f1(3) = (1.0/Re) * 32.0 * ( y(3)*(1.0-y(3)) + z(3)*(1.0-z(3)) );
	f1 = 0.0;

	f2(0) = 0.001 * sin(10.0*pi*x(0));
	f2(1) = 0.001 * sin(10.0*pi*x(1));
	f2(2) = 0.001 * sin(10.0*pi*x(2));
	f2(3) = 0.001 * sin(10.0*pi*x(3));
	f2 = 0.0;

	f3(0) = 0.001 * sin(10.0*pi*x(0));
	f3(1) = 0.001 * sin(10.0*pi*x(1));
	f3(2) = 0.001 * sin(10.0*pi*x(2));
	f3(3) = 0.001 * sin(10.0*pi*x(3));
	f3 = 0.0;
      }
      
      if ( bf == 0 ){ 
	uv1 = Iii;
	uv2 = uv3 = uv4 = Iij;
	vx = FE.dN1x();
	vy = FE.dN1y();
	vz = FE.dN1z();
      }else if ( bf == 1 ){
	uv2 = Iii;
	uv1 = uv3 = uv4 = Iij;
	vx = FE.dN2x();
	vy = FE.dN2y();
	vz = FE.dN2z();
      }else if ( bf == 2 ){
	uv3 = Iii;
	uv1 = uv2 = uv4 = Iij;
	vx = FE.dN3x();
	vy = FE.dN3y();
	vz = FE.dN3z();
      }else if ( bf == 3 ){
	uv4 = Iii;
	uv1 = uv2 = uv3 = Iij;
	vx = FE.dN4x();
	vy = FE.dN4y();
	vz = FE.dN4z();
      }	

      if ( bf == 0 ){
	umid1_1 = theta*uprev_TS1(0) + (1.0-theta)*uprev_NL1(0);
	umid1_2 = theta*uprev_TS1(1) + (1.0-theta)*uprev_NL1(1);
	umid1_3 = theta*uprev_TS1(2) + (1.0-theta)*uprev_NL1(2);
	umid1_4 = theta*uprev_TS1(3) + (1.0-theta)*uprev_NL1(3);

	umid2_1 = theta*uprev_TS2(0) + (1.0-theta)*uprev_NL2(0);
	umid2_2 = theta*uprev_TS2(1) + (1.0-theta)*uprev_NL2(1);
	umid2_3 = theta*uprev_TS2(2) + (1.0-theta)*uprev_NL2(2);
	umid2_4 = theta*uprev_TS2(3) + (1.0-theta)*uprev_NL2(3);

	umid3_1 = theta*uprev_TS3(0) + (1.0-theta)*uprev_NL3(0);
	umid3_2 = theta*uprev_TS3(1) + (1.0-theta)*uprev_NL3(1);
	umid3_3 = theta*uprev_TS3(2) + (1.0-theta)*uprev_NL3(2);
	umid3_4 = theta*uprev_TS3(3) + (1.0-theta)*uprev_NL3(3);

	umid1 = umid1_1 + umid1_2 + umid1_3 + umid1_4;
	umid2 = umid2_1 + umid2_2 + umid2_3 + umid2_4;
	umid3 = umid3_1 + umid3_2 + umid3_3 + umid3_4;
	
	uNL1 = uprev_NL1(0) + uprev_NL1(1) + uprev_NL1(2) + uprev_NL1(3);
	uNL2 = uprev_NL2(0) + uprev_NL2(1) + uprev_NL2(2) + uprev_NL2(3);
	uNL3 = uprev_NL3(0) + uprev_NL3(1) + uprev_NL3(2) + uprev_NL3(3);
	
	psum = p(0) + p(1) + p(2) + p(3);

	u1u1 = umid1_1*umid1_1*Iii + umid1_1*umid1_2*Iij + umid1_1*umid1_3*Iij + umid1_1*umid1_4*Iij
	     + umid1_2*umid1_1*Iij + umid1_2*umid1_2*Iii + umid1_2*umid1_3*Iij + umid1_2*umid1_4*Iij
	     + umid1_3*umid1_1*Iij + umid1_3*umid1_2*Iij + umid1_3*umid1_3*Iii + umid1_3*umid1_4*Iij
	     + umid1_4*umid1_1*Iij + umid1_4*umid1_2*Iij + umid1_4*umid1_3*Iij + umid1_4*umid1_4*Iii;
	u1u2 = umid1_1*umid2_1*Iii + umid1_1*umid2_2*Iij + umid1_1*umid2_3*Iij + umid1_1*umid2_4*Iij
	     + umid1_2*umid2_1*Iij + umid1_2*umid2_2*Iii + umid1_2*umid2_3*Iij + umid1_2*umid2_4*Iij
	     + umid1_3*umid2_1*Iij + umid1_3*umid2_2*Iij + umid1_3*umid2_3*Iii + umid1_3*umid2_4*Iij
	     + umid1_4*umid2_1*Iij + umid1_4*umid2_2*Iij + umid1_4*umid2_3*Iij + umid1_4*umid2_4*Iii;
	u1u3 = umid1_1*umid3_1*Iii + umid1_1*umid3_2*Iij + umid1_1*umid3_3*Iij + umid1_1*umid3_4*Iij
	     + umid1_2*umid3_1*Iij + umid1_2*umid3_2*Iii + umid1_2*umid3_3*Iij + umid1_2*umid3_4*Iij
	     + umid1_3*umid3_1*Iij + umid1_3*umid3_2*Iij + umid1_3*umid3_3*Iii + umid1_3*umid3_4*Iij
	     + umid1_4*umid3_1*Iij + umid1_4*umid3_2*Iij + umid1_4*umid3_3*Iij + umid1_4*umid3_4*Iii;
	u2u1 = u1u2;
	u2u2 = umid2_1*umid2_1*Iii + umid2_1*umid2_2*Iij + umid2_1*umid2_3*Iij + umid2_1*umid2_4*Iij
	     + umid2_2*umid2_1*Iij + umid2_2*umid2_2*Iii + umid2_2*umid2_3*Iij + umid2_2*umid2_4*Iij
	     + umid2_3*umid2_1*Iij + umid2_3*umid2_2*Iij + umid2_3*umid2_3*Iii + umid2_3*umid2_4*Iij
	     + umid2_4*umid2_1*Iij + umid2_4*umid2_2*Iij + umid2_4*umid2_3*Iij + umid2_4*umid2_4*Iii;
	u2u3 = umid2_1*umid3_1*Iii + umid2_1*umid3_2*Iij + umid2_1*umid3_3*Iij + umid2_1*umid3_4*Iij
	     + umid2_2*umid3_1*Iij + umid2_2*umid3_2*Iii + umid2_2*umid3_3*Iij + umid2_2*umid3_4*Iij
	     + umid2_3*umid3_1*Iij + umid2_3*umid3_2*Iij + umid2_3*umid3_3*Iii + umid2_3*umid3_4*Iij
	     + umid2_4*umid3_1*Iij + umid2_4*umid3_2*Iij + umid2_4*umid3_3*Iij + umid2_4*umid3_4*Iii;
	u3u1 = u1u3;
	u3u2 = u2u3;
	u3u3 = umid3_1*umid3_1*Iii + umid3_1*umid3_2*Iij + umid3_1*umid3_3*Iij + umid3_1*umid3_4*Iij
	     + umid3_2*umid3_1*Iij + umid3_2*umid3_2*Iii + umid3_2*umid3_3*Iij + umid3_2*umid3_4*Iij
	     + umid3_3*umid3_1*Iij + umid3_3*umid3_2*Iij + umid3_3*umid3_3*Iii + umid3_3*umid3_4*Iij
	     + umid3_4*umid3_1*Iij + umid3_4*umid3_2*Iij + umid3_4*umid3_3*Iij + umid3_4*umid3_4*Iii;


	uNL1uTS1 = uprev_NL1(0)*uprev_TS1(0)*Iii+uprev_NL1(0)*uprev_TS1(1)*Iij+uprev_NL1(0)*uprev_TS1(2)*Iij+uprev_NL1(0)*uprev_TS1(3)*Iij
	         + uprev_NL1(1)*uprev_TS1(0)*Iij+uprev_NL1(1)*uprev_TS1(1)*Iii+uprev_NL1(1)*uprev_TS1(2)*Iij+uprev_NL1(1)*uprev_TS1(3)*Iij
	         + uprev_NL1(2)*uprev_TS1(0)*Iij+uprev_NL1(2)*uprev_TS1(1)*Iii+uprev_NL1(2)*uprev_TS1(2)*Iij+uprev_NL1(2)*uprev_TS1(3)*Iij
  	         + uprev_NL1(3)*uprev_TS1(0)*Iij+uprev_NL1(3)*uprev_TS1(1)*Iii+uprev_NL1(3)*uprev_TS1(2)*Iij+uprev_NL1(3)*uprev_TS1(3)*Iij;
	uNL1uTS2 = uprev_NL1(0)*uprev_TS2(0)*Iii+uprev_NL1(0)*uprev_TS2(1)*Iij+uprev_NL1(0)*uprev_TS2(2)*Iij+uprev_NL1(0)*uprev_TS2(3)*Iij
	         + uprev_NL1(1)*uprev_TS2(0)*Iij+uprev_NL1(1)*uprev_TS2(1)*Iii+uprev_NL1(1)*uprev_TS2(2)*Iij+uprev_NL1(1)*uprev_TS2(3)*Iij
	         + uprev_NL1(2)*uprev_TS2(0)*Iij+uprev_NL1(2)*uprev_TS2(1)*Iii+uprev_NL1(2)*uprev_TS2(2)*Iij+uprev_NL1(2)*uprev_TS2(3)*Iij
  	         + uprev_NL1(3)*uprev_TS2(0)*Iij+uprev_NL1(3)*uprev_TS2(1)*Iii+uprev_NL1(3)*uprev_TS2(2)*Iij+uprev_NL1(3)*uprev_TS2(3)*Iij;
	uNL1uTS3 = uprev_NL1(0)*uprev_TS3(0)*Iii+uprev_NL1(0)*uprev_TS3(1)*Iij+uprev_NL1(0)*uprev_TS3(2)*Iij+uprev_NL1(0)*uprev_TS3(3)*Iij
	         + uprev_NL1(1)*uprev_TS3(0)*Iij+uprev_NL1(1)*uprev_TS3(1)*Iii+uprev_NL1(1)*uprev_TS3(2)*Iij+uprev_NL1(1)*uprev_TS3(3)*Iij
	         + uprev_NL1(2)*uprev_TS3(0)*Iij+uprev_NL1(2)*uprev_TS3(1)*Iii+uprev_NL1(2)*uprev_TS3(2)*Iij+uprev_NL1(2)*uprev_TS3(3)*Iij
  	         + uprev_NL1(3)*uprev_TS3(0)*Iij+uprev_NL1(3)*uprev_TS3(1)*Iii+uprev_NL1(3)*uprev_TS3(2)*Iij+uprev_NL1(3)*uprev_TS3(3)*Iij;
	uNL2uTS1 = uprev_NL2(0)*uprev_TS1(0)*Iii+uprev_NL2(0)*uprev_TS1(1)*Iij+uprev_NL2(0)*uprev_TS1(2)*Iij+uprev_NL2(0)*uprev_TS1(3)*Iij
	         + uprev_NL2(1)*uprev_TS1(0)*Iij+uprev_NL2(1)*uprev_TS1(1)*Iii+uprev_NL2(1)*uprev_TS1(2)*Iij+uprev_NL2(1)*uprev_TS1(3)*Iij
	         + uprev_NL2(2)*uprev_TS1(0)*Iij+uprev_NL2(2)*uprev_TS1(1)*Iii+uprev_NL2(2)*uprev_TS1(2)*Iij+uprev_NL2(2)*uprev_TS1(3)*Iij
  	         + uprev_NL2(3)*uprev_TS1(0)*Iij+uprev_NL2(3)*uprev_TS1(1)*Iii+uprev_NL2(3)*uprev_TS1(2)*Iij+uprev_NL2(3)*uprev_TS1(3)*Iij;
	uNL2uTS2 = uprev_NL2(0)*uprev_TS2(0)*Iii+uprev_NL2(0)*uprev_TS2(1)*Iij+uprev_NL2(0)*uprev_TS2(2)*Iij+uprev_NL2(0)*uprev_TS2(3)*Iij
	         + uprev_NL2(1)*uprev_TS2(0)*Iij+uprev_NL2(1)*uprev_TS2(1)*Iii+uprev_NL2(1)*uprev_TS2(2)*Iij+uprev_NL2(1)*uprev_TS2(3)*Iij
	         + uprev_NL2(2)*uprev_TS2(0)*Iij+uprev_NL2(2)*uprev_TS2(1)*Iii+uprev_NL2(2)*uprev_TS2(2)*Iij+uprev_NL2(2)*uprev_TS2(3)*Iij
  	         + uprev_NL2(3)*uprev_TS2(0)*Iij+uprev_NL2(3)*uprev_TS2(1)*Iii+uprev_NL2(3)*uprev_TS2(2)*Iij+uprev_NL2(3)*uprev_TS2(3)*Iij;
	uNL2uTS3 = uprev_NL2(0)*uprev_TS3(0)*Iii+uprev_NL2(0)*uprev_TS3(1)*Iij+uprev_NL2(0)*uprev_TS3(2)*Iij+uprev_NL2(0)*uprev_TS3(3)*Iij
	         + uprev_NL2(1)*uprev_TS3(0)*Iij+uprev_NL2(1)*uprev_TS3(1)*Iii+uprev_NL2(1)*uprev_TS3(2)*Iij+uprev_NL2(1)*uprev_TS3(3)*Iij
	         + uprev_NL2(2)*uprev_TS3(0)*Iij+uprev_NL2(2)*uprev_TS3(1)*Iii+uprev_NL2(2)*uprev_TS3(2)*Iij+uprev_NL2(2)*uprev_TS3(3)*Iij
  	         + uprev_NL2(3)*uprev_TS3(0)*Iij+uprev_NL2(3)*uprev_TS3(1)*Iii+uprev_NL2(3)*uprev_TS3(2)*Iij+uprev_NL2(3)*uprev_TS3(3)*Iij;
	uNL3uTS1 = uprev_NL3(0)*uprev_TS1(0)*Iii+uprev_NL3(0)*uprev_TS1(1)*Iij+uprev_NL3(0)*uprev_TS1(2)*Iij+uprev_NL3(0)*uprev_TS1(3)*Iij
	         + uprev_NL3(1)*uprev_TS1(0)*Iij+uprev_NL3(1)*uprev_TS1(1)*Iii+uprev_NL3(1)*uprev_TS1(2)*Iij+uprev_NL3(1)*uprev_TS1(3)*Iij
	         + uprev_NL3(2)*uprev_TS1(0)*Iij+uprev_NL3(2)*uprev_TS1(1)*Iii+uprev_NL3(2)*uprev_TS1(2)*Iij+uprev_NL3(2)*uprev_TS1(3)*Iij
  	         + uprev_NL3(3)*uprev_TS1(0)*Iij+uprev_NL3(3)*uprev_TS1(1)*Iii+uprev_NL3(3)*uprev_TS1(2)*Iij+uprev_NL3(3)*uprev_TS1(3)*Iij;
	uNL3uTS2 = uprev_NL3(0)*uprev_TS2(0)*Iii+uprev_NL3(0)*uprev_TS2(1)*Iij+uprev_NL3(0)*uprev_TS2(2)*Iij+uprev_NL3(0)*uprev_TS2(3)*Iij
	         + uprev_NL3(1)*uprev_TS2(0)*Iij+uprev_NL3(1)*uprev_TS2(1)*Iii+uprev_NL3(1)*uprev_TS2(2)*Iij+uprev_NL3(1)*uprev_TS2(3)*Iij
	         + uprev_NL3(2)*uprev_TS2(0)*Iij+uprev_NL3(2)*uprev_TS2(1)*Iii+uprev_NL3(2)*uprev_TS2(2)*Iij+uprev_NL3(2)*uprev_TS2(3)*Iij
  	         + uprev_NL3(3)*uprev_TS2(0)*Iij+uprev_NL3(3)*uprev_TS2(1)*Iii+uprev_NL3(3)*uprev_TS2(2)*Iij+uprev_NL3(3)*uprev_TS2(3)*Iij;
	uNL3uTS3 = uprev_NL3(0)*uprev_TS3(0)*Iii+uprev_NL3(0)*uprev_TS3(1)*Iij+uprev_NL3(0)*uprev_TS3(2)*Iij+uprev_NL3(0)*uprev_TS3(3)*Iij
	         + uprev_NL3(1)*uprev_TS3(0)*Iij+uprev_NL3(1)*uprev_TS3(1)*Iii+uprev_NL3(1)*uprev_TS3(2)*Iij+uprev_NL3(1)*uprev_TS3(3)*Iij
	         + uprev_NL3(2)*uprev_TS3(0)*Iij+uprev_NL3(2)*uprev_TS3(1)*Iii+uprev_NL3(2)*uprev_TS3(2)*Iij+uprev_NL3(2)*uprev_TS3(3)*Iij
  	         + uprev_NL3(3)*uprev_TS3(0)*Iij+uprev_NL3(3)*uprev_TS3(1)*Iii+uprev_NL3(3)*uprev_TS3(2)*Iij+uprev_NL3(3)*uprev_TS3(3)*Iij;

	f1u1 = f1(0)*uprev_NL1(0)*Iii + f1(0)*uprev_NL1(1)*Iij + f1(0)*uprev_NL1(2)*Iij + f1(0)*uprev_NL1(3)*Iij
	     + f1(1)*uprev_NL1(0)*Iii + f1(1)*uprev_NL1(1)*Iij + f1(1)*uprev_NL1(2)*Iij + f1(1)*uprev_NL1(3)*Iij
	     + f1(2)*uprev_NL1(0)*Iii + f1(2)*uprev_NL1(1)*Iij + f1(2)*uprev_NL1(2)*Iij + f1(2)*uprev_NL1(3)*Iij
	     + f1(3)*uprev_NL1(0)*Iii + f1(3)*uprev_NL1(1)*Iij + f1(3)*uprev_NL1(2)*Iij + f1(3)*uprev_NL1(3)*Iij;
	f1u2 = f1(0)*uprev_NL2(0)*Iii + f1(0)*uprev_NL2(1)*Iij + f1(0)*uprev_NL2(2)*Iij + f1(0)*uprev_NL2(3)*Iij
	     + f1(1)*uprev_NL2(0)*Iii + f1(1)*uprev_NL2(1)*Iij + f1(1)*uprev_NL2(2)*Iij + f1(1)*uprev_NL2(3)*Iij
	     + f1(2)*uprev_NL2(0)*Iii + f1(2)*uprev_NL2(1)*Iij + f1(2)*uprev_NL2(2)*Iij + f1(2)*uprev_NL2(3)*Iij
	     + f1(3)*uprev_NL2(0)*Iii + f1(3)*uprev_NL2(1)*Iij + f1(3)*uprev_NL2(2)*Iij + f1(3)*uprev_NL2(3)*Iij;
	f1u3 = f1(0)*uprev_NL3(0)*Iii + f1(0)*uprev_NL3(1)*Iij + f1(0)*uprev_NL3(2)*Iij + f1(0)*uprev_NL3(3)*Iij
	     + f1(1)*uprev_NL3(0)*Iii + f1(1)*uprev_NL3(1)*Iij + f1(1)*uprev_NL3(2)*Iij + f1(1)*uprev_NL3(3)*Iij
	     + f1(2)*uprev_NL3(0)*Iii + f1(2)*uprev_NL3(1)*Iij + f1(2)*uprev_NL3(2)*Iij + f1(2)*uprev_NL3(3)*Iij
	     + f1(3)*uprev_NL3(0)*Iii + f1(3)*uprev_NL3(1)*Iij + f1(3)*uprev_NL3(2)*Iij + f1(3)*uprev_NL3(3)*Iij;
	f2u1 = f2(0)*uprev_NL1(0)*Iii + f2(0)*uprev_NL1(1)*Iij + f2(0)*uprev_NL1(2)*Iij + f2(0)*uprev_NL1(3)*Iij
	     + f2(1)*uprev_NL1(0)*Iii + f2(1)*uprev_NL1(1)*Iij + f2(1)*uprev_NL1(2)*Iij + f2(1)*uprev_NL1(3)*Iij
	     + f2(2)*uprev_NL1(0)*Iii + f2(2)*uprev_NL1(1)*Iij + f2(2)*uprev_NL1(2)*Iij + f2(2)*uprev_NL1(3)*Iij
	     + f2(3)*uprev_NL1(0)*Iii + f2(3)*uprev_NL1(1)*Iij + f2(3)*uprev_NL1(2)*Iij + f2(3)*uprev_NL1(3)*Iij;
	f2u2 = f2(0)*uprev_NL2(0)*Iii + f2(0)*uprev_NL2(1)*Iij + f2(0)*uprev_NL2(2)*Iij + f2(0)*uprev_NL2(3)*Iij
	     + f2(1)*uprev_NL2(0)*Iii + f2(1)*uprev_NL2(1)*Iij + f2(1)*uprev_NL2(2)*Iij + f2(1)*uprev_NL2(3)*Iij
	     + f2(2)*uprev_NL2(0)*Iii + f2(2)*uprev_NL2(1)*Iij + f2(2)*uprev_NL2(2)*Iij + f2(2)*uprev_NL2(3)*Iij
	     + f2(3)*uprev_NL2(0)*Iii + f2(3)*uprev_NL2(1)*Iij + f2(3)*uprev_NL2(2)*Iij + f2(3)*uprev_NL2(3)*Iij;
	f2u3 = f2(0)*uprev_NL3(0)*Iii + f2(0)*uprev_NL3(1)*Iij + f2(0)*uprev_NL3(2)*Iij + f2(0)*uprev_NL3(3)*Iij
	     + f2(1)*uprev_NL3(0)*Iii + f2(1)*uprev_NL3(1)*Iij + f2(1)*uprev_NL3(2)*Iij + f2(1)*uprev_NL3(3)*Iij
	     + f2(2)*uprev_NL3(0)*Iii + f2(2)*uprev_NL3(1)*Iij + f2(2)*uprev_NL3(2)*Iij + f2(2)*uprev_NL3(3)*Iij
	     + f2(3)*uprev_NL3(0)*Iii + f2(3)*uprev_NL3(1)*Iij + f2(3)*uprev_NL3(2)*Iij + f2(3)*uprev_NL3(3)*Iij;
	f3u1 = f3(0)*uprev_NL1(0)*Iii + f3(0)*uprev_NL1(1)*Iij + f3(0)*uprev_NL1(2)*Iij + f3(0)*uprev_NL1(3)*Iij
	     + f3(1)*uprev_NL1(0)*Iii + f3(1)*uprev_NL1(1)*Iij + f3(1)*uprev_NL1(2)*Iij + f3(1)*uprev_NL1(3)*Iij
	     + f3(2)*uprev_NL1(0)*Iii + f3(2)*uprev_NL1(1)*Iij + f3(2)*uprev_NL1(2)*Iij + f3(2)*uprev_NL1(3)*Iij
	     + f3(3)*uprev_NL1(0)*Iii + f3(3)*uprev_NL1(1)*Iij + f3(3)*uprev_NL1(2)*Iij + f3(3)*uprev_NL1(3)*Iij;
	f3u2 = f3(0)*uprev_NL2(0)*Iii + f3(0)*uprev_NL2(1)*Iij + f3(0)*uprev_NL2(2)*Iij + f3(0)*uprev_NL2(3)*Iij
	     + f3(1)*uprev_NL2(0)*Iii + f3(1)*uprev_NL2(1)*Iij + f3(1)*uprev_NL2(2)*Iij + f3(1)*uprev_NL2(3)*Iij
	     + f3(2)*uprev_NL2(0)*Iii + f3(2)*uprev_NL2(1)*Iij + f3(2)*uprev_NL2(2)*Iij + f3(2)*uprev_NL2(3)*Iij
	     + f3(3)*uprev_NL2(0)*Iii + f3(3)*uprev_NL2(1)*Iij + f3(3)*uprev_NL2(2)*Iij + f3(3)*uprev_NL2(3)*Iij;
	f3u3 = f3(0)*uprev_NL3(0)*Iii + f3(0)*uprev_NL3(1)*Iij + f3(0)*uprev_NL3(2)*Iij + f3(0)*uprev_NL3(3)*Iij
	     + f3(1)*uprev_NL3(0)*Iii + f3(1)*uprev_NL3(1)*Iij + f3(1)*uprev_NL3(2)*Iij + f3(1)*uprev_NL3(3)*Iij
	     + f3(2)*uprev_NL3(0)*Iii + f3(2)*uprev_NL3(1)*Iij + f3(2)*uprev_NL3(2)*Iij + f3(2)*uprev_NL3(3)*Iij
	     + f3(3)*uprev_NL3(0)*Iii + f3(3)*uprev_NL3(1)*Iij + f3(3)*uprev_NL3(2)*Iij + f3(3)*uprev_NL3(3)*Iij;

      }

      f_uDv_1 = (1.0-theta) * (vx*f1u1 + vy*f1u2 + vz*f1u3);
      f_uDv_2 = (1.0-theta) * (vx*f2u1 + vy*f2u2 + vz*f2u3);
      f_uDv_3 = (1.0-theta) * (vx*f3u1 + vy*f3u2 + vz*f3u3);

      uDu_Dq = ( (1.0-theta)*(umid1*u1Grad(0)+umid2*u1Grad(1)+umid3*u1Grad(2)) 
	       + (1.0-theta)*theta*(uNL1*uprev_TS1Grad(0)+uNL2*uprev_TS1Grad(1)+uNL3*uprev_TS1Grad(2)) )*Ii*vx
	     + ( (1.0-theta)*(umid1*u2Grad(0)+umid2*u2Grad(1)+umid3*u2Grad(2)) 
	       + (1.0-theta)*theta*(uNL1*uprev_TS2Grad(0)+uNL2*uprev_TS2Grad(1)+uNL3*uprev_TS2Grad(2)) )*Ii*vy
	     + ( (1.0-theta)*(umid1*u3Grad(0)+umid2*u3Grad(1)+umid3*u3Grad(2)) 
	       + (1.0-theta)*theta*(uNL1*uprev_TS3Grad(0)+uNL2*uprev_TS3Grad(1)+uNL3*uprev_TS3Grad(2)) )*Ii*vz;

      
      uDu_uDv_1 = (1.0-theta)*u1Grad(0) * (vx*u1u1 + vy*u1u2 + vz*u1u3)  
	        + (1.0-theta)*u1Grad(1) * (vx*u2u1 + vy*u2u2 + vz*u2u3)
	        + (1.0-theta)*u1Grad(2) * (vx*u3u1 + vy*u3u2 + vz*u3u3)
                + theta*uprev_TS1Grad(0) * theta*(1.0-theta)*(vx*uNL1uTS1 + vy*uNL1uTS2 + vz*uNL1uTS3)  
	        + theta*uprev_TS1Grad(1) * theta*(1.0-theta)*(vx*uNL2uTS1 + vy*uNL2uTS2 + vz*uNL2uTS3)
	        + theta*uprev_TS1Grad(2) * theta*(1.0-theta)*(vx*uNL3uTS1 + vy*uNL3uTS2 + vz*uNL3uTS3)
                + theta*uprev_TS1Grad(0) * theta*(1.0-theta)*(vx*uNL1uTS1 + vy*uNL2uTS1 + vz*uNL3uTS1)  
	        + theta*uprev_TS1Grad(1) * theta*(1.0-theta)*(vx*uNL1uTS2 + vy*uNL2uTS2 + vz*uNL3uTS2)
	        + theta*uprev_TS1Grad(2) * theta*(1.0-theta)*(vx*uNL1uTS3 + vy*uNL2uTS3 + vz*uNL3uTS3);

      uDu_uDv_2 = (1.0-theta)*u2Grad(0) * (vx*u1u1 + vy*u1u2 + vz*u1u3)  
	        + (1.0-theta)*u2Grad(1) * (vx*u2u1 + vy*u2u2 + vz*u2u3)
	        + (1.0-theta)*u2Grad(2) * (vx*u3u1 + vy*u3u2 + vz*u3u3)
                + theta*uprev_TS2Grad(0) * theta*(1.0-theta)*(vx*uNL1uTS1 + vy*uNL1uTS2 + vz*uNL1uTS3)  
	        + theta*uprev_TS2Grad(1) * theta*(1.0-theta)*(vx*uNL2uTS1 + vy*uNL2uTS2 + vz*uNL2uTS3)
	        + theta*uprev_TS2Grad(2) * theta*(1.0-theta)*(vx*uNL3uTS1 + vy*uNL3uTS2 + vz*uNL3uTS3)
                + theta*uprev_TS2Grad(0) * theta*(1.0-theta)*(vx*uNL1uTS1 + vy*uNL2uTS1 + vz*uNL3uTS1)  
	        + theta*uprev_TS2Grad(1) * theta*(1.0-theta)*(vx*uNL1uTS2 + vy*uNL2uTS2 + vz*uNL3uTS2)
	        + theta*uprev_TS2Grad(2) * theta*(1.0-theta)*(vx*uNL1uTS3 + vy*uNL2uTS3 + vz*uNL3uTS3);

      uDu_uDv_3 = (1.0-theta)*u3Grad(0) * (vx*u1u1 + vy*u1u2 + vz*u1u3)  
	        + (1.0-theta)*u3Grad(1) * (vx*u2u1 + vy*u2u2 + vz*u2u3)
	        + (1.0-theta)*u3Grad(2) * (vx*u3u1 + vy*u3u2 + vz*u3u3)
                + theta*uprev_TS3Grad(0) * theta*(1.0-theta)*(vx*uNL1uTS1 + vy*uNL1uTS2 + vz*uNL1uTS3)  
	        + theta*uprev_TS3Grad(1) * theta*(1.0-theta)*(vx*uNL2uTS1 + vy*uNL2uTS2 + vz*uNL2uTS3)
	        + theta*uprev_TS3Grad(2) * theta*(1.0-theta)*(vx*uNL3uTS1 + vy*uNL3uTS2 + vz*uNL3uTS3)
                + theta*uprev_TS3Grad(0) * theta*(1.0-theta)*(vx*uNL1uTS1 + vy*uNL2uTS1 + vz*uNL3uTS1)  
	        + theta*uprev_TS3Grad(1) * theta*(1.0-theta)*(vx*uNL1uTS2 + vy*uNL2uTS2 + vz*uNL3uTS2)
	        + theta*uprev_TS3Grad(2) * theta*(1.0-theta)*(vx*uNL1uTS3 + vy*uNL2uTS3 + vz*uNL3uTS3);

      NL1   = (umid1_1*uv1+umid1_2*uv2+umid1_3*uv3+umid1_4*uv4) * (1.0-theta)*u1Grad(0)
            + (umid2_1*uv1+umid2_2*uv2+umid2_3*uv3+umid2_4*uv4) * (1.0-theta)*u1Grad(1)
	    + (umid3_1*uv1+umid3_2*uv2+umid3_3*uv3+umid3_4*uv4) * (1.0-theta)*u1Grad(2)
            + (1.0-theta)*(uprev_NL1(0)*uv1+uprev_NL1(1)*uv2+uprev_NL1(2)*uv3+uprev_NL1(3)*uv4) * theta*uprev_TS1Grad(0)
            + (1.0-theta)*(uprev_NL2(0)*uv1+uprev_NL2(1)*uv2+uprev_NL2(2)*uv3+uprev_NL2(3)*uv4) * theta*uprev_TS1Grad(1)
            + (1.0-theta)*(uprev_NL3(0)*uv1+uprev_NL3(1)*uv2+uprev_NL3(2)*uv3+uprev_NL3(3)*uv4) * theta*uprev_TS1Grad(2);

      NL2   = (umid1_1*uv1+umid1_2*uv2+umid1_3*uv3+umid1_4*uv4) * (1.0-theta)*u2Grad(0)
	    + (umid2_1*uv1+umid2_2*uv2+umid2_3*uv3+umid2_4*uv4) * (1.0-theta)*u2Grad(1)
            + (umid3_1*uv1+umid3_2*uv2+umid3_3*uv3+umid3_4*uv4) * (1.0-theta)*u2Grad(2)
            + (1.0-theta)*(uprev_NL1(0)*uv1+uprev_NL1(1)*uv2+uprev_NL1(2)*uv3+uprev_NL1(3)*uv4) * theta*uprev_TS2Grad(0)
            + (1.0-theta)*(uprev_NL2(0)*uv1+uprev_NL2(1)*uv2+uprev_NL2(2)*uv3+uprev_NL2(3)*uv4) * theta*uprev_TS2Grad(1)
            + (1.0-theta)*(uprev_NL3(0)*uv1+uprev_NL3(1)*uv2+uprev_NL3(2)*uv3+uprev_NL3(3)*uv4) * theta*uprev_TS2Grad(2);

      NL3   = (umid1_1*uv1+umid1_2*uv2+umid1_3*uv3+umid1_4*uv4) * (1.0-theta)*u3Grad(0)
	    + (umid2_1*uv1+umid2_2*uv2+umid2_3*uv3+umid2_4*uv4) * (1.0-theta)*u3Grad(1)
	    + (umid3_1*uv1+umid3_2*uv2+umid3_3*uv3+umid3_4*uv4) * (1.0-theta)*u3Grad(2)
            + (1.0-theta)*(uprev_NL1(0)*uv1+uprev_NL1(1)*uv2+uprev_NL1(2)*uv3+uprev_NL1(3)*uv4) * theta*uprev_TS3Grad(0)
            + (1.0-theta)*(uprev_NL2(0)*uv1+uprev_NL2(1)*uv2+uprev_NL2(2)*uv3+uprev_NL2(3)*uv4) * theta*uprev_TS3Grad(1)
            + (1.0-theta)*(uprev_NL3(0)*uv1+uprev_NL3(1)*uv2+uprev_NL3(2)*uv3+uprev_NL3(3)*uv4) * theta*uprev_TS3Grad(2);


      if ( smagorinsky_eddy_viscosity ){
	if ( bf == 0 ){
	  S_abs = sqrt( sqr(uprev_TS1Grad(0)) + sqr(uprev_TS1Grad(1)) + sqr(uprev_TS1Grad(2))
		      + sqr(uprev_TS2Grad(0)) + sqr(uprev_TS2Grad(1)) + sqr(uprev_TS2Grad(2))
	              + sqr(uprev_TS3Grad(0)) + sqr(uprev_TS3Grad(1)) + sqr(uprev_TS3Grad(2)) );
	  nu_T  = pow(Cs*h,2.0) * S_abs; 
	  nu = (1.0/Re) + nu_T;
	}
      } else{
	nu = (1.0/Re);
      }
      LAP1 = nu * (1.0-theta)*( u1Grad(0)*vx + u1Grad(1)*vy + u1Grad(2)*vz );
      LAP2 = nu * (1.0-theta)*( u2Grad(0)*vx + u2Grad(1)*vy + u2Grad(2)*vz );
      LAP3 = nu * (1.0-theta)*( u3Grad(0)*vx + u3Grad(1)*vy + u3Grad(2)*vz );

      u_v_1 = u1(0)*uv1+u1(1)*uv2+u1(2)*uv3+u1(3)*uv4;
      u_v_2 = u2(0)*uv1+u2(1)*uv2+u2(2)*uv3+u2(3)*uv4;
      u_v_3 = u3(0)*uv1+u3(1)*uv2+u3(2)*uv3+u3(3)*uv4;
      
      p_div_v_1 = psum*vx*Ii;
      p_div_v_2 = psum*vy*Ii;
      p_div_v_3 = psum*vz*Ii;

      u_Dv_1 = ( umid1*vx + umid2*vy + umid3*vz ) * Ii; 
      u_Dv_2 = u_Dv_1;
      u_Dv_3 = u_Dv_1;
      
      SD1 = uDu_uDv_1 + pGrad(0)*u_Dv_1 - f_uDv_1;
      SD2 = uDu_uDv_2 + pGrad(1)*u_Dv_2 - f_uDv_2;
      SD3 = uDu_uDv_3 + pGrad(2)*u_Dv_3 - f_uDv_3;
      
      LAPp = pGrad(0)*vx + pGrad(1)*vy + pGrad(2)*vz;

      if ( bf == 0 ){
	divU = (1.0-theta) * (u1Grad(0)+u2Grad(1)+u3Grad(2));
      }      

      LSdiv1 = divU*vx;
      LSdiv2 = divU*vy;
      LSdiv3 = divU*vz;
            
      bres(0,bf) = u_v_1/dt + LAP1 + NL1 - p_div_v_1 + delta1*SD1 + delta2*LSdiv1 + r*u_v_1;
      bres(1,bf) = u_v_2/dt + LAP2 + NL2 - p_div_v_2 + delta1*SD2 + delta2*LSdiv2 + r*u_v_2;
      bres(2,bf) = u_v_3/dt + LAP3 + NL3 - p_div_v_3 + delta1*SD3 + delta2*LSdiv3 + r*u_v_3;
      bres(3,bf) = delta1*LAPp + delta1*uDu_Dq + (divU*Ii); 
  }


  inline
  void NonLinearResidualExactInt( MV_ColMat<real>& bres, int& bf, int& el, FET4n3D& FE ){

      if ( bf == 0 ){
	h = 2.0*FE.circumRadius();
	
	x(0) = FE.x1();
	x(1) = FE.x2();
	x(2) = FE.x3();
	x(3) = FE.x4();
	
	y(0) = FE.y1();
	y(1) = FE.y2();
	y(2) = FE.y3();
	y(3) = FE.y4();
	
	z(0) = FE.z1();
	z(1) = FE.z2();
	z(2) = FE.z3();
	z(3) = FE.z4();
	
	cout << "check: " << COEFF(0)->GetRep() << endl;
	cout << "check: " << COEFF(0)->GetRep() << endl;
	cout << "check:1 " << COEFF(1)->GetRep() << endl;
	uprev_TS1(0)     = COEFF(0) -> GetNodalValue( FE.n1() );    
	cout << "check:1 " << COEFF(1)->GetRep() << endl;
	uprev_TS1(1)     = COEFF(0) -> GetNodalValue( FE.n2() );
	uprev_TS1(2)     = COEFF(0) -> GetNodalValue( FE.n3() );
	uprev_TS1(3)     = COEFF(0) -> GetNodalValue( FE.n4() );
	FE.evalGradientLoc(uprev_TS1,uprev_TS1Grad);
	cout << "check:2 " << COEFF(1)->GetRep() << endl;
	uprev_TS2(0)     = COEFF(1) -> GetNodalValue( FE.n1() );
	uprev_TS2(1)     = COEFF(1) -> GetNodalValue( FE.n2() );
	uprev_TS2(2)     = COEFF(1) -> GetNodalValue( FE.n3() );
	uprev_TS2(3)     = COEFF(1) -> GetNodalValue( FE.n4() );
	FE.evalGradientLoc(uprev_TS2,uprev_TS2Grad);
	cout << "check:3 " << COEFF(1)->GetRep() << endl;
	uprev_TS3(0)     = COEFF(2) -> GetNodalValue( FE.n1() );
	uprev_TS3(1)     = COEFF(2) -> GetNodalValue( FE.n2() );
	uprev_TS3(2)     = COEFF(2) -> GetNodalValue( FE.n3() );
	uprev_TS3(3)     = COEFF(2) -> GetNodalValue( FE.n4() );
	FE.evalGradientLoc(uprev_TS3,uprev_TS3Grad);
	//pprev_TS(0)      = COEFF(3) -> GetNodalValue( FE.n1() );
	//pprev_TS(1)      = COEFF(3) -> GetNodalValue( FE.n2() );
	//pprev_TS(2)      = COEFF(3) -> GetNodalValue( FE.n3() );
	//pprev_TS(3)      = COEFF(3) -> GetNodalValue( FE.n4() );
	//FE.evalGradientLoc(pprev_TS,pprev_TSGrad);
	
	cout << "check" << endl;
	uprev_NL1(0)     = COEFF(4) -> GetNodalValue( FE.n1() );    
	uprev_NL1(1)     = COEFF(4) -> GetNodalValue( FE.n2() );
	uprev_NL1(2)     = COEFF(4) -> GetNodalValue( FE.n3() );
	uprev_NL1(3)     = COEFF(4) -> GetNodalValue( FE.n4() );
	//FE.evalGradientLoc(uprev_NL1,uprev_NL1Grad);
	uprev_NL2(0)     = COEFF(5) -> GetNodalValue( FE.n1() );
	uprev_NL2(1)     = COEFF(5) -> GetNodalValue( FE.n2() );
	uprev_NL2(2)     = COEFF(5) -> GetNodalValue( FE.n3() );
	uprev_NL2(3)     = COEFF(5) -> GetNodalValue( FE.n4() );
	//FE.evalGradientLoc(uprev_NL2,uprev_NL2Grad);
	uprev_NL3(0)     = COEFF(6) -> GetNodalValue( FE.n1() );
	uprev_NL3(1)     = COEFF(6) -> GetNodalValue( FE.n2() );
	uprev_NL3(2)     = COEFF(6) -> GetNodalValue( FE.n3() );
	uprev_NL3(3)     = COEFF(6) -> GetNodalValue( FE.n4() );
	//FE.evalGradientLoc(uprev_NL3,uprev_NL3Grad);
	//pprev_NL(0)      = COEFF(7) -> GetNodalValue( FE.n1() );
	//pprev_NL(1)      = COEFF(7) -> GetNodalValue( FE.n2() );
	//pprev_NL(2)      = COEFF(7) -> GetNodalValue( FE.n3() );
	//pprev_NL(3)      = COEFF(7) -> GetNodalValue( FE.n4() );
	//FE.evalGradientLoc(pprev_NL,pprev_NLGrad);
	
	cout << "check" << endl;
	u1(0)     = COEFF(8)  -> GetNodalValue( FE.n1() );
	u1(1)     = COEFF(8)  -> GetNodalValue( FE.n2() );
	u1(2)     = COEFF(8)  -> GetNodalValue( FE.n3() );
	u1(3)     = COEFF(8)  -> GetNodalValue( FE.n4() );
	FE.evalGradientLoc(u1,u1Grad);
	u2(0)     = COEFF(9) -> GetNodalValue( FE.n1() );
	u2(1)     = COEFF(9) -> GetNodalValue( FE.n2() );
	u2(2)     = COEFF(9) -> GetNodalValue( FE.n3() );
	u2(3)     = COEFF(9) -> GetNodalValue( FE.n4() );
	FE.evalGradientLoc(u2,u2Grad);
	u3(0)     = COEFF(10) -> GetNodalValue( FE.n1() );
	u3(1)     = COEFF(10) -> GetNodalValue( FE.n2() );
	u3(2)     = COEFF(10) -> GetNodalValue( FE.n3() );
	u3(3)     = COEFF(10) -> GetNodalValue( FE.n4() );
	FE.evalGradientLoc(u3,u3Grad);
	p(0)      = COEFF(11) -> GetNodalValue( FE.n1() );
	p(1)      = COEFF(11) -> GetNodalValue( FE.n2() );
	p(2)      = COEFF(11) -> GetNodalValue( FE.n3() );
	p(3)      = COEFF(11) -> GetNodalValue( FE.n4() );
	FE.evalGradientLoc(p,pGrad);
	
	cout << "check" << endl;
	dt           = COEFF(12) -> Eval();
	//t            = COEFF(13) -> Eval();
	Re           = COEFF(14) -> Eval();
	
	cout << "check" << endl;
	f1(0) = (1.0/Re) * 32.0 * ( y(0)*(1.0-y(0)) + z(0)*(1.0-z(0)) );
	f1(1) = (1.0/Re) * 32.0 * ( y(1)*(1.0-y(1)) + z(1)*(1.0-z(1)) );
	f1(2) = (1.0/Re) * 32.0 * ( y(2)*(1.0-y(2)) + z(2)*(1.0-z(2)) );
	f1(3) = (1.0/Re) * 32.0 * ( y(3)*(1.0-y(3)) + z(3)*(1.0-z(3)) );
	f1 = 0.0;

	f2(0) = 0.001 * sin(10.0*pi*x(0));
	f2(1) = 0.001 * sin(10.0*pi*x(1));
	f2(2) = 0.001 * sin(10.0*pi*x(2));
	f2(3) = 0.001 * sin(10.0*pi*x(3));
	f2 = 0.0;

	f3(0) = 0.001 * sin(10.0*pi*x(0));
	f3(1) = 0.001 * sin(10.0*pi*x(1));
	f3(2) = 0.001 * sin(10.0*pi*x(2));
	f3(3) = 0.001 * sin(10.0*pi*x(3));
	f3 = 0.0;
	
	u1sum = u1(0) + u1(1) + u1(2) + u1(3);
	u2sum = u2(0) + u2(1) + u2(2) + u2(3);
	u3sum = u3(0) + u3(1) + u3(2) + u3(3);
	normU = sqrt( sqr(u1sum) + sqr(u2sum) + sqr(u3sum) ) / 4;
	
	divU = u1Grad(0) + u2Grad(1) + u3Grad(2);
	
	if ( h * Re > 1.0 ){ 
	  //delta1 = 0.5 / sqrt( 1.0/sqr(dt) + sqr(normU/h) );
	  delta1 = C1 * h;
	}
	else{ 
	  delta1 = C1 * sqr(h);
	}
	if ( h * Re > 1.0 ){ 
	  delta2 = C2 * h;
	}
	else{ 
	  delta2 = C2 * sqr(h);
	}
	if ( divU > K ){ 
	  r = 0.5 * divU;
	}
	else{ 
	  r = 0.0;
	}
	r = 0.0;
      }
      
      if ( bf == 0 ){ 
	uv1 = Iii;
	uv2 = uv3 = uv4 = Iij;
	vx = FE.dN1x();
	vy = FE.dN1y();
	vz = FE.dN1z();
      }else if ( bf == 1 ){
	uv2 = Iii;
	uv1 = uv3 = uv4 = Iij;
	vx = FE.dN2x();
	vy = FE.dN2y();
	vz = FE.dN2z();
      }else if ( bf == 2 ){
	uv3 = Iii;
	uv1 = uv2 = uv4 = Iij;
	vx = FE.dN3x();
	vy = FE.dN3y();
	vz = FE.dN3z();
      }else if ( bf == 3 ){
	uv4 = Iii;
	uv1 = uv2 = uv3 = Iij;
	vx = FE.dN4x();
	vy = FE.dN4y();
	vz = FE.dN4z();
      }	

      if ( bf == 0 ){
	umid1_1 = theta*uprev_TS1(0) + (1.0-theta)*uprev_NL1(0);
	umid1_2 = theta*uprev_TS1(1) + (1.0-theta)*uprev_NL1(1);
	umid1_3 = theta*uprev_TS1(2) + (1.0-theta)*uprev_NL1(2);
	umid1_4 = theta*uprev_TS1(3) + (1.0-theta)*uprev_NL1(3);

	umid2_1 = theta*uprev_TS2(0) + (1.0-theta)*uprev_NL2(0);
	umid2_2 = theta*uprev_TS2(1) + (1.0-theta)*uprev_NL2(1);
	umid2_3 = theta*uprev_TS2(2) + (1.0-theta)*uprev_NL2(2);
	umid2_4 = theta*uprev_TS2(3) + (1.0-theta)*uprev_NL2(3);

	umid3_1 = theta*uprev_TS3(0) + (1.0-theta)*uprev_NL3(0);
	umid3_2 = theta*uprev_TS3(1) + (1.0-theta)*uprev_NL3(1);
	umid3_3 = theta*uprev_TS3(2) + (1.0-theta)*uprev_NL3(2);
	umid3_4 = theta*uprev_TS3(3) + (1.0-theta)*uprev_NL3(3);

	umid1 = umid1_1 + umid1_2 + umid1_3 + umid1_4;
	umid2 = umid2_1 + umid2_2 + umid2_3 + umid2_4;
	umid3 = umid3_1 + umid3_2 + umid3_3 + umid3_4;
	
	D1umid1 = theta*uprev_TS1Grad(0) + (1.0-theta)*u1Grad(0);
	D2umid1 = theta*uprev_TS1Grad(1) + (1.0-theta)*u1Grad(1);
	D3umid1 = theta*uprev_TS1Grad(2) + (1.0-theta)*u1Grad(2);
	
	D1umid2 = theta*uprev_TS2Grad(0) + (1.0-theta)*u2Grad(0);
	D2umid2 = theta*uprev_TS2Grad(1) + (1.0-theta)*u2Grad(1);
	D3umid2 = theta*uprev_TS2Grad(2) + (1.0-theta)*u2Grad(2);
	
	D1umid3 = theta*uprev_TS3Grad(0) + (1.0-theta)*u3Grad(0);
	D2umid3 = theta*uprev_TS3Grad(1) + (1.0-theta)*u3Grad(1);
	D3umid3 = theta*uprev_TS3Grad(2) + (1.0-theta)*u3Grad(2);

	psum = p(0) + p(1) + p(2) + p(3);

	u1u1 = umid1_1*umid1_1*Iii + umid1_1*umid1_2*Iij + umid1_1*umid1_3*Iij + umid1_1*umid1_4*Iij
	     + umid1_2*umid1_1*Iij + umid1_2*umid1_2*Iii + umid1_2*umid1_3*Iij + umid1_2*umid1_4*Iij
	     + umid1_3*umid1_1*Iij + umid1_3*umid1_2*Iij + umid1_3*umid1_3*Iii + umid1_3*umid1_4*Iij
	     + umid1_4*umid1_1*Iij + umid1_4*umid1_2*Iij + umid1_4*umid1_3*Iij + umid1_4*umid1_4*Iii;
	u1u2 = umid1_1*umid2_1*Iii + umid1_1*umid2_2*Iij + umid1_1*umid2_3*Iij + umid1_1*umid2_4*Iij
	     + umid1_2*umid2_1*Iij + umid1_2*umid2_2*Iii + umid1_2*umid2_3*Iij + umid1_2*umid2_4*Iij
	     + umid1_3*umid2_1*Iij + umid1_3*umid2_2*Iij + umid1_3*umid2_3*Iii + umid1_3*umid2_4*Iij
	     + umid1_4*umid2_1*Iij + umid1_4*umid2_2*Iij + umid1_4*umid2_3*Iij + umid1_4*umid2_4*Iii;
	u1u3 = umid1_1*umid3_1*Iii + umid1_1*umid3_2*Iij + umid1_1*umid3_3*Iij + umid1_1*umid3_4*Iij
	     + umid1_2*umid3_1*Iij + umid1_2*umid3_2*Iii + umid1_2*umid3_3*Iij + umid1_2*umid3_4*Iij
	     + umid1_3*umid3_1*Iij + umid1_3*umid3_2*Iij + umid1_3*umid3_3*Iii + umid1_3*umid3_4*Iij
	     + umid1_4*umid3_1*Iij + umid1_4*umid3_2*Iij + umid1_4*umid3_3*Iij + umid1_4*umid3_4*Iii;
	u2u1 = u1u2;
	u2u2 = umid2_1*umid2_1*Iii + umid2_1*umid2_2*Iij + umid2_1*umid2_3*Iij + umid2_1*umid2_4*Iij
	     + umid2_2*umid2_1*Iij + umid2_2*umid2_2*Iii + umid2_2*umid2_3*Iij + umid2_2*umid2_4*Iij
	     + umid2_3*umid2_1*Iij + umid2_3*umid2_2*Iij + umid2_3*umid2_3*Iii + umid2_3*umid2_4*Iij
	     + umid2_4*umid2_1*Iij + umid2_4*umid2_2*Iij + umid2_4*umid2_3*Iij + umid2_4*umid2_4*Iii;
	u2u3 = umid2_1*umid3_1*Iii + umid2_1*umid3_2*Iij + umid2_1*umid3_3*Iij + umid2_1*umid3_4*Iij
	     + umid2_2*umid3_1*Iij + umid2_2*umid3_2*Iii + umid2_2*umid3_3*Iij + umid2_2*umid3_4*Iij
	     + umid2_3*umid3_1*Iij + umid2_3*umid3_2*Iij + umid2_3*umid3_3*Iii + umid2_3*umid3_4*Iij
	     + umid2_4*umid3_1*Iij + umid2_4*umid3_2*Iij + umid2_4*umid3_3*Iij + umid2_4*umid3_4*Iii;
	u3u1 = u1u3;
	u3u2 = u2u3;
	u3u3 = umid3_1*umid3_1*Iii + umid3_1*umid3_2*Iij + umid3_1*umid3_3*Iij + umid3_1*umid3_4*Iij
	     + umid3_2*umid3_1*Iij + umid3_2*umid3_2*Iii + umid3_2*umid3_3*Iij + umid3_2*umid3_4*Iij
	     + umid3_3*umid3_1*Iij + umid3_3*umid3_2*Iij + umid3_3*umid3_3*Iii + umid3_3*umid3_4*Iij
	     + umid3_4*umid3_1*Iij + umid3_4*umid3_2*Iij + umid3_4*umid3_3*Iij + umid3_4*umid3_4*Iii;

	f1u1 = f1(0)*umid1_1*Iii + f1(0)*umid1_2*Iij + f1(0)*umid1_3*Iij + f1(0)*umid1_4*Iij
	     + f1(1)*umid1_1*Iij + f1(1)*umid1_2*Iii + f1(1)*umid1_3*Iij + f1(1)*umid1_4*Iij
	     + f1(2)*umid1_1*Iij + f1(2)*umid1_2*Iij + f1(2)*umid1_3*Iii + f1(2)*umid1_4*Iij
	     + f1(3)*umid1_1*Iij + f1(3)*umid1_2*Iij + f1(3)*umid1_3*Iij + f1(3)*umid1_4*Iii;
	f1u2 = f1(0)*umid2_1*Iii + f1(0)*umid2_2*Iij + f1(0)*umid2_3*Iij + f1(0)*umid2_4*Iij
	     + f1(1)*umid2_1*Iij + f1(1)*umid2_2*Iii + f1(1)*umid2_3*Iij + f1(1)*umid2_4*Iij
	     + f1(2)*umid2_1*Iij + f1(2)*umid2_2*Iij + f1(2)*umid2_3*Iii + f1(2)*umid2_4*Iij
	     + f1(3)*umid2_1*Iij + f1(3)*umid2_2*Iij + f1(3)*umid2_3*Iij + f1(3)*umid2_4*Iii;
	f1u3 = f1(0)*umid3_1*Iii + f1(0)*umid3_2*Iij + f1(0)*umid3_3*Iij + f1(0)*umid3_4*Iij
	     + f1(1)*umid3_1*Iij + f1(1)*umid3_2*Iii + f1(1)*umid3_3*Iij + f1(1)*umid3_4*Iij
	     + f1(2)*umid3_1*Iij + f1(2)*umid3_2*Iij + f1(2)*umid3_3*Iii + f1(2)*umid3_4*Iij
	     + f1(3)*umid3_1*Iij + f1(3)*umid3_2*Iij + f1(3)*umid3_3*Iij + f1(3)*umid3_4*Iii;
	f2u1 = f2(0)*umid1_1*Iii + f2(0)*umid1_2*Iij + f2(0)*umid1_3*Iij + f2(0)*umid1_4*Iij
	     + f2(1)*umid1_1*Iij + f2(1)*umid1_2*Iii + f2(1)*umid1_3*Iij + f2(1)*umid1_4*Iij
	     + f2(2)*umid1_1*Iij + f2(2)*umid1_2*Iij + f2(2)*umid1_3*Iii + f2(2)*umid1_4*Iij
	     + f2(3)*umid1_1*Iij + f2(3)*umid1_2*Iij + f2(3)*umid1_3*Iij + f2(3)*umid1_4*Iii;
	f2u2 = f2(0)*umid2_1*Iii + f2(0)*umid2_2*Iij + f2(0)*umid2_3*Iij + f2(0)*umid2_4*Iij
	     + f2(1)*umid2_1*Iij + f2(1)*umid2_2*Iii + f2(1)*umid2_3*Iij + f2(1)*umid2_4*Iij
	     + f2(2)*umid2_1*Iij + f2(2)*umid2_2*Iij + f2(2)*umid2_3*Iii + f2(2)*umid2_4*Iij
	     + f2(3)*umid2_1*Iij + f2(3)*umid2_2*Iij + f2(3)*umid2_3*Iij + f2(3)*umid2_4*Iii;
	f2u3 = f2(0)*umid3_1*Iii + f2(0)*umid3_2*Iij + f2(0)*umid3_3*Iij + f2(0)*umid3_4*Iij
	     + f2(1)*umid3_1*Iij + f2(1)*umid3_2*Iii + f2(1)*umid3_3*Iij + f2(1)*umid3_4*Iij
	     + f2(2)*umid3_1*Iij + f2(2)*umid3_2*Iij + f2(2)*umid3_3*Iii + f2(2)*umid3_4*Iij
	     + f2(3)*umid3_1*Iij + f2(3)*umid3_2*Iij + f2(3)*umid3_3*Iij + f2(3)*umid3_4*Iii;
	f3u1 = f3(0)*umid1_1*Iii + f3(0)*umid1_2*Iij + f3(0)*umid1_3*Iij + f3(0)*umid1_4*Iij
	     + f3(1)*umid1_1*Iij + f3(1)*umid1_2*Iii + f3(1)*umid1_3*Iij + f3(1)*umid1_4*Iij
	     + f3(2)*umid1_1*Iij + f3(2)*umid1_2*Iij + f3(2)*umid1_3*Iii + f3(2)*umid1_4*Iij
	     + f3(3)*umid1_1*Iij + f3(3)*umid1_2*Iij + f3(3)*umid1_3*Iij + f3(3)*umid1_4*Iii;
	f3u2 = f3(0)*umid2_1*Iii + f3(0)*umid2_2*Iij + f3(0)*umid2_3*Iij + f3(0)*umid2_4*Iij
	     + f3(1)*umid2_1*Iij + f3(1)*umid2_2*Iii + f3(1)*umid2_3*Iij + f3(1)*umid2_4*Iij
	     + f3(2)*umid2_1*Iij + f3(2)*umid2_2*Iij + f3(2)*umid2_3*Iii + f3(2)*umid2_4*Iij
	     + f3(3)*umid2_1*Iij + f3(3)*umid2_2*Iij + f3(3)*umid2_3*Iij + f3(3)*umid2_4*Iii;
	f3u3 = f3(0)*umid3_1*Iii + f3(0)*umid3_2*Iij + f3(0)*umid3_3*Iij + f3(0)*umid3_4*Iij
	     + f3(1)*umid3_1*Iij + f3(1)*umid3_2*Iii + f3(1)*umid3_3*Iij + f3(1)*umid3_4*Iij
	     + f3(2)*umid3_1*Iij + f3(2)*umid3_2*Iij + f3(2)*umid3_3*Iii + f3(2)*umid3_4*Iij
	     + f3(3)*umid3_1*Iij + f3(3)*umid3_2*Iij + f3(3)*umid3_3*Iij + f3(3)*umid3_4*Iii;
      }

      f_uDv_1 = vx*f1u1 + vy*f1u2 + vz*f1u3;
      f_uDv_2 = vx*f2u1 + vy*f2u2 + vz*f2u3;
      f_uDv_3 = vx*f3u1 + vy*f3u2 + vz*f3u3;


      f_v_1 = f1(0)*uv1+f1(1)*uv2+f1(2)*uv3+f1(3)*uv4;
      f_v_2 = f2(0)*uv1+f2(1)*uv2+f2(2)*uv3+f2(3)*uv4;
      f_v_3 = f3(0)*uv1+f3(1)*uv2+f3(2)*uv3+f3(3)*uv4;

      f_Dq = ( f1(0)+f1(1)+f1(2)+f1(3) )*Ii*vx;
           + ( f2(0)+f2(1)+f2(2)+f2(3) )*Ii*vy;
	   + ( f3(0)+f3(1)+f3(2)+f3(3) )*Ii*vz;


      uDu_Dq = ( umid1*D1umid1+umid2*D2umid1+umid3*D3umid1 )*Ii*vx
	     + ( umid1*D1umid2+umid2*D2umid2+umid3*D3umid2 )*Ii*vy;
             + ( umid1*D1umid3+umid2*D2umid3+umid3*D3umid3 )*Ii*vz;

      
      uDu_uDv_1 = D1umid1 * (vx*u1u1 + vy*u1u2 + vz*u1u3)  
	        + D2umid1 * (vx*u2u1 + vy*u2u2 + vz*u2u3)
	        + D3umid1 * (vx*u3u1 + vy*u3u2 + vz*u3u3);

      uDu_uDv_2 = D1umid2 * (vx*u1u1 + vy*u1u2 + vz*u1u3)  
	        + D2umid2 * (vx*u2u1 + vy*u2u2 + vz*u2u3)
	        + D3umid2 * (vx*u3u1 + vy*u3u2 + vz*u3u3);

      uDu_uDv_3 = D1umid3 * (vx*u1u1 + vy*u1u2 + vz*u1u3)  
	        + D2umid3 * (vx*u2u1 + vy*u2u2 + vz*u2u3)
	        + D3umid3 * (vx*u3u1 + vy*u3u2 + vz*u3u3);


      NL1   = (umid1_1*uv1+umid1_2*uv2+umid1_3*uv3+umid1_4*uv4) * D1umid1
            + (umid2_1*uv1+umid2_2*uv2+umid2_3*uv3+umid2_4*uv4) * D2umid1
	    + (umid3_1*uv1+umid3_2*uv2+umid3_3*uv3+umid3_4*uv4) * D3umid1;

      NL2   = (umid1_1*uv1+umid1_2*uv2+umid1_3*uv3+umid1_4*uv4) * D1umid2
	    + (umid2_1*uv1+umid2_2*uv2+umid2_3*uv3+umid2_4*uv4) * D2umid2
            + (umid3_1*uv1+umid3_2*uv2+umid3_3*uv3+umid3_4*uv4) * D3umid2;

      NL3   = (umid1_1*uv1+umid1_2*uv2+umid1_3*uv3+umid1_4*uv4) * D1umid3
	    + (umid2_1*uv1+umid2_2*uv2+umid2_3*uv3+umid2_4*uv4) * D2umid3
	    + (umid3_1*uv1+umid3_2*uv2+umid3_3*uv3+umid3_4*uv4) * D3umid3;


      if ( smagorinsky_eddy_viscosity ){
	if ( bf == 0 ){
	  S_abs = sqrt( sqr(uprev_TS1Grad(0)) + sqr(uprev_TS1Grad(1)) + sqr(uprev_TS1Grad(2))
		      + sqr(uprev_TS2Grad(0)) + sqr(uprev_TS2Grad(1)) + sqr(uprev_TS2Grad(2))
	              + sqr(uprev_TS3Grad(0)) + sqr(uprev_TS3Grad(1)) + sqr(uprev_TS3Grad(2)) );
	  nu_T  = pow(Cs*h,2.0) * S_abs; 
	  nu = (1.0/Re) + nu_T;
	}
      } else{
	nu = (1.0/Re);
      }
      LAP1 = nu * ( D1umid1*vx + D2umid1*vy + D3umid1 * vz );
      LAP2 = nu * ( D1umid2*vx + D2umid2*vy + D3umid2 * vz );
      LAP3 = nu * ( D1umid3*vx + D2umid3*vy + D3umid3 * vz );

      uprev_TS_v_1 = uprev_TS1(0)*uv1+uprev_TS1(1)*uv2+uprev_TS1(2)*uv3+uprev_TS1(3)*uv4;
      uprev_TS_v_2 = uprev_TS2(0)*uv1+uprev_TS2(1)*uv2+uprev_TS2(2)*uv3+uprev_TS2(3)*uv4;
      uprev_TS_v_3 = uprev_TS3(0)*uv1+uprev_TS3(1)*uv2+uprev_TS3(2)*uv3+uprev_TS3(3)*uv4;
      
      u_v_1 = u1(0)*uv1+u1(1)*uv2+u1(2)*uv3+u1(3)*uv4;
      u_v_2 = u2(0)*uv1+u2(1)*uv2+u2(2)*uv3+u2(3)*uv4;
      u_v_3 = u3(0)*uv1+u3(1)*uv2+u3(2)*uv3+u3(3)*uv4;
      
      p_div_v_1 = psum*vx*Ii;
      p_div_v_2 = psum*vy*Ii;
      p_div_v_3 = psum*vz*Ii;

      u_Dv_1 = ( umid1*vx + umid2*vy + umid3*vz ) * Ii; 
      u_Dv_2 = u_Dv_1;
      u_Dv_3 = u_Dv_1;
      
      SD1 = uDu_uDv_1 + pGrad(0)*u_Dv_1 - f_uDv_1;
      SD2 = uDu_uDv_2 + pGrad(1)*u_Dv_2 - f_uDv_2;
      SD3 = uDu_uDv_3 + pGrad(2)*u_Dv_3 - f_uDv_3;
      
      LSdiv1 = D1umid1 + D2umid2 + D3umid3 * vx;
      LSdiv1 = D1umid1 + D2umid2 + D3umid3 * vy;
      LSdiv1 = D1umid1 + D2umid2 + D3umid3 * vz;
      
      LAPp = pGrad(0)*vx + pGrad(1)*vy + pGrad(2)*vz;

      if ( bf == 0 ){
	divUmid = theta*uprev_TS1Grad(0) + (1.0-theta)*u1Grad(0)
	  + theta*uprev_TS2Grad(1) + (1.0-theta)*u2Grad(1)
	  + theta*uprev_TS3Grad(2) + (1.0-theta)*u3Grad(2);
      }      
      
      bres(0,bf) = (u_v_1 - uprev_TS_v_1)/dt + LAP1 + NL1 - p_div_v_1 - f_v_1 + delta1*SD1 + delta2*LSdiv1 + r*u_v_1;
      bres(1,bf) = (u_v_2 - uprev_TS_v_2)/dt + LAP2 + NL2 - p_div_v_2 - f_v_2 + delta1*SD2 + delta2*LSdiv2 + r*u_v_2;
      bres(2,bf) = (u_v_3 - uprev_TS_v_3)/dt + LAP3 + NL3 - p_div_v_3 - f_v_3 + delta1*SD3 + delta2*LSdiv3 + r*u_v_3;
      bres(3,bf) = delta1*LAPp + delta1*(uDu_Dq-f_Dq) + (divUmid*Ii); 
  }


};


#endif

