#ifndef __EQUATION_PRE_CG1_CG1_HH
#define __EQUATION_PRE_CG1_CG1_HH

#include <ShapeFunction.hh>
#include <LocalField.hh>
#include <GlobalField.hh>
#include <Settings.hh>
#include <EquationSystem.hh>
#include <TetLinSpace.hh>
#include <math.h>

extern Settings *settings;

class EquationPre_cG1cG1:public Equation {
private:

  int cell_no;

  real delta1, delta2, r;
  real C1, C2, K;

  real Re, dt, t;

  real Cs, S_abs, nu, nu_T;

  real normU;

  real x[4];
  real y[4];
  real z[4];

  real f1[4];
  real f2[4];
  real f3[4];

  real MASS, NL, NLSD, LAP, DIV, FORCE, FORCESD, PRE, PRESD;

  LocalField *u_fine;
  LocalField *p_fine;
  LocalField *u_coarse;
  LocalField *p_coarse;

  LocalField *upTS;
  LocalField *ppTS;
  LocalField *upNL;
  LocalField *ppNL;
  LocalField *unow;
  LocalField *pnow;

  LocalField *force;

  LocalField w11,w12,w13,w21,w22,w23,w31,w32,w33;

public:

  EquationPre_cG1cG1() : Equation(3) {

    start_vector_component = 3;

    cell_no = -1;

    pi = KW_PI;

    C1 = 1.0;
    C2 = 1.0;
    K = 20.0;

    Cs = 0.1;
  }
  
  real InitLocalFields(FiniteElement *fe){

    upTS = new LocalField[3];
    for (int i=0;i<3;i++) upTS[i].AttachFiniteElement(fe);

    ppTS = new LocalField[1];
    ppTS[0].AttachFiniteElement(fe);

    upNL = new LocalField[3];
    for (int i=0;i<3;i++) upNL[i].AttachFiniteElement(fe);

    ppNL = new LocalField[1];
    ppNL[0].AttachFiniteElement(fe);

    unow = new LocalField[3];
    for (int i=0;i<3;i++) unow[i].AttachFiniteElement(fe);

    pnow = new LocalField[1];
    pnow[0].AttachFiniteElement(fe);

    u_fine = new LocalField[3];
    for (int i=0;i<3;i++) u_fine[i].AttachFiniteElement(fe);

    p_fine = new LocalField[1];
    p_fine[0].AttachFiniteElement(fe);

    u_coarse = new LocalField[3];
    for (int i=0;i<3;i++) u_coarse[i].AttachFiniteElement(fe);

    p_coarse = new LocalField[1];
    p_coarse[0].AttachFiniteElement(fe);
    
    force = new LocalField[3];
    for (int i=0;i<3;i++) force[i].AttachFiniteElement(fe);

  }

  real DeleteLocalFields(){

    delete [] upTS;
    delete [] ppTS;
    delete [] upNL;
    delete [] ppNL;
    delete [] unow;
    delete [] pnow;
    delete [] u_fine;
    delete [] p_fine;
    delete [] u_coarse;
    delete [] p_coarse;
     
    delete [] force;
  }

  real IntegrateLHS(ShapeFunction &p, ShapeFunction &q){
    
    LAP = p.dx()*q.dx() + p.dy()*q.dy() + p.dz()*q.dz();
    
    return(
	   
	   LAP
	   
	   );
    
  }
  
  //--------------------------------------------------------------------------------

  real IntegrateRHS(ShapeFunction &q){

    cell_no = q.GetCellNo();

    if ( (cell_no == 0) && (q.GetDof() == 0) ) InitLocalFields(q.GetFiniteElement());

    if ( q.GetDof() == 0 ) {

      h = 2.0 * q.GetCircumRadius();
      
      for (int i=0;i<3;i++) upTS[i].Refill(cell_no,i,coeff[0]);
      //ppTS[0].Refill(cell_no,0,coeff[1]);

      for (int i=0;i<3;i++) upNL[i].Refill(cell_no,i,coeff[2]);
      //ppNL[0].Refill(cell_no,0,coeff[3]);

      for (int i=0;i<3;i++) unow[i].Refill(cell_no,i,coeff[4]);
      //pnow[0].Refill(cell_no,0,coeff[5]);

      dt = coeff[6]->GetConstantValue();
      //t = coeff[7]->GetConstantValue();
      Re = coeff[8]->GetConstantValue();

      for (int i=0;i<4;i++) x[i] = q.GetCoord(i,0);
      for (int i=0;i<4;i++) y[i] = q.GetCoord(i,1);
      for (int i=0;i<4;i++) z[i] = q.GetCoord(i,2);
      
      //for (int i=0;i<4;i++) f1[i] = (1.0 / Re) * 32.0 * (y[i] * (1.0 - y[i]) + z[i] * (1.0 - z[i])); 
      for (int i=0;i<4;i++) f1[i] = 0.0;
      
      //for (int i=0;i<4;i++) f2[i] = 0.001 * sin(10.0 * pi * x[i]);
      for (int i=0;i<4;i++) f2[i] = 0.0;
      
      //for (int i=0;i<4;i++) f3[i] = 0.001 * sin(10.0 * pi * x[i]);
      for (int i=0;i<4;i++) f3[i] = 0.0;
      
      force[0].Refill(cell_no,f1);
      force[1].Refill(cell_no,f2);
      force[2].Refill(cell_no,f3);

      normU = sqrt(sqr(unow[0].GetMeanValue()) + sqr(unow[1].GetMeanValue()) + sqr(unow[2].GetMeanValue()));

      if (h * Re > 1.0) {
	//delta1 = 0.5 / sqrt( 1.0/sqr(dt) + sqr(normU/h) );
	delta1 = C1 * h;
      } else {
	delta1 = C1 * sqr(h);
      }
      if (h * Re > 1.0) {
	delta2 = C2 * h;
      } else {
	delta2 = C2 * sqr(h);
      }
    }

    w11 = upTS[0]; w11.Add(upNL[0]);
    w22 = upTS[1]; w22.Add(upNL[1]);
    w33 = upTS[2]; w33.Add(upNL[2]);
    
    NL    = (upTS[0].dx()+unow[0].dx())*w11*q.dx() + 
            (upTS[0].dy()+unow[0].dy())*w22*q.dx() + 
            (upTS[0].dz()+unow[0].dz())*w33*q.dx() + 
            (upTS[1].dx()+unow[1].dx())*w11*q.dy() + 
            (upTS[1].dy()+unow[1].dy())*w22*q.dy() + 
            (upTS[1].dz()+unow[1].dz())*w33*q.dy() + 
            (upTS[2].dx()+unow[2].dx())*w11*q.dz() + 
            (upTS[2].dy()+unow[2].dy())*w22*q.dz() + 
            (upTS[2].dz()+unow[2].dz())*w33*q.dz();

    NL *= 0.25;


    DIV   = ( upTS[0].dx() + unow[0].dx() +
              upTS[1].dy() + unow[1].dy() +
              upTS[2].dz() + unow[2].dz() ) * q;

    DIV *= 0.5;

    FORCE = force[0]*q.dx() + force[1]*q.dy() + force[2]*q.dz();

    if ( (cell_no >= (q.GetNoCells()-1)) && (q.GetDof() == q.GetDim()) ) DeleteLocalFields();


    return(
	   
	   FORCE - NL - (1.0/delta1)*DIV
	   
           );

  }

  
};

#endif
