#ifndef __SETTINGS_NAVIER_STOKES_HH
#define __SETTINGS_NAVIER_STOKES_HH

#include "Settings.hh"

///
class SettingsNavierStokes : public Settings {
public:
  
  /// Constructor
  SettingsNavierStokes() : Settings() {

	 // Create the parameters and specify default values
	 
	 Add("reynolds number",                       type_real,   100.0);
	 
	 Add("turbulent flow",                        type_bool, false);
    Add("turbulent inflow",                      type_bool, false);
	 
    Add("compute reynolds stresses",             type_bool, false);
    Add("write reynolds stresses",               type_bool, false);

    Add("compute projections",                   type_bool, false);
    Add("variational multscale eddy viscosity",  type_bool, false);
    Add("smagorinsky eddy viscosity",            type_bool, false);

    Add("write couette pertubations",            type_bool, false);
    Add("write poiseuille pertubations",         type_bool, false);

    // FIXME: All the below are related to boundary conditions
	 
    Add("backward facing step",                  type_bool, false);
    Add("bluff body",                            type_bool, false);
    Add("jet",                                   type_bool, false);
    Add("driven cavity",                         type_bool, false);
    Add("LES benchmark",                         type_bool, false);

    Add("slip bc x0",                            type_bool, false);
    Add("slip bc x1",                            type_bool, false);
    Add("slip bc y0",                            type_bool, false);
    Add("slip bc y1",                            type_bool, false);
    Add("slip bc z0",                            type_bool, false);
    Add("slip bc z1",                            type_bool, false);

    Add("no slip bc x0",                         type_bool, false);
    Add("no slip bc x1",                         type_bool, false);
    Add("no slip bc y0",                         type_bool, false);
    Add("no slip bc y1",                         type_bool, false);
    Add("no slip bc z0",                         type_bool, false);
    Add("no slip bc z1",                         type_bool, false);

    Add("periodic bc x",                         type_bool, false);
    Add("periodic bc y",                         type_bool, false);
    Add("periodic bc z",                         type_bool, false);

    Add("inflow bc x0",                          type_bool, false);
    Add("outflow bc x0",                         type_bool, false);
    Add("outflow bc x1",                         type_bool, false);

    Add("couette bc y",                          type_bool, false);

    Add("driven cavity bc x0",                   type_bool, false);
    Add("driven cavity bc y1",                   type_bool, false);

    Add("dual drag bc",                          type_bool, false);
 
    Add("bnd x0",                                type_real, 0.0);
    Add("bnd x1",                                type_real, 1.0);
    Add("bnd y0",                                type_real, 0.0);
    Add("bnd y1",                                type_real, 1.0);
    Add("bnd z0",                                type_real, 0.0);
    Add("bnd z1",                                type_real, 1.0);
	 
  };
  
  /// Destructor
  ~SettingsNavierStokes() {};
};

#endif
