// Original author: Johan Hoffman. Added to Kraftwerk 2002-01-14.
// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#ifndef StatProblemDualPressure_h
#define StatProblemDualPressure_h

#include "StatProblem.hh"

class StatProblemDualPressure: 
public StatProblem {
  
 public:
  
  StatProblemDualPressure(Grid *grid);
  
  ~StatProblemDualPressure(){}
  
  int Timestep( MV_ColMat<real> &val, real time, real deltat );
  
  int ComputeDiscreteResidual( real time, real deltat, real& norm_2, real& norm_inf );
  
 private:

  int InitEquationData( real time, real deltat );

  int SetDomain();
  
  Field u1_prev_TS,u2_prev_TS,u3_prev_TS,p_prev_TS;
  Field U1,U2,U3;
  Field u1,u2,u3,p;
  Field dt,t,Re;
  
  real ReNo;
  real bnd_x0,bnd_x1,bnd_y0,bnd_y1,bnd_z0,bnd_z1;
  
  MV_ColMat<real> bndx0;
  MV_ColMat<real> bndx1;
  MV_ColMat<real> bndy0;
  MV_ColMat<real> bndy1;
  MV_ColMat<real> bndz0;
  MV_ColMat<real> bndz1;
  
  bool periodic_bc_x;
  bool periodic_bc_y;
  bool periodic_bc_z;

  bool outflow_bc_x0;
  bool outflow_bc_x1;

};


#endif
