#include "StatProblemNS.hh"

StatProblemNS::StatProblemNS( MV_Vector<Grid>* g ) : StatProblem( g ){ 
  
  noeq = 4;

  settings->Get("reynolds number", &ReNo);

  settings->Get("compute projections",&compute_projections);

  settings->Get("channel length", &L);
 
  settings->Get("reynolds number", &ReNo);



  // define Fields to be attached to equation
  u1_prev_TS = new Field(&(*GRID)(compGrid));
  u2_prev_TS = new Field(&(*GRID)(compGrid));
  u3_prev_TS = new Field(&(*GRID)(compGrid));
  p_prev_TS = new Field(&(*GRID)(compGrid));
  
  u1_prev_NL = new Field(&(*GRID)(compGrid));
  u2_prev_NL = new Field(&(*GRID)(compGrid));
  u3_prev_NL = new Field(&(*GRID)(compGrid));
  p_prev_NL = new Field(&(*GRID)(compGrid));
  
  u1 = new Field(&(*GRID)(compGrid));
  u2 = new Field(&(*GRID)(compGrid));
  u3 = new Field(&(*GRID)(compGrid));
  p = new Field(&(*GRID)(compGrid));
  
  dt = new Field(&(*GRID)(compGrid));
  t = new Field(&(*GRID)(compGrid));
  Re = new Field(&(*GRID)(compGrid));
  
  if ( compute_projections ){
    u1_coarse = new Field(&(*GRID)(compGrid));
    u2_coarse = new Field(&(*GRID)(compGrid));
    u3_coarse = new Field(&(*GRID)(compGrid));
    p_coarse = new Field(&(*GRID)(compGrid));

    u1_fine = new Field(&(*GRID)(compGrid));
    u2_fine = new Field(&(*GRID)(compGrid));
    u3_fine = new Field(&(*GRID)(compGrid));
    p_fine = new Field(&(*GRID)(compGrid));
  }

  settings->Get("driven cavity", &driven_cavity);
  settings->Get("backward facing step", &backward_facing_step);
  settings->Get("bluff body", &bluff_body);
  settings->Get("jet", &jet);

  settings->Get("slip bc x0", &slip_bc_x0);
  settings->Get("slip bc x1", &slip_bc_x1);
  settings->Get("slip bc y0", &slip_bc_y0);
  settings->Get("slip bc y1", &slip_bc_y1);
  settings->Get("slip bc z0", &slip_bc_z0);
  settings->Get("slip bc z1", &slip_bc_z1);

  settings->Get("no slip bc x0", &no_slip_bc_x0);
  settings->Get("no slip bc x1", &no_slip_bc_x1);
  settings->Get("no slip bc y0", &no_slip_bc_y0);
  settings->Get("no slip bc y1", &no_slip_bc_y1);
  settings->Get("no slip bc z0", &no_slip_bc_z0);
  settings->Get("no slip bc z1", &no_slip_bc_z1);

  settings->Get("periodic bc x", &periodic_bc_x);
  settings->Get("periodic bc y", &periodic_bc_y);
  settings->Get("periodic bc z", &periodic_bc_z);

  settings->Get("inflow bc x0", &inflow_bc_x0);
  settings->Get("outflow bc x0", &outflow_bc_x0);
  settings->Get("outflow bc x1", &outflow_bc_x1);

  settings->Get("couette bc y", &couette_bc_y);

  settings->Get("driven cavity bc x0", &driven_cavity_bc_x0);
  settings->Get("driven cavity bc y1", &driven_cavity_bc_y1);

  settings->Get("dual drag bc", &dual_drag_bc);

  settings->Get("compute projections",&compute_projections);

  // define the computational domain
  SetDomain();

}
  
  
int StatProblemNS::Timestep( MV_ColMat<real>& val, real time, real deltat ){

  return 0;
  
}



int StatProblemNS::CompMatrixVectorProduct( MV_ColMat<real>* vmat, int comp ){

  MV_ColMat<real> val(noNodes,4);

  int i,j;
  for ( i=0; i < noNodes; i++ ){
    for ( j=0; j < 4; j++ ){
      val(i,j) = (*vmat)(j+i*noeq,comp);
      //     cout << "vmat(" << i << "," << j << ") = " << val(i,j) << endl;
    } 
  }
  MV_Vector<Field*> data(15);
  if ( compute_projections ) data.newsize(23);

  data(0)  = u1_prev_TS;
  data(1)  = u2_prev_TS;
  data(2)  = u3_prev_TS;
  data(3)  = p_prev_TS;
  
  data(4)  = u1_prev_NL;
  data(5)  = u2_prev_NL;
  data(6)  = u3_prev_NL;
  data(7)  = p_prev_NL;
  
  data(8) = u1;
  data(9) = u2;
  data(10) = u3;
  data(11) = p;
  
  data(12) = dt;
  data(13) = t;
  data(14) = Re;
  
  if ( compute_projections ){
    u1_coarse->InitByNodes(uk_coarse,0);
    u2_coarse->InitByNodes(uk_coarse,1);
    u3_coarse->InitByNodes(uk_coarse,2);
    p_coarse->InitByNodes(uk_coarse,3);

    u1_fine->InitByNodes(uk_fine,0);
    u2_fine->InitByNodes(uk_fine,1);
    u3_fine->InitByNodes(uk_fine,2);
    p_fine->InitByNodes(uk_fine,3);

    data(15) = u1_coarse;
    data(16) = u2_coarse;
    data(17) = u3_coarse;
    data(18) = p_coarse;

    data(19) = u1_fine;
    data(20) = u2_fine;
    data(21) = u3_fine;
    data(22) = p_fine;
  }
  
  // attach field vector to equation
  equation->attachField(data);
  
  // initialize fields
  u1_prev_TS->InitByNodes(uk_prev_TS,0);
  u2_prev_TS->InitByNodes(uk_prev_TS,1);
  u3_prev_TS->InitByNodes(uk_prev_TS,2);
  //p_prev_TS->InitByNodes(uk_prev_TS,3);
  
  /*
  u1_prev_NL->InitByNodes(uk_prev_NL,0);
  u2_prev_NL->InitByNodes(uk_prev_NL,1);
  u3_prev_NL->InitByNodes(uk_prev_NL,2);
  p_prev_NL->InitByNodes(uk_prev_NL,3);
  */

  u1_prev_NL->InitByNodes(&val,0);
  u2_prev_NL->InitByNodes(&val,1);
  u3_prev_NL->InitByNodes(&val,2);
  p_prev_NL->InitByNodes(&val,3);
  
  u1->InitByNodes(&val,0);
  u2->InitByNodes(&val,1);
  u3->InitByNodes(&val,2);
  p->InitByNodes(&val,3);
  
  dt->InitByConstant(&deltat);
  t->InitByConstant(&time);
  Re->InitByConstant(&ReNo);
      

  // initialize inner variables 
  Scalar *temp;
  int ierr;
  
  PetscFunctionBegin;

  int noBF    = 4;

  Vec Au;
  ierr = VecCreate(MPI_COMM_WORLD, PETSC_DECIDE, noeq*noNodes, &Au);
  ierr = VecSetFromOptions(Au);CHKERRQ(ierr);

  // compute norm of Au 
  real norm;
  ierr = VecNorm( Au, NORM_2, &norm );
  if ( debug_mode ) cout << "Au norm: " << norm << endl<<flush;

  MV_ColMat<real> bres(noeq,noBF);

  FET4n3D FE(&(*GRID)(compGrid)); 

  MV_Vector<real> resultElVec(noBF*noeq);
  MV_Vector<int>  rowIdx(noBF*noeq);

  real Vol;
  int bf,n,cnt,el,nd;
  equation->setLocSolSize( noBF, noeq );
  for( el=0; el < noElms; el++ ){
    FE.refill(el);
    Vol = FE.volume();
    for ( bf=0; bf < noBF; bf++ ){
      equation->MatrixVectorProductExactInt( bres, bf, el, FE );
      for ( j=0; j < noeq; j++ ){
	//cout << "Au = " << bres(j,bf) << endl;
	resultElVec(j + bf*noeq) = bres(j,bf) * Vol;	
      }
    }
    cnt = 0;
    for ( j=0; j < noeq; j++ ) rowIdx(cnt++) = j + noeq*FE.n1();	
    for ( j=0; j < noeq; j++ ) rowIdx(cnt++) = j + noeq*FE.n2();	
    for ( j=0; j < noeq; j++ ) rowIdx(cnt++) = j + noeq*FE.n3();	
    for ( j=0; j < noeq; j++ ) rowIdx(cnt++) = j + noeq*FE.n4();	
    n = cnt;
    ierr = VecSetValues(Au,n,&rowIdx(0),&resultElVec(0),ADD_VALUES); CHKERRQ(ierr);
  }

  ierr = VecAssemblyBegin(Au); CHKERRQ(ierr);
  ierr = VecAssemblyEnd(Au); CHKERRQ(ierr);

  // compute norm of Au 
  ierr = VecNorm( Au, NORM_2, &norm );
  if ( debug_mode ) cout << "Au norm: " << norm << endl<<flush;
  
  MV_Vector<real> DC_test(noeq*noNodes);
  for ( i=0; i < DirichletIdx.size(); i++ ) DC_test(DirichletIdx(i))++; 

  // update solution vector
  ierr = VecGetArray(Au, &temp); CHKERRQ( ierr );
  for ( i=0; i < 4*noNodes; i++ ){
    if ( DC_test(i) == 0 ) (*vmat)(i,comp+1) = temp[i];
  }
  ierr = VecRestoreArray(Au, &temp); CHKERRQ( ierr );

  // destroy solution variables 
  ierr = VecDestroy(Au); CHKERRQ(ierr);
  
  return 0;
  
}


int StatProblemNS::CompRHS( MV_Vector<real>* bvec ){

  // define Fields to be attached to equation
  MV_Vector<Field*> data(15);
  if ( compute_projections ) data.newsize(23);

  data(0)  = u1_prev_TS;
  data(1)  = u2_prev_TS;
  data(2)  = u3_prev_TS;
  data(3)  = p_prev_TS;
  
  data(4)  = u1_prev_NL;
  data(5)  = u2_prev_NL;
  data(6)  = u3_prev_NL;
  data(7)  = p_prev_NL;
  
  data(8) = u1;
  data(9) = u2;
  data(10) = u3;
  data(11) = p;
  
  data(12) = dt;
  data(13) = t;
  data(14) = Re;
  
  if ( compute_projections ){
    u1_coarse->InitByNodes(uk_coarse,0);
    u2_coarse->InitByNodes(uk_coarse,1);
    u3_coarse->InitByNodes(uk_coarse,2);
    p_coarse->InitByNodes(uk_coarse,3);

    u1_fine->InitByNodes(uk_fine,0);
    u2_fine->InitByNodes(uk_fine,1);
    u3_fine->InitByNodes(uk_fine,2);
    p_fine->InitByNodes(uk_fine,3);

    data(15) = u1_coarse;
    data(16) = u2_coarse;
    data(17) = u3_coarse;
    data(18) = p_coarse;

    data(19) = u1_fine;
    data(20) = u2_fine;
    data(21) = u3_fine;
    data(22) = p_fine;
  }
  
  // attach field vector to equation
  equation->attachField(data);
  
  // initialize fields
  u1_prev_TS->InitByNodes(uk_prev_TS,0);
  u2_prev_TS->InitByNodes(uk_prev_TS,1);
  u3_prev_TS->InitByNodes(uk_prev_TS,2);
  //p_prev_TS->InitByNodes(uk_prev_TS,3);
  
  /*
  u1_prev_NL->InitByNodes(val,0);
  u2_prev_NL->InitByNodes(val,1);
  u3_prev_NL->InitByNodes(val,2);
  p_prev_NL->InitByNodes(val,3);
  
  u1->InitByNodes(val,0);
  u2->InitByNodes(val,1);
  u3->InitByNodes(val,2);
  p->InitByNodes(val,3);
  */
  
  dt->InitByConstant(&deltat);
  t->InitByConstant(&time);
  Re->InitByConstant(&ReNo);
      
  // initialize inner variables 
  Scalar *temp;
  int ierr;
  
  PetscFunctionBegin;

  int noBF    = 4;

  Vec b;
  ierr = VecCreate(MPI_COMM_WORLD, PETSC_DECIDE, noeq*noNodes, &b);
  ierr = VecSetFromOptions(b);CHKERRQ(ierr);

  MV_ColMat<real> bres(noeq,noBF);

  FET4n3D FE(&(*GRID)(compGrid)); 

  MV_Vector<real> resultElVec(noBF*noeq);
  MV_Vector<int>  rowIdx(noBF*noeq);

  real Vol;
  int bf,n,cnt,el,nd,i,j;
  equation->setLocSolSize( noBF, noeq );
  for( el=0; el < noElms; el++ ){
    FE.refill(el);
    Vol = FE.volume();
    for ( bf=0; bf < noBF; bf++ ){
      equation->residualExactInt( bres, bf, el, &(*GRID)(compGrid) );
      for ( j=0; j < noeq; j++ ){
	resultElVec(j + bf*noeq) = bres(j,bf) * Vol;	
      }
    }
    cnt = 0;
    for ( j=0; j < noeq; j++ ) rowIdx(cnt++) = j + noeq*FE.n1();	
    for ( j=0; j < noeq; j++ ) rowIdx(cnt++) = j + noeq*FE.n2();	
    for ( j=0; j < noeq; j++ ) rowIdx(cnt++) = j + noeq*FE.n3();	
    for ( j=0; j < noeq; j++ ) rowIdx(cnt++) = j + noeq*FE.n4();	
    n = cnt;
    ierr = VecSetValues(b,n,&rowIdx(0),&resultElVec(0),ADD_VALUES); CHKERRQ(ierr);
  }

  ierr = VecAssemblyBegin(b); CHKERRQ(ierr);
  ierr = VecAssemblyEnd(b); CHKERRQ(ierr);


  // compute norm of b 
  real norm;
  ierr = VecNorm( b, NORM_2, &norm );
  if ( debug_mode ) cout << "Load norm: " << norm << endl<<flush;
  
  // set dirichlet bc
  SetBoundaryConditions(b);

  // update solution vector
  bvec->newsize(4*noNodes);
  ierr = VecGetArray(b, &temp); CHKERRQ( ierr );
  for ( i=0; i < 4*noNodes; i++ ){
    (*bvec)(i) = temp[i];
  }
  ierr = VecRestoreArray(b, &temp); CHKERRQ( ierr );

  // destroy solution variables 
  ierr = VecDestroy(b); CHKERRQ(ierr);
  
  return 0;
  
}

/*
int StatProblemNS::CompNonLinearResidual( MV_ColMat<real>* vmat, int comp ){

  cout << "check:" << vmat->size(0) << ", " << vmat->size(1) << endl;
  MV_ColMat<real> val(noNodes,4);

  int i,j;
  for ( i=0; i < noNodes; i++ ){
    for ( j=0; j < 4; j++ ) val(i,j) = (*vmat)(j+i*noeq,comp);
    cout << val(i,j) << endl;
  }

  // define Fields to be attached to equation
  Field u1_prev_TS(&(*GRID)(compGrid));
  Field u2_prev_TS(&(*GRID)(compGrid));
  Field u3_prev_TS(&(*GRID)(compGrid));
  Field p_prev_TS(&(*GRID)(compGrid));
  
  Field u1_prev_NL(&(*GRID)(compGrid));
  Field u2_prev_NL(&(*GRID)(compGrid));
  Field u3_prev_NL(&(*GRID)(compGrid));
  Field p_prev_NL(&(*GRID)(compGrid));
  
  Field u1(&(*GRID)(compGrid));
  Field u2(&(*GRID)(compGrid));
  Field u3(&(*GRID)(compGrid));
  Field p(&(*GRID)(compGrid));
  
  Field dt(&(*GRID)(compGrid));
  Field t(&(*GRID)(compGrid));
  Field Re(&(*GRID)(compGrid));
  
  MV_Vector<Field*> data(15);
  if ( compute_projections ) data.newsize(23);

  data(0)  = u1_prev_TS;
  data(1)  = u2_prev_TS;
  data(2)  = u3_prev_TS;
  data(3)  = p_prev_TS;
  
  data(4)  = u1_prev_NL;
  data(5)  = u2_prev_NL;
  data(6)  = u3_prev_NL;
  data(7)  = p_prev_NL;
  
  data(8) = u1;
  data(9) = u2;
  data(10) = u3;
  data(11) = p;
  
  data(12) = dt;
  data(13) = t;
  data(14) = Re;
  
  if ( compute_projections ){
    Field u1_coarse(&(*GRID)(compGrid));
    Field u2_coarse(&(*GRID)(compGrid));
    Field u3_coarse(&(*GRID)(compGrid));
    Field p_coarse(&(*GRID)(compGrid));

    Field u1_fine(&(*GRID)(compGrid));
    Field u2_fine(&(*GRID)(compGrid));
    Field u3_fine(&(*GRID)(compGrid));
    Field p_fine(&(*GRID)(compGrid));

    u1_coarse->InitByNodes(uk_coarse,0);
    u2_coarse->InitByNodes(uk_coarse,1);
    u3_coarse->InitByNodes(uk_coarse,2);
    p_coarse->InitByNodes(uk_coarse,3);

    u1_fine->InitByNodes(uk_fine,0);
    u2_fine->InitByNodes(uk_fine,1);
    u3_fine->InitByNodes(uk_fine,2);
    p_fine->InitByNodes(uk_fine,3);

    data(15) = u1_coarse;
    data(16) = u2_coarse;
    data(17) = u3_coarse;
    data(18) = p_coarse;

    data(19) = u1_fine;
    data(20) = u2_fine;
    data(21) = u3_fine;
    data(22) = p_fine;
  }
  
  // attach field vector to equation
  equation->attachField(data);
  
  // initialize fields
  u1_prev_TS->InitByNodes(uk_prev_TS,0);
  u2_prev_TS->InitByNodes(uk_prev_TS,1);
  u3_prev_TS->InitByNodes(uk_prev_TS,2);
  //p_prev_TS->InitByNodes(uk_prev_TS,3);
  
  cout << "check: " << u1_prev_TS.getRep() << endl;
  cout << "check: " << data(0)->getRep() << endl;
  cout << "check: " << equation->COEFF(0)->getRep() << endl;
  cout << "check: " << equation->COEFF(0)->getNoNodes() << endl;
  cout << "check: " << equation->COEFF(0)->getRep() << endl;
  cout << "check: " << equation->COEFF(0)->getNoNodes() << endl;
  cout << "check: " << endl;

  u1_prev_NL->InitByNodes(&val,0);
  u2_prev_NL->InitByNodes(&val,1);
  u3_prev_NL->InitByNodes(&val,2);
  p_prev_NL->InitByNodes(&val,3);
  
  u1->InitByNodes(&val,0);
  u2->InitByNodes(&val,1);
  u3->InitByNodes(&val,2);
  p->InitByNodes(&val,3);
  
  dt->InitByConstant(&deltat);
  t->InitByConstant(&time);
  Re->InitByConstant(&ReNo);
      
  cout << "checkb" << endl;
  //InitFields( &val );
  cout << "check: " << equation->COEFF(0)->getNoNodes() << endl;
  cout << "check: " << equation->COEFF(0)->getNoNodes() << endl;
  cout << "check: " << equation->COEFF(0)->getRep() << endl;
  cout << "check: " << equation->noEq() << endl;
  cout << "check: " << equation->COEFF(0)->getRep() << endl;
  cout << "check: " << equation->COEFF(0)->getNoNodes() << endl;
  cout << "checka" << endl;
  cout << "check: " << equation->COEFF(0)->getRep() << endl;

  // initialize inner variables 
  Scalar *temp;
  int ierr;
  
  PetscFunctionBegin;

  int noBF = 4;

  Vec res;
  ierr = VecCreate(MPI_COMM_WORLD, PETSC_DECIDE, noeq*noNodes, &res);
  ierr = VecSetFromOptions(res);CHKERRQ(ierr);

  cout << "check: " << equation->COEFF(0)->getRep() << endl;
  MV_ColMat<real> bres(noeq,noBF);

  FET4n3D FE(&(*GRID)(compGrid)); 

  MV_Vector<real> resultElVec(noBF*noeq);
  MV_Vector<int>  rowIdx(noBF*noeq);

  cout << "checkb" << endl;
  real Vol;
  int bf,n,cnt,el,nd;
  equation->setLocSolSize( noBF, noeq );
  for( el=0; el < noElms; el++ ){
    FE.refill(el);
    Vol = FE.volume();
    for ( bf=0; bf < noBF; bf++ ){
      equation->NonLinearResidualExactInt( bres, bf, el, FE );
      for ( j=0; j < noeq; j++ ){
	resultElVec(j + bf*noeq) = bres(j,bf) * Vol;	
      }
    }
    cnt = 0;
    for ( j=0; j < noeq; j++ ) rowIdx(cnt++) = j + noeq*FE.n1();	
    for ( j=0; j < noeq; j++ ) rowIdx(cnt++) = j + noeq*FE.n2();	
    for ( j=0; j < noeq; j++ ) rowIdx(cnt++) = j + noeq*FE.n3();	
    for ( j=0; j < noeq; j++ ) rowIdx(cnt++) = j + noeq*FE.n4();	
    n = cnt;
    ierr = VecSetValues(res,n,&rowIdx(0),&resultElVec(0),ADD_VALUES); CHKERRQ(ierr);
  }

  ierr = VecAssemblyBegin(res); CHKERRQ(ierr);
  ierr = VecAssemblyEnd(res); CHKERRQ(ierr);
  cout << "checkb" << endl;


  // compute norm of res 
  real norm;
  ierr = VecNorm( res, NORM_2, &norm );
  if ( debug_mode ) cout << "Load norm: " << norm << endl<<flush;
  
  // update solution vector
  ierr = VecGetArray(res, &temp); CHKERRQ( ierr );
  for ( i=0; i < 4*noNodes; i++ ){
    (*vmat)(i,comp+1) = temp[i];
  }
  ierr = VecRestoreArray(res, &temp); CHKERRQ( ierr );

  // destroy solution variables 
  ierr = VecDestroy(res); CHKERRQ(ierr);
  
  return 0;
  
}


int StatProblemNS::InitFields( MV_ColMat<real>* val ){

  // define Fields to be attached to equation
  Field u1_prev_TS(&(*GRID)(compGrid));
  Field u2_prev_TS(&(*GRID)(compGrid));
  Field u3_prev_TS(&(*GRID)(compGrid));
  Field p_prev_TS(&(*GRID)(compGrid));
  
  Field u1_prev_NL(&(*GRID)(compGrid));
  Field u2_prev_NL(&(*GRID)(compGrid));
  Field u3_prev_NL(&(*GRID)(compGrid));
  Field p_prev_NL(&(*GRID)(compGrid));
  
  Field u1(&(*GRID)(compGrid));
  Field u2(&(*GRID)(compGrid));
  Field u3(&(*GRID)(compGrid));
  Field p(&(*GRID)(compGrid));
  
  Field dt(&(*GRID)(compGrid));
  Field t(&(*GRID)(compGrid));
  Field Re(&(*GRID)(compGrid));
  
  MV_Vector<Field*> data(15);
  if ( compute_projections ) data.newsize(23);

  data(0)  = u1_prev_TS;
  data(1)  = u2_prev_TS;
  data(2)  = u3_prev_TS;
  data(3)  = p_prev_TS;
  
  data(4)  = u1_prev_NL;
  data(5)  = u2_prev_NL;
  data(6)  = u3_prev_NL;
  data(7)  = p_prev_NL;
  
  data(8) = u1;
  data(9) = u2;
  data(10) = u3;
  data(11) = p;
  
  data(12) = dt;
  data(13) = t;
  data(14) = Re;
  
  if ( compute_projections ){
    Field u1_coarse(&(*GRID)(compGrid));
    Field u2_coarse(&(*GRID)(compGrid));
    Field u3_coarse(&(*GRID)(compGrid));
    Field p_coarse(&(*GRID)(compGrid));

    Field u1_fine(&(*GRID)(compGrid));
    Field u2_fine(&(*GRID)(compGrid));
    Field u3_fine(&(*GRID)(compGrid));
    Field p_fine(&(*GRID)(compGrid));

    u1_coarse->InitByNodes(uk_coarse,0);
    u2_coarse->InitByNodes(uk_coarse,1);
    u3_coarse->InitByNodes(uk_coarse,2);
    p_coarse->InitByNodes(uk_coarse,3);

    u1_fine->InitByNodes(uk_fine,0);
    u2_fine->InitByNodes(uk_fine,1);
    u3_fine->InitByNodes(uk_fine,2);
    p_fine->InitByNodes(uk_fine,3);

    data(15) = u1_coarse;
    data(16) = u2_coarse;
    data(17) = u3_coarse;
    data(18) = p_coarse;

    data(19) = u1_fine;
    data(20) = u2_fine;
    data(21) = u3_fine;
    data(22) = p_fine;
  }
  
  // attach field vector to equation
  equation->attachField(data);
  
  // initialize fields
  u1_prev_TS->InitByNodes(uk_prev_TS,0);
  u2_prev_TS->InitByNodes(uk_prev_TS,1);
  u3_prev_TS->InitByNodes(uk_prev_TS,2);
  //p_prev_TS->InitByNodes(uk_prev_TS,3);
  
  cout << "check: " << u1_prev_TS.getRep() << endl;
  cout << "check: " << data(0)->getRep() << endl;
  cout << "check: " << equation->COEFF(0)->getRep() << endl;
  cout << "check: " << equation->COEFF(0)->getNoNodes() << endl;
  cout << "check: " << equation->COEFF(0)->getRep() << endl;
  cout << "check: " << equation->COEFF(0)->getNoNodes() << endl;
  cout << "check: " << endl;


  u1_prev_NL->InitByNodes(val,0);
  u2_prev_NL->InitByNodes(val,1);
  u3_prev_NL->InitByNodes(val,2);
  p_prev_NL->InitByNodes(val,3);
  
  u1->InitByNodes(val,0);
  u2->InitByNodes(val,1);
  u3->InitByNodes(val,2);
  p->InitByNodes(val,3);
  
  dt->InitByConstant(&deltat);
  t->InitByConstant(&time);
  Re->InitByConstant(&ReNo);
      
  return 0;
}
*/


int StatProblemNS::SetBoundaryConditions( Vec& b ){
  
  // set dirichlet bc
  DirichletIdx.newsize(0); // Should be of size 0
  DirichletVal.newsize(0); // Should be of size 0
  
  int ierr;
  
    
    if ( backward_facing_step ){
      ierr = dirichletBC( b, bfs_bndx1, 0.0, &(*GRID)(compGrid), 4, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bfs_bndx1, 0.0, &(*GRID)(compGrid), 4, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bfs_bndx1, 0.0, &(*GRID)(compGrid), 4, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, bfs_bndy1, 0.0, &(*GRID)(compGrid), 4, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bfs_bndy1, 0.0, &(*GRID)(compGrid), 4, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bfs_bndy1, 0.0, &(*GRID)(compGrid), 4, 2, DirichletIdx, DirichletVal );
    }	  
    
    if ( jet ){
      ierr = dirichletBC_GRID( b, &(*GRID)(compGrid), 4, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC_GRID( b, &(*GRID)(compGrid), 4, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC_GRID( b, &(*GRID)(compGrid), 4, 2, DirichletIdx, DirichletVal );
    }
    
    if ( bluff_body ){
      ierr = dirichletBC( b, ob_bndx0, 0.0, &(*GRID)(compGrid), 4, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndx0, 0.0, &(*GRID)(compGrid), 4, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndx0, 0.0, &(*GRID)(compGrid), 4, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndx1, 0.0, &(*GRID)(compGrid), 4, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndx1, 0.0, &(*GRID)(compGrid), 4, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndx1, 0.0, &(*GRID)(compGrid), 4, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndy0, 0.0, &(*GRID)(compGrid), 4, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndy0, 0.0, &(*GRID)(compGrid), 4, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndy0, 0.0, &(*GRID)(compGrid), 4, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndy1, 0.0, &(*GRID)(compGrid), 4, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndy1, 0.0, &(*GRID)(compGrid), 4, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndy1, 0.0, &(*GRID)(compGrid), 4, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndz0, 0.0, &(*GRID)(compGrid), 4, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndz0, 0.0, &(*GRID)(compGrid), 4, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndz0, 0.0, &(*GRID)(compGrid), 4, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndz1, 0.0, &(*GRID)(compGrid), 4, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndz1, 0.0, &(*GRID)(compGrid), 4, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndz1, 0.0, &(*GRID)(compGrid), 4, 2, DirichletIdx, DirichletVal );
    }  
    
    if ( couette_bc_y ){
      ierr = dirichletBC( b, bndy1, 1.0, &(*GRID)(compGrid), 4, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy1, 0.0, &(*GRID)(compGrid), 4, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy1, 0.0, &(*GRID)(compGrid), 4, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, bndy0, -1.0, &(*GRID)(compGrid), 4, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy0,  0.0, &(*GRID)(compGrid), 4, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy0,  0.0, &(*GRID)(compGrid), 4, 2, DirichletIdx, DirichletVal );
    }
    
    if ( inflow_bc_x0 ){
      ierr = dirichletBCInflow( b, bndx0, &(*GRID)(compGrid), 4, 0, 1.0, DirichletIdx, DirichletVal );
      ierr = dirichletBCInflow( b, bndx0, &(*GRID)(compGrid), 4, 1, 0.0, DirichletIdx, DirichletVal );
      ierr = dirichletBCInflow( b, bndx0, &(*GRID)(compGrid), 4, 2, 0.0, DirichletIdx, DirichletVal );
    }
    
    if ( outflow_bc_x1 ){
      ierr = outflowBC( b, bndx1, *u1_prev_TS, ReNo, &(*GRID)(compGrid), 4, 3, DirichletIdx, DirichletVal );
    }
    
    if ( driven_cavity_bc_x0 ){
      ierr = dirichletBC( b, bndx0, 0.0, &(*GRID)(compGrid), 4, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndx0, 1.0, &(*GRID)(compGrid), 4, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndx0, 0.0, &(*GRID)(compGrid), 4, 2, DirichletIdx, DirichletVal );
    }
    
    if ( driven_cavity_bc_y1 ){
      ierr = dirichletBC( b, bndy1, 0.0, &(*GRID)(compGrid), 4, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy1, 0.0, &(*GRID)(compGrid), 4, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy1, 1.0, &(*GRID)(compGrid), 4, 2, DirichletIdx, DirichletVal );
    }
    
    if ( slip_bc_x0 ){
      ierr = dirichletBC( b, bndx0, 0.0, &(*GRID)(compGrid), 4, 0, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_x0 ){
      ierr = dirichletBC( b, bndx0, 0.0, &(*GRID)(compGrid), 4, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndx0, 0.0, &(*GRID)(compGrid), 4, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndx0, 0.0, &(*GRID)(compGrid), 4, 2, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_x1 ){
      ierr = dirichletBC( b, bndx1, 0.0, &(*GRID)(compGrid), 4, 0, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_x1 ){ 
      ierr = dirichletBC( b, bndx1, 0.0, &(*GRID)(compGrid), 4, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndx1, 0.0, &(*GRID)(compGrid), 4, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndx1, 0.0, &(*GRID)(compGrid), 4, 2, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_y0 ){
      ierr = dirichletBC( b, bndy0, 0.0, &(*GRID)(compGrid), 4, 1, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_y0 ){
      ierr = dirichletBC( b, bndy0, 0.0, &(*GRID)(compGrid), 4, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy0, 0.0, &(*GRID)(compGrid), 4, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy0, 0.0, &(*GRID)(compGrid), 4, 2, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_y1 ){
      ierr = dirichletBC( b, bndy1, 0.0, &(*GRID)(compGrid), 4, 1, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_y1 ){
      ierr = dirichletBC( b, bndy1, 0.0, &(*GRID)(compGrid), 4, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy1, 0.0, &(*GRID)(compGrid), 4, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy1, 0.0, &(*GRID)(compGrid), 4, 2, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_z0 ){
      ierr = dirichletBC( b, bndz0, 0.0, &(*GRID)(compGrid), 4, 2, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_z0 ){
      ierr = dirichletBC( b, bndz0, 0.0, &(*GRID)(compGrid), 4, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndz0, 0.0, &(*GRID)(compGrid), 4, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndz0, 0.0, &(*GRID)(compGrid), 4, 2, DirichletIdx, DirichletVal );
    }
    if ( slip_bc_z1 ){
      ierr = dirichletBC( b, bndz1, 0.0, &(*GRID)(compGrid), 4, 2, DirichletIdx, DirichletVal );	  
    } else if ( no_slip_bc_z1 ){
      ierr = dirichletBC( b, bndz1, 0.0, &(*GRID)(compGrid), 4, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndz1, 0.0, &(*GRID)(compGrid), 4, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndz1, 0.0, &(*GRID)(compGrid), 4, 2, DirichletIdx, DirichletVal );
    }
        
    if ( periodic_bc_x ){
      ierr = dirichletBC( b, bndx1, 0.0, &(*GRID)(compGrid), 4, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndx1, 0.0, &(*GRID)(compGrid), 4, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndx1, 0.0, &(*GRID)(compGrid), 4, 2, DirichletIdx, DirichletVal );		
      ierr = dirichletBC( b, bndx1, 0.0, &(*GRID)(compGrid), 4, 3, DirichletIdx, DirichletVal );		
    }
    
    if ( periodic_bc_y ){
      ierr = dirichletBC( b, bndy1, 0.0, &(*GRID)(compGrid), 4, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy1, 0.0, &(*GRID)(compGrid), 4, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndy1, 0.0, &(*GRID)(compGrid), 4, 2, DirichletIdx, DirichletVal );		
      ierr = dirichletBC( b, bndy1, 0.0, &(*GRID)(compGrid), 4, 3, DirichletIdx, DirichletVal );		
    }
    
    if ( periodic_bc_z ){
      ierr = dirichletBC( b, bndz1, 0.0, &(*GRID)(compGrid), 4, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndz1, 0.0, &(*GRID)(compGrid), 4, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, bndz1, 0.0, &(*GRID)(compGrid), 4, 2, DirichletIdx, DirichletVal );		
      ierr = dirichletBC( b, bndz1, 0.0, &(*GRID)(compGrid), 4, 3, DirichletIdx, DirichletVal );		
    }
    
    if ( dual_drag_bc ){
      ierr = dirichletBC( b, ob_bndx0, 1.0, &(*GRID)(compGrid), 4, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndx0, 0.0, &(*GRID)(compGrid), 4, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndx0, 0.0, &(*GRID)(compGrid), 4, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndx1, 1.0, &(*GRID)(compGrid), 4, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndx1, 0.0, &(*GRID)(compGrid), 4, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndx1, 0.0, &(*GRID)(compGrid), 4, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndy0, 1.0, &(*GRID)(compGrid), 4, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndy0, 0.0, &(*GRID)(compGrid), 4, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndy0, 0.0, &(*GRID)(compGrid), 4, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndy1, 1.0, &(*GRID)(compGrid), 4, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndy1, 0.0, &(*GRID)(compGrid), 4, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndy1, 0.0, &(*GRID)(compGrid), 4, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndz0, 1.0, &(*GRID)(compGrid), 4, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndz0, 0.0, &(*GRID)(compGrid), 4, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndz0, 0.0, &(*GRID)(compGrid), 4, 2, DirichletIdx, DirichletVal );
      
      ierr = dirichletBC( b, ob_bndz1, 1.0, &(*GRID)(compGrid), 4, 0, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndz1, 0.0, &(*GRID)(compGrid), 4, 1, DirichletIdx, DirichletVal );
      ierr = dirichletBC( b, ob_bndz1, 0.0, &(*GRID)(compGrid), 4, 2, DirichletIdx, DirichletVal );
    }  
    
  return 0;
  
}


int StatProblemNS::SetTime( real t ){

  time = t;

  return 0;
}
  
int StatProblemNS::SetTimeStep( real k ){

  deltat = k;

  return 0;
}
  
int StatProblemNS::SetEquation( Equation* eq ){

  equation = eq;

  return 0;
}
  

int StatProblemNS::CompUPResidual( real time, real deltat, real& norm_2, real& norm_inf ){
  
  return 0;
  
}



int StatProblemNS::CompNLResidual( real time, real deltat, real& norm_2, real& norm_inf ){
      
  
  return 0;
  
}




int StatProblemNS::SetDomain(){
  
  bndx0.newsize(nsd,2);
  bndx1.newsize(nsd,2);
  bndy0.newsize(nsd,2);
  bndy1.newsize(nsd,2);
  bndz0.newsize(nsd,2);
  bndz1.newsize(nsd,2);
  
  ob_bndx0.newsize(nsd,2);
  ob_bndx1.newsize(nsd,2);
  ob_bndy0.newsize(nsd,2);
  ob_bndy1.newsize(nsd,2);
  ob_bndz0.newsize(nsd,2);
  ob_bndz1.newsize(nsd,2);
  
  bfs_bndx1.newsize(nsd,2);
  bfs_bndy1.newsize(nsd,2);
  
  bndx0(0,0) = 0.0;
  bndx0(0,1) = 0.0;
  bndx0(1,0) = 0.0;
  bndx0(1,1) = 1.0;
  bndx0(2,0) = 0.0;
  bndx0(2,1) = 1.0;
  
  bndx1(0,0) = L;
  bndx1(0,1) = L;
  bndx1(1,0) = 0.0;
  bndx1(1,1) = 1.0;
  bndx1(2,0) = 0.0;
  bndx1(2,1) = 1.0;
  
  bndy0(0,0) = 0.0;
  bndy0(0,1) = L;
  bndy0(1,0) = 0.0;
  bndy0(1,1) = 0.0;
  bndy0(2,0) = 0.0;
  bndy0(2,1) = 1.0;
  
  bndy1(0,0) = 0.0;
  bndy1(0,1) = L;
  bndy1(1,0) = 1.0;
  bndy1(1,1) = 1.0;
  bndy1(2,0) = 0.0;
  bndy1(2,1) = 1.0;
  
  bndz0(0,0) = 0.0;
  bndz0(0,1) = L;
  bndz0(1,0) = 0.0;
  bndz0(1,1) = 1.0;
  bndz0(2,0) = 0.0;
  bndz0(2,1) = 0.0;
  
  bndz1(0,0) = 0.0;
  bndz1(0,1) = L;
  bndz1(1,0) = 0.0;
  bndz1(1,1) = 1.0;
  bndz1(2,0) = 1.0;
  bndz1(2,1) = 1.0;
  
  ob_bndx0(0,0) = 0.375;
  ob_bndx0(0,1) = 0.375;
  ob_bndx0(1,0) = 0.375;
  ob_bndx0(1,1) = 0.625;
  ob_bndx0(2,0) = 0.375;
  ob_bndx0(2,1) = 0.625;
  
  ob_bndx1(0,0) = 0.625;
  ob_bndx1(0,1) = 0.625;
  ob_bndx1(1,0) = 0.375;
  ob_bndx1(1,1) = 0.625;
  ob_bndx1(2,0) = 0.375;
  ob_bndx1(2,1) = 0.625;
  
  ob_bndy0(0,0) = 0.375;
  ob_bndy0(0,1) = 0.625;
  ob_bndy0(1,0) = 0.375;
  ob_bndy0(1,1) = 0.375;
  ob_bndy0(2,0) = 0.375;
  ob_bndy0(2,1) = 0.625;
  
  ob_bndy1(0,0) = 0.375;
  ob_bndy1(0,1) = 0.625;
  ob_bndy1(1,0) = 0.625;
  ob_bndy1(1,1) = 0.625;
  ob_bndy1(2,0) = 0.375;
  ob_bndy1(2,1) = 0.625;
  
  ob_bndz0(0,0) = 0.375;
  ob_bndz0(0,1) = 0.625;
  ob_bndz0(1,0) = 0.375;
  ob_bndz0(1,1) = 0.625;
  ob_bndz0(2,0) = 0.375;
  ob_bndz0(2,1) = 0.375;
  
  ob_bndz1(0,0) = 0.375;
  ob_bndz1(0,1) = 0.625;
  ob_bndz1(1,0) = 0.375;
  ob_bndz1(1,1) = 0.625;
  ob_bndz1(2,0) = 0.625;
  ob_bndz1(2,1) = 0.625;
  
  bfs_bndx1(0,0) = 0.5;
  bfs_bndx1(0,1) = 0.5;
  bfs_bndx1(1,0) = 0.0;
  bfs_bndx1(1,1) = 0.5;
  bfs_bndx1(2,0) = 0.0;
  bfs_bndx1(2,1) = 1.0;
  
  bfs_bndy1(0,0) = 0.0;
  bfs_bndy1(0,1) = 0.5;
  bfs_bndy1(1,0) = 0.5;
  bfs_bndy1(1,1) = 0.5;
  bfs_bndy1(2,0) = 0.0;
  bfs_bndy1(2,1) = 1.0;
  
  return 0;
}



