#ifndef __STAT_PROBLEM_NS_HH
#define __STAT_PROBLEM_NS_HH

#include "StatProblem.hh"

class StatProblemNS: public StatProblem{
  
public:
  
  StatProblemNS( MV_Vector<Grid>* g );
  
  ~StatProblemNS(){}
  
  int CompMatrixVectorProduct( MV_ColMat<real>* vmat, int comp );

  int CompRHS( MV_Vector<real>* bvec);

  //  int CompNonLinearResidual( MV_ColMat<real>* vmat, int comp );

  //  int InitFields( MV_ColMat<real>* val );

  int Timestep( MV_ColMat<real> &val, real time, real deltat );
  
  int CompUPResidual( real time, real deltat, real& norm_2, real& norm_inf );
  
  int CompNLResidual( real time, real deltat, real& norm_2, real& norm_inf );
  
  int SetDomain();
  
  int SetTime( real t );

  int SetTimeStep( real k );

  int SetEquation( Equation* eq );
  
private:
  
  int SetBoundaryConditions( Vec& b );

  real ReNo,deltat,time;

  MV_Vector<int> DirichletIdx;
  MV_Vector<real> DirichletVal;
  
  real L;
  
  Field* u1_prev_TS;
  Field* u2_prev_TS;
  Field* u3_prev_TS;
  Field* p_prev_TS;
  
  Field* u1_prev_NL;
  Field* u2_prev_NL;
  Field* u3_prev_NL;
  Field* p_prev_NL;
  
  Field* u1;
  Field* u2;
  Field* u3;
  Field* p;
  
  Field* dt;
  Field* t;
  Field* Re;
  
  Field* u1_coarse;
  Field* u2_coarse;
  Field* u3_coarse;
  Field* p_coarse;
  
  Field* u1_fine;
  Field* u2_fine;
  Field* u3_fine;
  Field* p_fine;
  
  MV_ColMat<real> bndx0;
  MV_ColMat<real> bndx1;
  MV_ColMat<real> bndy0;
  MV_ColMat<real> bndy1;
  MV_ColMat<real> bndz0;
  MV_ColMat<real> bndz1;
  
  MV_ColMat<real> ob_bndx0;
  MV_ColMat<real> ob_bndx1;
  MV_ColMat<real> ob_bndy0;
  MV_ColMat<real> ob_bndy1;
  MV_ColMat<real> ob_bndz0;
  MV_ColMat<real> ob_bndz1;
  
  MV_ColMat<real> bfs_bndx1;
  MV_ColMat<real> bfs_bndy1;
  
  bool driven_cavity;
  bool backward_facing_step;
  bool bluff_body;
  bool jet;

  bool slip_bc_x0;
  bool slip_bc_x1;
  bool slip_bc_y0;
  bool slip_bc_y1;
  bool slip_bc_z0;
  bool slip_bc_z1;

  bool no_slip_bc_x0;
  bool no_slip_bc_x1;
  bool no_slip_bc_y0;
  bool no_slip_bc_y1;
  bool no_slip_bc_z0;
  bool no_slip_bc_z1;

  bool periodic_bc_x;
  bool periodic_bc_y;
  bool periodic_bc_z;

  bool inflow_bc_x0;
  bool outflow_bc_x0;
  bool outflow_bc_x1;

  bool couette_bc_y;

  bool driven_cavity_bc_x0;
  bool driven_cavity_bc_y1;

  bool dual_drag_bc;


		
};

#endif
