// Original author: Johan Hoffman. Added to Kraftwerk 2002-01-14.
// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#ifndef __STAT_PROBLEM_VELOCITY_HH
#define __STAT_PROBLEM_VELOCITY_HH

#include "StatProblem.hh"

class StatProblemVelocity: public StatProblem{
  
public:
  
  StatProblemVelocity(Grid *grid);
  
  ~StatProblemVelocity(){}
  
  int Timestep( MV_ColMat<real> &val, real time, real deltat );
  
  int ComputeDiscreteResidual( real time, real deltat, real& norm_2, real& norm_inf );

  //int AssembleMatrix( MV_ColMat<real>& val, real time, real deltat, Mat& Matrix );

  int ReAssembleMatrix( bool am );
  
  int SetTurbulentInflow( real trb );

  int SetReynoldsStresses( MV_ColMat<real>* taupnt );

private:
  
  int InitEquationData( real time, real deltat );

  int SetDomain();
  
  int SetBoundaryConditions( SparseMatrix *A, Vector *b );
  
  bool re_assemble_matrix;
  
  Field u1_prev_TS,u2_prev_TS,u3_prev_TS,p_prev_TS;
  Field u1_prev_NL,u2_prev_NL,u3_prev_NL,p_prev_NL;
  Field u1,u2,u3,p;
  Field u1_coarse,u2_coarse,u3_coarse,p_coarse;
  Field u1_fine,u2_fine,u3_fine,p_fine;
  Field dt,t,Re;
  Field tau11,tau12,tau13,tau22,tau23,tau33;

  MV_ColMat<real>* tau;

  MV_ColMat<real> turbulence;

  real ReNo;
		
  real bnd_x0,bnd_x1,bnd_y0,bnd_y1,bnd_z0,bnd_z1;
  
  MV_ColMat<real> bndx0;
  MV_ColMat<real> bndx1;
  MV_ColMat<real> bndy0;
  MV_ColMat<real> bndy1;
  MV_ColMat<real> bndz0;
  MV_ColMat<real> bndz1;
  
  MV_ColMat<real> ob_bndx0;
  MV_ColMat<real> ob_bndx1;
  MV_ColMat<real> ob_bndy0;
  MV_ColMat<real> ob_bndy1;
  MV_ColMat<real> ob_bndz0;
  MV_ColMat<real> ob_bndz1;
  
  MV_ColMat<real> les_bndx0;
  MV_ColMat<real> les_bndx1;
  MV_ColMat<real> les_bndy1;
  MV_ColMat<real> les_bndz0;
  MV_ColMat<real> les_bndz1;
  
  MV_ColMat<real> bfs_bndx1;
  MV_ColMat<real> bfs_bndy1;
  
  bool turbulent_flow;

  bool driven_cavity;
  bool backward_facing_step;
  bool bluff_body;
  bool jet;
  bool LES_benchmark;

  bool slip_bc_x0;
  bool slip_bc_x1;
  bool slip_bc_y0;
  bool slip_bc_y1;
  bool slip_bc_z0;
  bool slip_bc_z1;

  bool no_slip_bc_x0;
  bool no_slip_bc_x1;
  bool no_slip_bc_y0;
  bool no_slip_bc_y1;
  bool no_slip_bc_z0;
  bool no_slip_bc_z1;

  bool periodic_bc_x;
  bool periodic_bc_y;
  bool periodic_bc_z;

  bool inflow_bc_x0;
  bool outflow_bc_x0;
  bool outflow_bc_x1;

  bool couette_bc_y;

  bool driven_cavity_bc_x0;
  bool driven_cavity_bc_y1;

  bool dual_drag_bc;

};

#endif
