// Original author: Johan Hoffman. Added to Kraftwerk 2002-01-28.
// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#ifndef __TRAN_PROBLEM_DUAL_NS_HH
#define __TRAN_PROBLEM_DUAL_NS_HH

#include "kraftwerk.h"
#include "StatProblem.hh"
#include "TranProblem.hh"
#include "EquationDualVel_cG1cG1.hh"
#include "EquationDualPre_cG1cG1.hh"
#include "ElementTetLin.hh"
#include "ElementHexLin.hh"
#include "StatProblemDualVelocity.hh"
#include "StatProblemDualPressure.hh"

enum initialDataDualNS { zero_initial_data_dual };

class TranProblemDualNS:public TranProblem {
public:
  
  TranProblemDualNS(Grid *grid,
						  MV_Vector < GridRelation > *P);
  ~TranProblemDualNS();
  
  int Solve();
  
private:
		
  // Check convergence of velocity (U) - pressure (P) iteration
  bool CheckUPConvergence();
		
  int CompContResidual(MV_ColMat < real > *U,
							  MV_ColMat < real > *Uprev, real t, real dt,
							  MV_Vector < real > &contResidual);
		
  int WriteData(MV_ColMat < real > *U, MV_ColMat < real > *Uprev,
					 real t, real dt);
		
  int SetInitialData();
		
  int SetTolerances();
		
  // Initial variables
  OpenDX *opendx;
  OpenDX *opendx_primal;
		
  StatProblemDualVelocity *sp_vel;
  StatProblemDualPressure *sp_pre;
		
  MV_ColMat < real > print_uk;
		
  //NeighborFE neighbor;
		
  real reynolds_number;
		
  initialDataDualNS initial_data;
		
  int noeq;
  int noeq_vel;
  int noeq_pre;
		
  int input_level_primal;
		
  real UP_tol, stat_tol, UP_max_tol;
  int max_no_UP_iter;
  bool UP_conv;
		
  int UP_it;
  real UP_res_norm_2, UP_res_norm_inf, NL_res_norm_2;
  real pre_res_norm_2, pre_res_norm_inf, vel_res_norm_2, vel_res_norm_inf;
  real maxnorm_du;
		
};

#endif
