// Original author: Johan Hoffman. Added to Kraftwerk 2002-01-28.
// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#include "TranProblemNS.hh"

//----------------------------------------------------------------------------------------------------
TranProblemNS::TranProblemNS(Grid *grid, MV_Vector<GridRelation>* P ) : TranProblem(grid,P) 
{
  // Open the OpenDX file
  char output_file[KW_LINELENGTH];
  settings->Get("output file primal",output_file);
  opendx = new OpenDX(output_file,2,3,1);
  opendx->SetLabel(0,"Velocity");
  opendx->SetLabel(1,"Pressure");
  
  // Get inner variables
  settings->Get("reynolds number",&reynolds_number);
  settings->Get("compute projections",&compute_projections);

  // Check if pertubations should be written
  settings->Get("write couette pertubations",&pert_couette);
  settings->Get("write poiseuille pertubations",&pert_poiseuille);
  pertubation = false;
  if ( pert_poiseuille || pert_couette ) pertubation = true;

  // Initialize output data matlab files
  no_data_entries = 0;
  no_residual_entries = 0;
  no_pertubation_entries = 0;
  if ( write_data )     no_data_entries        = 39;
  if ( write_residuals) no_residual_entries    = 12;
  if ( pertubation )    no_pertubation_entries = 24;
  matlab_data = NULL;
  if ( write_data || write_residuals || pertubation ){
    matlab_data = new Matlab(no_data_entries + no_residual_entries + no_pertubation_entries);
  }
  
  if ( write_data ){
    matlab_data->SetLabel(0,"|| u ||_1");
    matlab_data->SetLabel(1,"|| u ||_2");
    matlab_data->SetLabel(2,"|| u ||_{max}");
    matlab_data->SetLabel(3,"| u |_{1,1}");
    matlab_data->SetLabel(4,"| u |_{1,2}");
    matlab_data->SetLabel(5,"| u |_{1,max}");
    matlab_data->SetLabel(6,"|| u ||_{1,1}");
    matlab_data->SetLabel(7,"|| u ||_{1,2}");
    matlab_data->SetLabel(8,"|| u ||_{1,max}");
    matlab_data->SetLabel(9,"| u |_{2,1}");
    matlab_data->SetLabel(10,"| u |_{2,2}");
    matlab_data->SetLabel(11,"| u |_{2,max}");
    matlab_data->SetLabel(12,"|| u ||_{2,1}");
    matlab_data->SetLabel(13,"|| u ||_{2,2}");
    matlab_data->SetLabel(14,"|| u ||_{2,max}");
    matlab_data->SetLabel(15,"|| p ||_1");
    matlab_data->SetLabel(16,"|| p ||_2");
    matlab_data->SetLabel(17,"|| p ||_{max}");
    matlab_data->SetLabel(18,"| p |_{1,1}");
    matlab_data->SetLabel(19,"| p |_{1,2}");
    matlab_data->SetLabel(20,"| p |_{1,max}");
    matlab_data->SetLabel(21,"|| p ||_{1,1}");
    matlab_data->SetLabel(22,"|| p ||_{1,2}");
    matlab_data->SetLabel(23,"|| p ||_{1,max}");
    matlab_data->SetLabel(24,"| p |_{2,1}");
    matlab_data->SetLabel(25,"| p |_{2,2}");
    matlab_data->SetLabel(26,"| p |_{2,max}");
    matlab_data->SetLabel(27,"|| p ||_{2,1}");
    matlab_data->SetLabel(28,"|| p ||_{2,2}");
    matlab_data->SetLabel(29,"|| p ||_{2,max}");
    matlab_data->SetLabel(30,"|| div u ||_1");
    matlab_data->SetLabel(31,"|| div u ||_2");
    matlab_data->SetLabel(32,"|| div u ||_{max}");
    matlab_data->SetLabel(33,"|| du/dt ||_1");
    matlab_data->SetLabel(34,"|| du/dt ||_2");
    matlab_data->SetLabel(35,"|| du/dt ||_{max}");
    matlab_data->SetLabel(36,"|| dp/dt ||_1");
    matlab_data->SetLabel(37,"|| dp/dt ||_2");
    matlab_data->SetLabel(38,"|| dp/dt ||_{max}");
  }

  if ( write_residuals ){
    matlab_data->SetLabel(no_data_entries+0,"|| R1 ||_1");
    matlab_data->SetLabel(no_data_entries+1,"|| R1 ||_2");
    matlab_data->SetLabel(no_data_entries+2,"|| R1 ||_{max}");
    matlab_data->SetLabel(no_data_entries+3,"|| R2 ||_1");
    matlab_data->SetLabel(no_data_entries+4,"|| R2 ||_2");
    matlab_data->SetLabel(no_data_entries+5,"|| R2 ||_{max}");
    matlab_data->SetLabel(no_data_entries+6,"|| R3 ||_1");
    matlab_data->SetLabel(no_data_entries+7,"|| R3 ||_2");
    matlab_data->SetLabel(no_data_entries+8,"|| R3 ||_{max}");
    matlab_data->SetLabel(no_data_entries+9,"|| R4 ||_1");
    matlab_data->SetLabel(no_data_entries+10,"|| R4 ||_2");
    matlab_data->SetLabel(no_data_entries+11,"|| R4 ||_{max}");
  }

  if ( pertubation ){
    matlab_data->SetLabel(no_data_entries+no_residual_entries+0,"|| u1_{pert} ||_2");
    matlab_data->SetLabel(no_data_entries+no_residual_entries+1,"|| u1_{pert} ||_{max}");
    matlab_data->SetLabel(no_data_entries+no_residual_entries+2,"|| u2_{pert} ||_2");
    matlab_data->SetLabel(no_data_entries+no_residual_entries+3,"|| u2_{pert} ||_{max}");
    matlab_data->SetLabel(no_data_entries+no_residual_entries+4,"|| u3_{pert} ||_2");
    matlab_data->SetLabel(no_data_entries+no_residual_entries+5,"|| u3_{pert} ||_{max}");
    matlab_data->SetLabel(no_data_entries+no_residual_entries+6,"|| du_1/dx_1 ||_2");
    matlab_data->SetLabel(no_data_entries+no_residual_entries+7,"|| du_1/dx_1 ||_{max}");
    matlab_data->SetLabel(no_data_entries+no_residual_entries+8,"|| du_1/dx_2 ||_2");
    matlab_data->SetLabel(no_data_entries+no_residual_entries+9,"|| du_1/dx_2 ||_{max}");
    matlab_data->SetLabel(no_data_entries+no_residual_entries+10,"|| du_1/dx_3 ||_2");
    matlab_data->SetLabel(no_data_entries+no_residual_entries+11,"|| du_1/dx_3 ||_{max}");
    matlab_data->SetLabel(no_data_entries+no_residual_entries+12,"|| du_2/dx_1 ||_2");
    matlab_data->SetLabel(no_data_entries+no_residual_entries+13,"|| du_2/dx_1 ||_{max}");
    matlab_data->SetLabel(no_data_entries+no_residual_entries+14,"|| du_2/dx_2 ||_2");
    matlab_data->SetLabel(no_data_entries+no_residual_entries+15,"|| du_2/dx_2 ||_{max}");
    matlab_data->SetLabel(no_data_entries+no_residual_entries+16,"|| du_2/dx_3 ||_2");
    matlab_data->SetLabel(no_data_entries+no_residual_entries+17,"|| du_2/dx_3 ||_{max}");
    matlab_data->SetLabel(no_data_entries+no_residual_entries+18,"|| du_3/dx_1 ||_2");
    matlab_data->SetLabel(no_data_entries+no_residual_entries+19,"|| du_3/dx_1 ||_{max}");
    matlab_data->SetLabel(no_data_entries+no_residual_entries+20,"|| du_3/dx_2 ||_2");
    matlab_data->SetLabel(no_data_entries+no_residual_entries+21,"|| du_3/dx_2 ||_{max}");
    matlab_data->SetLabel(no_data_entries+no_residual_entries+22,"|| du_3/dx_3 ||_2");
    matlab_data->SetLabel(no_data_entries+no_residual_entries+23,"|| du_3/dx_3 ||_{max}");
  }
}
//----------------------------------------------------------------------------------------------------
TranProblemNS::~TranProblemNS()
{
  // Delete the OpenDX file
  delete opendx;
}      
//----------------------------------------------------------------------------------------------------
int TranProblemNS::Solve()
{
  char tmp[256];
  int i,j,k;
  
  display->Status(0,"Solving Navier-Stokes primal problem.");
  
  real progress = 0.0;
  
  // Update internal variables
  Update();
  
  // Reset OpenDX files
  opendx->Reset();
  opendx_residuals->Reset();
  
  // Reset Matlab file
  if ( matlab_data )
    matlab_data->Reset();

  // Set time t to starting time T0
  t = T0;

  // Set tolerances for solution method
  SetTolerances();

  // Check output levels
  CheckOutputLevel();

  // Reset output files
  opendx->Reset();
  opendx_residuals->Reset();
  
  // Set FEM discretisation 
  //Mapper                    mapper(nsd);
  //ElementTetLin             element;
  //QuadratureRules           quadrule(TET,1);
  EquationVel_cG1cG1        eq_vel;
  EquationPre_cG1cG1        eq_pre;
  Discretiser               discrete_vel(grid,&eq_vel);
  Discretiser               discrete_pre(grid,&eq_pre);
  //  EquationNS_cG1cG1    eq_NS;

  // Set number of equations 
  noeq_vel = eq_vel.noEq();
  noeq_pre = eq_pre.noEq();
  noeq     = noeq_vel + noeq_pre;
  
  // Initialize solution vectors
  uk.newsize(no_nodes,noeq);
  uk_prev_TS.newsize(no_nodes,noeq); 
  uk_prev_NL.newsize(no_nodes,noeq); 
  print_uk.newsize(printNoNodes,noeq); 

  // Set initial data
  initial_data = zero_initial_data;
  if ( pert_couette )    initial_data = couette_flow; 
  if ( pert_poiseuille ) initial_data = poiseuille_flow; 
  SetInitialData();

  // Initialize stationary problems
  sp_vel = new StatProblemVelocity(grid); 
  sp_pre = new StatProblemPressure(grid); 
  //sp_NS  = new StatProblemNS( GRID ); 

  // Attach FEM discretisation to stationary problem
  sp_pre->AttachFEM( &discrete_pre, &eq_pre );
  sp_vel->AttachFEM( &discrete_vel, &eq_vel );
  //sp_NS->AttachFEM( &discrete, &mapper, &quadrule, &element, &eq_NS );

  // Initialize projections (for subgrid modeling)
  if ( compute_projections ){
    uk_fine.newsize(no_nodes,noeq); 
    uk_coarse.newsize(no_nodes,noeq); 
  }

  // Initialize subgrid modeling utilities
  SetSubgridModelingUtilities();

  // Reassemble the momentum matrix or not
  sp_vel->ReAssembleMatrix(true);
  
  // Write data to file
  display->Status(2,"Write data to file");
  if ( write_data )      WriteData( &uk, &uk_prev_TS, t, dt );
  if ( write_residuals ) CompContResidual( &uk, &uk_prev_TS, t, dt, contResidual );
  if ( pertubation )     WritePertubationData( &uk, &uk_prev_TS, t, dt );
    
  // Start time marching 
  display->Status(2,"Start computation at t = %.1f",t);
  while ( t <= T ){
    
    // Save solution if we should
    if ( SaveSample() ){
      
      // Write solution to file
      for ( i=0; i < printNoNodes; i++ ){
	for ( j=0; j < noeq; j++ ) print_uk(i,j) = uk(i,j);
      }
      opendx->AddFrame(grid,&print_uk,t);
      
      // Write Reynolds stresses to file
      if ( write_reynolds_stresses ){
	for ( i=0; i < N; i++ ){
	  (*GRIDREL)(1).projNestedElData(tau,tau_print);
	  
	  // FIXME: Was GRID(compGrid-1) !!!
	  sprintf(tmp,"tau%i_%.6f.inp",i+1,t); writeInp(tmp, grid, tau_print, 1, i);
	  sprintf(tmp,"tau%i_%.6f.m",i+1,t); writeMfile(tmp, grid, tau_print, 1, i);
	}	
      }
      
    }
    
    // Set time step
    GetMeshSize(grid_h);
    dt = grid_h;
    t += dt;
    
    // Set turbulent inflow condition
    sp_vel->SetTurbulentInflow(0.04);

    // Write some output for gdisp
    progress = (t-T0) / (T-T0);
    display->Progress(0,progress,"Primal solution: %.1f %%. Currently at t = %f, dt = %f",progress*100.0,t,dt);
	 
    // Set solution from previous time step
    uk_prev_TS = uk;
    sp_pre->Set_uk_prev_TS(&uk_prev_TS);
    sp_vel->Set_uk_prev_TS(&uk_prev_TS);
    //sp_NS->Set_uk_prev_TS(&uk_prev_TS);
    
    // Set solution from previous outer non linear iteration
    uk_prev_NL = uk;
    sp_pre->Set_uk_prev_NL(&uk_prev_NL);
    sp_vel->Set_uk_prev_NL(&uk_prev_NL);
    //sp_NS->Set_uk_prev_NL(&uk_prev_NL);
    
    // Set coarse-fine projections for subgrid model
    if ( compute_projections ) sp_vel->SetMultiScale(&uk_coarse,&uk_fine);
    
    // Solve non linear problem
    SolveNonLinearProblemUsingFixPoint();
    //SolveNonLinearProblemUsingGMRES();
	 
    // Write solution to file
    for ( i=0; i < printNoNodes; i++ ){
      for ( j=0; j < noeq; j++ ) print_uk(i,j) = uk(i,j);
    }
    //sprintf(tmp,"unow.inp"); writeInp(tmp, &(*GRID)(compGrid-1), print_uk, 3, 0);
    //sprintf(tmp,"pnow.inp"); writeInp(tmp, &(*GRID)(compGrid-1), print_uk, 1, 3);
    //sprintf(tmp,"u_restart.inp"); writeInp(tmp, &(*GRID)(compGrid), uk, 3, 0);
    
    // Write data to file
    display->Status(2,"Write data to file");
    if ( write_data )      WriteData( &uk, &uk_prev_TS, t, dt );
    if ( write_residuals ) CompContResidual( &uk, &uk_prev_TS, t, dt, contResidual );
    if ( pertubation )     WritePertubationData( &uk, &uk_prev_TS, t, dt );
    
    // Check if the solution is stationary
    if ( CheckIfStationary() ) break;
  }

  // Solution computed to end time
  display->Status(0,"Solution has been computed to endtime T = %f",t);

  return 0;
}
//----------------------------------------------------------------------------------------------------
int TranProblemNS::SolveNonLinearProblemUsingFixPoint()
{
  // Start non linear iteration
  for ( NL_it=0; NL_it < max_no_NL_iter; NL_it++ ){ 
      
    // Display information
    display->Status(1,"Starting non-linear iteration %d.",NL_it+1);
    
    // Compute Reynolds stresses - for subgrid model
    if ( compute_reynolds_stresses ){
      display->Status(5,"Estimate Reynolds stresses (Update subgrid model)");
      //scale_expol.compRS( uk, tau, (*GRIDREL)(0), (*GRIDREL)(1), (*GRIDREL)(2), h, H, HH );
      //scale_expol.compRSCov( uk, tau, (*GRIDREL)(0), (*GRIDREL)(1), (*GRIDREL)(2) );
      scale_expol.ComputeReynoldsStresses( &uk, &tau, &(*GRIDREL)(0), &(*GRIDREL)(1), 
					   &(*GRIDREL)(2), &(*GRIDREL)(3), &(*GRIDREL)(4) );
    }
    
    // Compute projections - for subgrid model
    int i,j;
    if ( compute_projections ){
      display->Status(5,"Extract coarse and fine scale field");
      (*GRIDREL)(0).projNodalData(uk,uk_coarse);
      (*GRIDREL)(0).projNodalData(uk_coarse);
      for ( i=0; i < no_nodes; i++ ){
		  for ( j=0; j < noeq; j++ ) uk_fine(i,j) = uk(i,j) - uk_coarse(i,j);
      }
    }
    
    // Start inner iteration
    for ( UP_it=0; UP_it < max_no_UP_iter; UP_it++ ){
      
      // Solve continuity equation
      display->Status(2,"Solving for the pressure.");
      sp_pre->Timestep( uk, t, dt );
      
      // Solve momentum equation
      display->Status(2,"Solving for the velocity.");
      sp_vel->Timestep( uk, t, dt );
      
      // Re assemble the momentum matrix or not
      sp_vel->ReAssembleMatrix(false);
      
      // Compute residuals for inner iteration
      sp_pre->Set_uk(&uk);
      sp_vel->Set_uk(&uk);
      sp_pre->ComputeDiscreteResidual( t, dt,pre_res_norm_2, pre_res_norm_inf );
      sp_vel->ComputeDiscreteResidual( t, dt,vel_res_norm_2, vel_res_norm_inf );
      
      // Check if inner loop has converged
      if ( CheckUPConvergence() ) break;
    }
    
    // Re assemble the momentum matrix or not
    sp_vel->ReAssembleMatrix(true);
    
    // Compute residuals for non linear outer iteration 
    uk_prev_NL = uk;
    sp_pre->Set_uk_prev_NL(&uk_prev_NL);
    sp_vel->Set_uk_prev_NL(&uk_prev_NL);
    sp_pre->ComputeDiscreteResidual( t, dt,pre_res_norm_2, pre_res_norm_inf );
    sp_vel->ComputeDiscreteResidual( t, dt,vel_res_norm_2, vel_res_norm_inf );
    
    // Re assemble the momentum matrix or not
    sp_vel->ReAssembleMatrix(false);
    
    // Check if outer loop has converged
    if ( CheckNonLinearConvergence() ) break;
  }
  
  return 0;
}
//----------------------------------------------------------------------------------------------------
int TranProblemNS::SolveNonLinearProblemUsingGMRES()
{
  /*
  // Solve preconditioned problem Ax = AP^(-1)Px = b, where x=(u,p), 
  // by first solving AP^(-1)v=b through GMRES, then 
  // solve Px = v to get x. Extrem cases is P = A (changes nothing) 
  // and P = I (corresponds to no preconditioner). 
  // AP^(-1)v=b is solved to a tolerance rho/|b| < tol, 
  // where rho is the norm of the residual b-AP^(-1)v = b-Ax,
  // at a maximal number of iterations kmax, 
  // starting from startvector v = Px. 

  sp_NS->SetTime(t);
  sp_NS->SetTimeStep(dt);

  int i,j,k;
  real tmp;

  int k_max = 10;
  real tol  = 0.001;

  real rho, beta, norm_v, norm_residual, nu, tmpvar;

  int sz = noeq * no_nodes;

  MV_ColMat<real> H(k_max+1,k_max+1);
  MV_Vector<real> g(k_max+1);
  MV_Vector<real> c(k_max+1);
  MV_Vector<real> s(k_max+1);
  MV_Vector<real> y(k_max+1);
  MV_Vector<real> w(k_max+1);
  MV_ColMat<real> r(k_max+1,k_max+1);
  

  MV_ColMat<real> v(sz,k_max+1);
  v = 0.0;

  // Compute start residual = b-AP^(-1)u = b-Ax.
  MV_Vector<real> b(sz);
  sp_NS->CompRHS( &b );
  sp_NS->CompMatrixVectorProduct( &v, 0 );
  for ( i=0; i < sz; i++ ) v(i,0) = b(i) - v(i,1);

  norm_residual = 0.0;
  for ( i=0; i < sz; i++ ) norm_residual += sqr(v(i,0));
  norm_residual = sqrt(norm_residual);
  
  for ( i=0; i < sz; i++ ) v(i,0) /= norm_residual;
      
  rho  = norm_residual;
  beta = rho;

  g.newsize(k_max+1);
  g    = 0.0;
  g(0) = rho;

  for ( k=0; k < k_max; k++ ){

  display->Status(1,"Start non linear GMRES iteration number %i",k+1);

  // Compute Krylov vector AP^(-1) v(k) 
  // 2 steps: First solve Px = v, then apply A to x.
  Solve_Px_v( &v, k );
  for ( i=0; i < sz; i++ ) cout << "v(i,k) = " << v(i,k) << endl;
  sp_NS->CompMatrixVectorProduct( &v, k );

  for ( j=0; j < k+1; j++ ){
  H(j,k) = 0.0;
  for ( i=0; i < sz; i++ ) H(j,k)   += v(i,k+1)*v(i,j);
  for ( i=0; i < sz; i++ ) v(i,k+1) -= H(j,k)*v(i,j);
  }

  cout << "H(j,k) = " << H(j,k) << endl;
  cout << "v(i,k+1) = " << v(i,k+1) << endl;

  norm_v = 0.0;
  for ( i=0; i < sz; i++ ) norm_v += sqr(v(i,k+1));
  norm_v = sqrt(norm_v);
    
  H(k+1,k) = norm_v;
    
  // Test for non orthogonality and reorthogonalise if necessary
  for ( i=0; i < sz; i++ ) v(i,k+1) /= norm_v;
    
  if ( k > 0 ){
  for ( j=0; j < k; j++ ) H(j,k) = c(j)*H(j,k) - s(j)*H(j+1,k);
  }
    
  nu   = sqrt( sqr(H(k,k)) + sqr(H(k+1,k)) );
  cout << "nu = " << nu << endl;
  c(k) = H(k,k) / nu;
  s(k) = - H(k+1,k) / nu;
    
  H(k,k)   = c(k)*H(k,k) - s(k)*H(k+1,k);
  H(k+1,k) = 0.0;
    
  tmp    = c(k)*g(k) - s(k)*g(k+1);
  g(k+1) = s(k)*g(k) + c(k)*g(k+1);
  g(k)   = tmp;
    
  rho = fabs(g(k+1));
    
  display->Status(1,"Non linear GMRES residual rho = %.5f",rho);
  if ( rho < tol * norm_residual ){
  display->Status(1,"Non linear GMRES iteration converged!");
  break;
  }
  }
  
  display->Status(2,"Postprocess to obtain solution");

  cout << "check exit" << endl;
  exit(1);

  r = H;
  w = g;
  
  y.newsize(k+1);

  // Solve triangular system ry = w
  y(k) = w(k) / r(k,k);
  for ( i=1; i < k+1; k++ ){
  y(k-i) = w(k-i);
  for ( j=0; j < i; j++ ) y(k-i) -= r(k-j,k-j)*y(k-j);
  y(k-i) /= r(k-i,k-i);
  }
  
  MV_Vector<real> x(sz);
  for ( i=0; i < sz; i++ ){
  for ( j=0; j < k+1; j++ ) x(i) += v(i,j)*y(j);
  }
  
  // Get solution from preconditioned problem (get u from Pu=x)
  Solve_Px_v( &v, k+1 );
      
  return 0;
  */
}
//----------------------------------------------------------------------------------------------------
int TranProblemNS::Solve_Px_v( MV_ColMat<real>* vmat, int comp )
{
  /*
	 Mat A;
	 Vec b;
	 Vec u;
	 Discretiser discrete;
	 discrete.createSeqRHS( &u, &b, (*GRID)(compGrid), noeq_vel );

	 int ierr;
	 Scalar *temp;


	 MV_ColMat<real> vsol(no_nodes,noeq);
	 int i,j;
	 for ( i=0; i < no_nodes; i++ ){
    for ( j=0; j < noeq; j++ ){
	 vsol(i,j) = (*vmat)(j+i*noeq,comp);
	 cout << "vsol = " << vsol(i,j) << endl;
    } 
	 }

	 // Compute Reynolds stresses - for subgrid model
	 if ( compute_reynolds_stresses ){
    if ( DEBUG_MODE ) cout << "Estimate Reynolds stresses (Update subgrid model)" << endl;
    //scale_expol.compRS( uk, tau, (*GRIDREL)(0), (*GRIDREL)(1), (*GRIDREL)(2), h, H, HH );
    //scale_expol.compRSCov( uk, tau, (*GRIDREL)(0), (*GRIDREL)(1), (*GRIDREL)(2) );
    scale_expol.ComputeReynoldsStresses( &uk, &tau, &(*GRIDREL)(0), &(*GRIDREL)(1), 
	 &(*GRIDREL)(2), &(*GRIDREL)(3), &(*GRIDREL)(4) );
	 }
  
	 // Compute projections - for subgrid model
	 int i,j;
	 if ( compute_projections ){
    if ( DEBUG_MODE ) cout << "Extract coarse and fine scale field" << endl;
    (*GRIDREL)(0).projNodalData(uk,uk_coarse);
    (*GRIDREL)(0).projNodalData(uk_coarse);
    for ( i=0; i < no_nodes; i++ ){
	 for ( j=0; j < noeq; j++ ) uk_fine(i,j) = uk(i,j) - uk_coarse(i,j);
    }
	 }
    
	 // Re assemble the momentum matrix or not
	 sp_vel->ReAssembleMatrix(true);

	 // Start inner iteration
	 for ( UP_it=0; UP_it < max_no_UP_iter; UP_it++ ){
    
    // Solve continuity equation
    display->Status(2,"Solving for the pressure.");
    //sp_pre->SolveGivenRHS( vsol, t, dt );
    
    // Solve momentum equation
    display->Status(2,"Solving for the velocity.");
    sp_vel->AssembleMatrix( vsol, t, dt, A );

    ierr = VecGetArray(b, &temp); CHKERRQ( ierr );
    int i;
    for ( i=0; i < no_nodes*noeq_vel; i++ ){
	 temp[i] = (*vmat)(i,comp);
	 cout << "vmat1 = " << (*vmat)(i,comp) << endl;
    }
    ierr = VecRestoreArray(b, &temp); CHKERRQ( ierr );
    
    discrete.solve( A, u, b );
  
    // update solution vector
    ierr = VecGetArray(u, &temp); CHKERRQ( ierr );
    for ( i=0; i < no_nodes*noeq; i++ ){
	 (*vmat)(i,comp) = temp[i];
	 cout << "vmat = " << (*vmat)(i,comp) << endl;
    }
    ierr = VecRestoreArray(u, &temp); CHKERRQ( ierr );
    
    // Re assemble the momentum matrix or not
    sp_vel->ReAssembleMatrix(false);
    
    // Compute residuals for inner iteration
    //sp_pre->Set_uk(&vsol);
    //sp_vel->Set_uk(&vsol);
    //sp_pre->CompUPResidual( t, dt,pre_res_norm_2, pre_res_norm_inf );
    //sp_vel->CompUPResidual( t, dt,vel_res_norm_2, vel_res_norm_inf );
    
    // Check if inner loop has converged
    //if ( CheckUPConvergence() ) break;
	 }
  
	 ierr = VecDestroy(u); CHKERRQ(ierr);
	 ierr = VecDestroy(b); CHKERRQ(ierr);

	 return 0;
  */
}
//----------------------------------------------------------------------------------------------------
bool TranProblemNS::CheckNonLinearConvergence()
{
  bool non_linear_convergence = false;

  NL_res_norm_2 = sqrt( sqr(pre_res_norm_2) + sqr(vel_res_norm_2) );
  if ( pre_res_norm_inf > vel_res_norm_inf ) NL_res_norm_inf = pre_res_norm_inf;
  else NL_res_norm_inf = vel_res_norm_inf;

  display->Status(4,"R(NL) = %.6f",NL_res_norm_2);

  display->Status(4,"l2-norm of NL-residuals: ");
  display->Status(4,"%.6f (momentum equation)",vel_res_norm_2);
  display->Status(4,"%.6f (continuity equation)",pre_res_norm_2);
  display->Status(4,"%.6f (total l2-residual)",NL_res_norm_2);
  display->Status(4,"%.6f (total max-residual)",NL_res_norm_inf);
    
  if (  NL_res_norm_2 < NL_tol ) non_linear_convergence = true;
  if ( (NL_res_norm_2 > NL_max_tol) && (NL_it == (max_no_NL_iter-1)) ){
    display->Error("NL residual too large.");
    exit(1);
  }

  return non_linear_convergence;
}
//----------------------------------------------------------------------------------------------------
bool TranProblemNS::CheckUPConvergence()
{
  bool UP_convergence = false;
  
  UP_res_norm_2 = sqrt( sqr(pre_res_norm_2) + sqr(vel_res_norm_2) );
  if ( pre_res_norm_inf > vel_res_norm_inf ) UP_res_norm_inf = pre_res_norm_inf;
  else UP_res_norm_inf = vel_res_norm_inf;	  

  display->Status(4,"R(UP) = %.6f",UP_res_norm_2);
	
  display->Status(4,"l2-norm of UP-residuals: ");
  display->Status(4,"%.6f (momentum equation)",vel_res_norm_2);
  display->Status(4,"%.6f (continuity equation)",pre_res_norm_2);
  display->Status(4,"%.6f (total l2-residual)",UP_res_norm_2);
  display->Status(4,"%.6f (total max-residual)",UP_res_norm_inf);

  if ( UP_res_norm_2 < UP_tol ) UP_convergence = true;
  if ( (UP_res_norm_2 > UP_max_tol) && (UP_it == (max_no_UP_iter-1)) ){
    display->Error("UP residual too large.");
    exit(1);
  }
  
  return UP_convergence;
}
//----------------------------------------------------------------------------------------------------
int TranProblemNS::SetSubgridModelingUtilities()
{
  settings->Get("compute reynolds stresses",&compute_reynolds_stresses);
  settings->Get("write reynolds stresses",&write_reynolds_stresses);

  subgrid_model_on_divergence_form = true;

  // N = (sqr(nsd)+nsd)/2;
  if ( subgrid_model_on_divergence_form ) N = 6; 
  // N =  sqr(nsd);
  else N = 9;             
  // 3 levels 
  N *= 3;

  if ( compute_reynolds_stresses ) tau.newsize(no_elms,N); 
  if ( write_reynolds_stresses )   tau_print.newsize(printNoElms,N);

  return 0;
}
//----------------------------------------------------------------------------------------------------
int TranProblemNS::SetInitialData()
{
  display->Status(1,"Set initial conditions");

  Point p;
  uk = 0.0;
  
  MV_ColMat<real> read_uk;

  char tmp[256];
  real x,y,z,pert,noRolls;
  real pi = KW_PI;
  int i,j;

  switch ( initial_data ){ 
  case zero_initial_data: 
    break;
  case white_noise: 
    for ( i=0; i < no_nodes; i++ ){
      for ( j=0; j < noeq_vel; j++ ){
		  uk(i,j) = 0.1 * 2.0*(drand48() - 0.5);
		  if ( j == 0 ) uk(i,j) += 1.0;
      }
    }
    break;
  case read_u0_inp:
    read_uk.newsize(printNoNodes,4);
    read_uk = 0.0;
    sprintf(tmp,"u_0.inp");
    readInp(tmp,grid, read_uk, noeq_vel, 0);
    if ( !(*GRIDREL)(0).NodalProlongationInitialized() ) (*GRIDREL)(0).initNodalProlongation();
    (*GRIDREL)(0).projNodalData(read_uk,uk);
    break;
  case restart_inp:
    sprintf(tmp,"u_restart.inp");
    readInp(tmp,grid, uk, noeq_vel, 0);
    break;
  case poiseuille_flow: 
    pert = 0.1;
    noRolls = 1.0;
    for ( i=0; i < no_nodes; i++ ){
		p = grid->GetNode(i)->GetCoord();
      x = real(p.x);
      y = real(p.y); 
      z = real(p.z);
      uk(i,0) = 16.0*y*(1.0-y)*z*(1-z);
      uk(i,1) = pert * (   sin(noRolls*2.0*pi*y) * cos(noRolls*pi*z) );
      uk(i,2) = pert * ( - cos(noRolls*2.0*pi*y) * sin(noRolls*pi*z) );
    }
    break;
  case couette_flow:
    pert = 0.1;
    for ( i=0; i < no_nodes; i++ ){
		p = grid->GetNode(i)->GetCoord();
      x = real(p.x);
      y = real(p.y); 
      z = real(p.z);
      uk(i,0) = 2.0*y-1.0;
      uk(i,1) = pert * (   sin(2.0*pi*y) * cos(pi*z) );
      uk(i,2) = pert * ( - cos(2.0*pi*y) * sin(pi*z) );
      //uk(i,1) = pert * ( - cos(2.0*pi*z) * sin(pi*y) );
      //uk(i,2) = pert * (   sin(2.0*pi*z) * cos(pi*y) );
    }
    break;
  default:
    display->Error("The initial condition is not implemented.");
    exit(1);
  }

  return 0;
}
//----------------------------------------------------------------------------------------------------
int TranProblemNS::WriteData( MV_ColMat<real>* U, MV_ColMat<real>* Uprev, real t, real dt )
{
  int el,nod,i;

  // Initializing Fields
  Field Ufld_1(grid);
  Field Ufld_2(grid);
  Field Ufld_3(grid);
  Field Pfld(grid);
  
  Ufld_1.InitByNodes(U,0);
  Ufld_2.InitByNodes(U,1);
  Ufld_3.InitByNodes(U,2);
  Pfld.InitByNodes(U,3);

  // Compute norms of solution
  real L1_norm_U   = Ufld_1.Lp_norm(1) + Ufld_2.Lp_norm(1) + Ufld_3.Lp_norm(1);
  real L2_norm_U   = sqrt( sqr(Ufld_1.Lp_norm(2)) + sqr(Ufld_2.Lp_norm(2)) + sqr(Ufld_3.Lp_norm(2)) );
  real Lmax_norm_U = max( max(Ufld_1.Lmax_norm(),Ufld_2.Lmax_norm()) , Ufld_3.Lmax_norm() );

  real W11_seminorm_U   = Ufld_1.Wkp_seminorm(1,1) + Ufld_2.Wkp_seminorm(1,1) + Ufld_3.Wkp_seminorm(1,1);
  real W12_seminorm_U   = sqrt( sqr(Ufld_1.Wkp_seminorm(1,2)) + sqr(Ufld_2.Wkp_seminorm(1,2)) + sqr(Ufld_3.Wkp_seminorm(1,2)) );
  real W1max_seminorm_U = max( max(Ufld_1.Wkmax_seminorm(1),Ufld_2.Wkmax_seminorm(1)), Ufld_3.Wkmax_seminorm(1) );

  real W11_norm_U   = L1_norm_U + W11_seminorm_U;
  real W12_norm_U   = sqrt( sqr(L2_norm_U) + sqr(W12_seminorm_U) );
  real W1max_norm_U = max( Lmax_norm_U, W1max_seminorm_U );

  real W21_seminorm_U,W22_seminorm_U,W2max_seminorm_U,W21_norm_U,W22_norm_U,W2max_norm_U;
  W21_seminorm_U = W22_seminorm_U = W2max_seminorm_U = W21_norm_U = W22_norm_U = W2max_norm_U = 0.0;
  /*
	 W21_seminorm_U   = Ufld_1.Wkp_seminorm(2,1) + Ufld_2.Wkp_seminorm(2,1) + Ufld_3.Wkp_seminorm(2,1);
	 W22_seminorm_U   = sqrt( sqr(Ufld_1.Wkp_seminorm(2,2)) + sqr(Ufld_2.Wkp_seminorm(2,2)) + sqr(Ufld_3.Wkp_seminorm(2,2)) );
	 W2max_seminorm_U = max( max(Ufld_1.Wkmax_seminorm(2),Ufld_2.Wkmax_seminorm(2)), Ufld_3.Wkmax_seminorm(2) );

	 W21_norm_U   = W11_norm_U + W21_seminorm_U;
	 W22_norm_U   = sqrt( sqr(W12_norm_U) + sqr(W22_seminorm_U) );
	 W2max_norm_U = max( W1max_norm_U, W2max_seminorm_U );
  */

  real L1_norm_P   = Pfld.Lp_norm(1);
  real L2_norm_P   = Pfld.Lp_norm(2);
  real Lmax_norm_P = Pfld.Lmax_norm();

  real W11_seminorm_P   = Pfld.Wkp_seminorm(1,1);
  real W12_seminorm_P   = Pfld.Wkp_seminorm(1,2);
  real W1max_seminorm_P = Pfld.Wkmax_seminorm(1);

  real W11_norm_P   = L1_norm_P + W11_seminorm_P;
  real W12_norm_P   = sqrt( sqr(L2_norm_P) + sqr(W12_seminorm_P) );
  real W1max_norm_P = max( Lmax_norm_P, W1max_seminorm_P );

  real W21_seminorm_P,W22_seminorm_P,W2max_seminorm_P,W21_norm_P,W22_norm_P,W2max_norm_P;
  W21_seminorm_P = W22_seminorm_P = W2max_seminorm_P = W21_norm_P = W22_norm_P = W2max_norm_P = 0.0;
  /*
	 W21_seminorm_P   = Pfld.Wkp_seminorm(2,1);
	 W22_seminorm_P   = Pfld.Wkp_seminorm(2,2);
	 W2max_seminorm_P = Pfld.Wkmax_seminorm(2);

	 W21_norm_P   = W11_norm_P + W21_seminorm_P;
	 W22_norm_P   = sqrt( sqr(W12_norm_P) + sqr(W22_seminorm_P) );
	 W2max_norm_P = max( W1max_norm_P, W2max_seminorm_P );
  */

  // Compute the divergence of the velocity field 
  MV_Vector<real> divU(no_elms);
  for ( el=0; el < no_elms; el++ ) divU(el) = Ufld_1.EvalGradient(el,0) + Ufld_2.EvalGradient(el,1) + Ufld_3.EvalGradient(el,2);
    
  Field Ufld_div(grid);
  Ufld_div.InitByElms(&divU);
  
  real L1_norm_divU   = Ufld_div.Lp_norm(1);
  real L2_norm_divU   = Ufld_div.Lp_norm(2);
  real Lmax_norm_divU = Ufld_div.Lmax_norm();

  // Compute time derivative   
  MV_ColMat<real> Udot(no_nodes,(*U).size(1));
  for ( nod=0; nod < no_nodes; nod++ ){
    for ( i=0; i < (*U).size(1); i++ ) Udot(nod,i) = ( (*U)(nod,i) - (*Uprev)(nod,i) ) / dt; 
  }
  Field Udotfld_1(grid);
  Field Udotfld_2(grid);
  Field Udotfld_3(grid);
  Field Pdotfld(grid);
  Udotfld_1.InitByNodes(&Udot,0);
  Udotfld_2.InitByNodes(&Udot,1);
  Udotfld_3.InitByNodes(&Udot,2);
  Pdotfld.InitByNodes(&Udot,3);
  
  real L1_norm_Udot   = Udotfld_1.Lp_norm(1) + Udotfld_2.Lp_norm(1) + Udotfld_3.Lp_norm(1);
  real L2_norm_Udot   = sqrt( sqr(Udotfld_1.Lp_norm(2)) + sqr(Udotfld_2.Lp_norm(2)) + sqr(Udotfld_3.Lp_norm(2)) );
  real Lmax_norm_Udot = max( max(Udotfld_1.Lmax_norm(),Udotfld_2.Lmax_norm()) , Udotfld_3.Lmax_norm() );

  real L1_norm_Pdot   = Pdotfld.Lp_norm(1);
  real L2_norm_Pdot   = Pdotfld.Lp_norm(2);
  real Lmax_norm_Pdot = Pdotfld.Lmax_norm();

  // Write norms of solution to matlab file
  matlab_data->SetTime(t);

  matlab_data->Set(0,L1_norm_U);
  matlab_data->Set(1,L2_norm_U);
  matlab_data->Set(2,Lmax_norm_U);
  matlab_data->Set(3,W11_seminorm_U);
  matlab_data->Set(4,W12_seminorm_U);
  matlab_data->Set(5,W1max_seminorm_U);
  matlab_data->Set(6,W11_norm_U);
  matlab_data->Set(7,W12_norm_U);
  matlab_data->Set(8,W1max_norm_U);
  matlab_data->Set(9,W21_seminorm_U);
  matlab_data->Set(10,W22_seminorm_U);
  matlab_data->Set(11,W2max_seminorm_U);
  matlab_data->Set(12,W21_norm_U);
  matlab_data->Set(13,W22_norm_U);
  matlab_data->Set(14,W2max_norm_U);
  matlab_data->Set(15,L1_norm_P);
  matlab_data->Set(16,L2_norm_P);
  matlab_data->Set(17,Lmax_norm_P);
  matlab_data->Set(18,W11_seminorm_P);
  matlab_data->Set(19,W12_seminorm_P);
  matlab_data->Set(20,W1max_seminorm_P);
  matlab_data->Set(21,W11_norm_P);
  matlab_data->Set(22,W12_norm_P);
  matlab_data->Set(23,W1max_norm_P);
  matlab_data->Set(24,W21_seminorm_P);
  matlab_data->Set(25,W22_seminorm_P);
  matlab_data->Set(26,W2max_seminorm_P);
  matlab_data->Set(27,W21_norm_P);
  matlab_data->Set(28,W22_norm_P);
  matlab_data->Set(29,W2max_norm_P);
  matlab_data->Set(30,L1_norm_divU);
  matlab_data->Set(31,L2_norm_divU);
  matlab_data->Set(32,Lmax_norm_divU);
  matlab_data->Set(33,L1_norm_Udot);
  matlab_data->Set(34,L2_norm_Udot);
  matlab_data->Set(35,Lmax_norm_Udot);
  matlab_data->Set(36,L1_norm_Pdot);
  matlab_data->Set(37,L2_norm_Pdot);
  matlab_data->Set(38,Lmax_norm_Pdot);
  matlab_data->Save();

  return 0;
}
//----------------------------------------------------------------------------------------------------
int TranProblemNS::CompContResidual( MV_ColMat<real>* U, MV_ColMat<real>* Uprev, real t, real dt, MV_Vector<real>& contResidual )
{
  int nod,el,sd;

  // R1 = | u_t + u*Du + Dp - f |  
  // R2 = max jump in gradient over element boundaries divided by h
  // R3 = time discontinuity (=0 for continuous time integration)
  // R4 = | div u |

  // Computing the R2 residual is expensive due to the approximation of second derivatives
  bool compute_R2 = false;

  //neighbor = &((*GRID)(compGrid).getNeighbor());
  //if ( compute_R2 ){
  //  if ( neighbor->elementSize() == 0 ) neighbor->init( (*GRID)(compGrid), false, false, true );
  // }


  Grid *grid = grid;
  
  Field Ufld_1(grid);
  Field Ufld_2(grid);
  Field Ufld_3(grid);
  Field Pfld(grid);
  
  Ufld_1.InitByNodes(U,0);
  Ufld_2.InitByNodes(U,1);
  Ufld_3.InitByNodes(U,2);
  Pfld.InitByNodes(U,3);

  MV_ColMat<real> Udot(no_nodes,nsd);
  for ( nod=0; nod < no_nodes; nod++ ){
    for ( sd=0; sd < nsd; sd++ ) Udot(nod,sd) = ( (*U)(nod,sd) - (*Uprev)(nod,sd) ) / dt; 
  }
  Field Udotfld_1(grid);
  Field Udotfld_2(grid);
  Field Udotfld_3(grid);
  Udotfld_1.InitByNodes(&Udot,0);
  Udotfld_2.InitByNodes(&Udot,1);
  Udotfld_3.InitByNodes(&Udot,2);
  
  // R3=0 for continuous time integration
  MV_ColMat<real> R1(no_elms,nsd);
  MV_Vector<real> R2(no_elms);
  //MV_ColMat<real> R3(no_elms,nsd);
  MV_Vector<real> R4(no_elms);

  Field Ffld_1(grid);
  Field Ffld_2(grid);
  Field Ffld_3(grid);
  real zero = 0.0;
  Ffld_1.InitByConstant(&zero);
  Ffld_2.InitByConstant(&zero);
  Ffld_3.InitByConstant(&zero);
    
  for ( el=0; el < no_elms; el++ ){
    R1(el,0) = fabs( Udotfld_1.Eval(el) + Ufld_1.Eval(el)*Ufld_1.EvalGradient(el,0) + Ufld_2.Eval(el)*Ufld_1.EvalGradient(el,1) 
							+ Ufld_3.Eval(el)*Ufld_1.EvalGradient(el,2) + Pfld.EvalGradient(el,0) - Ffld_1.Eval(el) );   
    R1(el,1) = fabs( Udotfld_2.Eval(el) + Ufld_1.Eval(el)*Ufld_2.EvalGradient(el,0) + Ufld_2.Eval(el)*Ufld_2.EvalGradient(el,1) 
							+ Ufld_3.Eval(el)*Ufld_2.EvalGradient(el,2) + Pfld.EvalGradient(el,1) - Ffld_2.Eval(el) );   
    R1(el,2) = fabs( Udotfld_3.Eval(el) + Ufld_1.Eval(el)*Ufld_3.EvalGradient(el,0) + Ufld_2.Eval(el)*Ufld_3.EvalGradient(el,1) 
							+ Ufld_3.Eval(el)*Ufld_3.EvalGradient(el,2) + Pfld.EvalGradient(el,2) - Ffld_3.Eval(el) );   
  }
  Field R1fld_1(grid);
  Field R1fld_2(grid);
  Field R1fld_3(grid);
  R1fld_1.InitByElms(&R1,0);
  R1fld_2.InitByElms(&R1,1);
  R1fld_3.InitByElms(&R1,2);

  // Compute norms of R1 residual
  real L1_norm_R1   = R1fld_1.Lp_norm(1) + R1fld_2.Lp_norm(1) + R1fld_3.Lp_norm(1);
  real L2_norm_R1   = sqrt( sqr(R1fld_1.Lp_norm(2)) + sqr(R1fld_2.Lp_norm(2)) + sqr(R1fld_3.Lp_norm(2)) );
  real Lmax_norm_R1 = max( max(R1fld_1.Lmax_norm(),R1fld_2.Lmax_norm()) , R1fld_3.Lmax_norm() );

  // Compute norms of R2 residual
  real L1_norm_R2,L2_norm_R2,Lmax_norm_R2;
  if ( compute_R2 ){
	 
    MV_Vector<real> locGrad(nsd);
    MV_Vector<real> neighborGrad(nsd);
	 
    int i,el_start,el_stop,noNeighbors;
	 Cell *c;
	 
    for ( el=0; el < no_elms; el++ ){

		c = grid->GetCell(el);
		
      R2(el) = 0.0;
      Ufld_1.EvalGradient(el,locGrad);
		
      for (int i=0;i<c->GetNoCellNeighbors();i++){
		  Ufld_1.EvalGradient(c->GetCellNeighbor(i),neighborGrad);
		  for (sd=0;sd<nsd;sd++)
			 if ( fabs(locGrad(sd) - neighborGrad(sd)) > R2(el) )
				R2(el) = fabs(locGrad(sd) - neighborGrad(sd));
      }
		
      Ufld_2.EvalGradient(el,locGrad);
		
      for (int i=0;i<c->GetNoCellNeighbors();i++){
		  Ufld_2.EvalGradient(c->GetCellNeighbor(i),neighborGrad);
		  for ( sd=0; sd < nsd; sd++ ){
			 if ( fabs(locGrad(sd) - neighborGrad(sd)) > R2(el) ) R2(el) = fabs(locGrad(sd) - neighborGrad(sd));
		  }
      }
		
      Ufld_3.EvalGradient(el,locGrad);
		
      for (int i=0;i<c->GetNoCellNeighbors();i++){
		  Ufld_3.EvalGradient(c->GetCellNeighbor(i),neighborGrad);
		  for ( sd=0; sd < nsd; sd++ ){
			 if ( fabs(locGrad(sd) - neighborGrad(sd)) > R2(el) ) R2(el) = fabs(locGrad(sd) - neighborGrad(sd));
		  }
      }
		
      R2(el) /= ( reynolds_number * LocalMeshSize(el) );
    }
    Field R2fld(grid);
    R2fld.InitByElms(&R2);

    L1_norm_R2   = R2fld.Lp_norm(1);
    L2_norm_R2   = R2fld.Lp_norm(2);
    Lmax_norm_R2 = R2fld.Lmax_norm();
  }
  
  // Compute norms of R4 residual
  for ( el=0; el < no_elms; el++ ) R4(el) = Ufld_1.EvalGradient(el,0) + Ufld_2.EvalGradient(el,1) + Ufld_3.EvalGradient(el,2);
  Field R4fld(grid);
  R4fld.InitByElms(&R4);
  
  real L1_norm_R4   = R4fld.Lp_norm(1);
  real L2_norm_R4   = R4fld.Lp_norm(2);
  real Lmax_norm_R4 = R4fld.Lmax_norm();
  
  if ( !compute_R2 ){
    L1_norm_R2   = 0.0;
    L2_norm_R2   = 0.0;
    Lmax_norm_R2 = 0.0;
  }
  real L1_norm_R3   = 0.0;
  real L2_norm_R3   = 0.0;
  real Lmax_norm_R3 = 0.0;

  // Write norms to matlab file
  matlab_data->SetTime(t);

  matlab_data->Set(no_data_entries+0,L1_norm_R1);
  matlab_data->Set(no_data_entries+1,L2_norm_R1);
  matlab_data->Set(no_data_entries+2,Lmax_norm_R1);
  matlab_data->Set(no_data_entries+3,L1_norm_R2);
  matlab_data->Set(no_data_entries+4,L2_norm_R2);
  matlab_data->Set(no_data_entries+5,Lmax_norm_R2);
  matlab_data->Set(no_data_entries+6,L1_norm_R3);
  matlab_data->Set(no_data_entries+7,L2_norm_R3);
  matlab_data->Set(no_data_entries+8,Lmax_norm_R3);
  matlab_data->Set(no_data_entries+9,L1_norm_R4);
  matlab_data->Set(no_data_entries+10,L2_norm_R4);
  matlab_data->Set(no_data_entries+11,Lmax_norm_R4);
  matlab_data->Save();

  // Write element representation of residual
  contResidual.newsize(no_elms);
  for ( el=0; el < no_elms; el++ ) contResidual(el) = R1(el,0) + R1(el,1) + R1(el,2) + R4(el);
  if ( compute_R2 ){
    for ( el=0; el < no_elms; el++ ) contResidual(el) += R2(el);
  }    

  /*
  // Write nodal representation of residual 
  MV_Vector<real> R1_1_vec;
  MV_Vector<real> R1_2_vec;
  MV_Vector<real> R1_3_vec;
  MV_Vector<real> R2_vec;
  MV_Vector<real> R4_vec;
  R1fld_1.getNodalVector(R1_1_vec);
  R1fld_2.getNodalVector(R1_2_vec);
  R1fld_3.getNodalVector(R1_3_vec);
  R2fld.getNodalVector(R2_vec);
  R4fld.getNodalVector(R4_vec);
  MV_ColMat<real> printRes(printNoNodes,5);
  for ( i=0; i < printNoNodes; i++ ){
  printRes(i,0) = R1_1_vec(i);
  printRes(i,1) = R1_2_vec(i);
  printRes(i,2) = R1_3_vec(i);
  printRes(i,3) = R2_vec(i);
  printRes(i,4) = R4_vec(i);
  }
  opendx_residuals->AddFrame(&(*GRID)(printGrid),&printRes,t);
  */

  return 0;
}
//----------------------------------------------------------------------------------------------------
int TranProblemNS::WritePertubationData( MV_ColMat<real>* U, MV_ColMat<real>* Uprev, real t, real dt )
{
  int nod,el;
  real x,y,z;
  Point p;
  Grid *grid = grid;
  
  // Write pertubation data to file
  Field Ufld_1(grid);
  Field Ufld_2(grid);
  Field Ufld_3(grid);
  
  Ufld_1.InitByNodes(U,0);
  Ufld_2.InitByNodes(U,1);
  Ufld_3.InitByNodes(U,2);

  Field Upertfld_1(grid);
  Field Upertfld_2(grid);
  Field Upertfld_3(grid);
  
  MV_Vector<real> Utmp(no_nodes);
  
  for (int i=0;i<no_nodes;i++){
	 p = grid->GetNode(i)->GetCoord();
    y = real(p.y);
    z = real(p.z);
    if ( pert_poiseuille ) Utmp(nod) = (*U)(nod,0) - ( 16.0*y*(1.0-y)*z*(1-z) );
    if ( pert_couette )    Utmp(nod) = (*U)(nod,0) - ( 2.0*y-1.0 );
  }
  Upertfld_1.InitByNodes(&Utmp);
  Upertfld_2.InitByNodes(U,1);
  Upertfld_3.InitByNodes(U,2);

  real L2_norm_U1pert, L2_norm_U2pert, L2_norm_U3pert;
  real Lmax_norm_U1pert, Lmax_norm_U2pert, Lmax_norm_U3pert;
  real L2_norm_D1U1, L2_norm_D2U1, L2_norm_D3U1;
  real L2_norm_D1U2, L2_norm_D2U2, L2_norm_D3U2;
  real L2_norm_D1U3, L2_norm_D2U3, L2_norm_D3U3;
  real Lmax_norm_D1U1, Lmax_norm_D2U1, Lmax_norm_D3U1;
  real Lmax_norm_D1U2, Lmax_norm_D2U2, Lmax_norm_D3U2;
  real Lmax_norm_D1U3, Lmax_norm_D2U3, Lmax_norm_D3U3;

  L2_norm_U1pert   = Upertfld_1.Lp_norm(2);
  L2_norm_U2pert   = Upertfld_2.Lp_norm(2);
  L2_norm_U3pert   = Upertfld_3.Lp_norm(2);
  Lmax_norm_U1pert = Upertfld_1.Lmax_norm();
  Lmax_norm_U2pert = Upertfld_2.Lmax_norm();
  Lmax_norm_U3pert = Upertfld_3.Lmax_norm();
  
  Field Ugradfld(grid);
  Utmp.newsize(no_elms);
  
  for ( el=0; el < no_elms; el++ ) Utmp(el) = Ufld_1.EvalGradient(el,0);
  Ugradfld.InitByElms(&Utmp);
  L2_norm_D1U1  = Ugradfld.Lp_norm(2);
  Lmax_norm_D1U1 = Ugradfld.Lmax_norm();
  for ( el=0; el < no_elms; el++ ) Utmp(el) = Ufld_1.EvalGradient(el,1);
  Ugradfld.InitByElms(&Utmp);
  L2_norm_D2U1  = Ugradfld.Lp_norm(2);
  Lmax_norm_D2U1 = Ugradfld.Lmax_norm();
  for ( el=0; el < no_elms; el++ ) Utmp(el) = Ufld_1.EvalGradient(el,2);
  Ugradfld.InitByElms(&Utmp);
  L2_norm_D3U1  = Ugradfld.Lp_norm(2);
  Lmax_norm_D3U1 = Ugradfld.Lmax_norm();
  
  for ( el=0; el < no_elms; el++ ) Utmp(el) = Ufld_2.EvalGradient(el,0);
  Ugradfld.InitByElms(&Utmp);
  L2_norm_D1U2  = Ugradfld.Lp_norm(2);
  Lmax_norm_D1U2 = Ugradfld.Lmax_norm();
  for ( el=0; el < no_elms; el++ ) Utmp(el) = Ufld_2.EvalGradient(el,1);
  Ugradfld.InitByElms(&Utmp);
  L2_norm_D2U2  = Ugradfld.Lp_norm(2);
  Lmax_norm_D2U2 = Ugradfld.Lmax_norm();
  for ( el=0; el < no_elms; el++ ) Utmp(el) = Ufld_2.EvalGradient(el,2);
  Ugradfld.InitByElms(&Utmp);
  L2_norm_D3U2  = Ugradfld.Lp_norm(2);
  Lmax_norm_D3U2 = Ugradfld.Lmax_norm();
  
  for ( el=0; el < no_elms; el++ ) Utmp(el) = Ufld_3.EvalGradient(el,0);
  Ugradfld.InitByElms(&Utmp);
  L2_norm_D1U3  = Ugradfld.Lp_norm(2);
  Lmax_norm_D1U3 = Ugradfld.Lmax_norm();
  for ( el=0; el < no_elms; el++ ) Utmp(el) = Ufld_3.EvalGradient(el,1);
  Ugradfld.InitByElms(&Utmp);
  L2_norm_D2U3  = Ugradfld.Lp_norm(2);
  Lmax_norm_D2U3 = Ugradfld.Lmax_norm();
  for ( el=0; el < no_elms; el++ ) Utmp(el) = Ufld_3.EvalGradient(el,2);
  Ugradfld.InitByElms(&Utmp);
  L2_norm_D3U3  = Ugradfld.Lp_norm(2);
  Lmax_norm_D3U3 = Ugradfld.Lmax_norm();
  
  // Write data to matlab file 
  matlab_data->SetTime(t);

  matlab_data->Set(no_data_entries+no_residual_entries+0,L2_norm_U1pert);
  matlab_data->Set(no_data_entries+no_residual_entries+1,Lmax_norm_U1pert);
  matlab_data->Set(no_data_entries+no_residual_entries+2,L2_norm_U2pert);
  matlab_data->Set(no_data_entries+no_residual_entries+3,Lmax_norm_U2pert);
  matlab_data->Set(no_data_entries+no_residual_entries+4,L2_norm_U3pert);
  matlab_data->Set(no_data_entries+no_residual_entries+5,Lmax_norm_U3pert);
  matlab_data->Set(no_data_entries+no_residual_entries+6,L2_norm_D1U1);
  matlab_data->Set(no_data_entries+no_residual_entries+7,Lmax_norm_D1U1);
  matlab_data->Set(no_data_entries+no_residual_entries+8,L2_norm_D2U1);
  matlab_data->Set(no_data_entries+no_residual_entries+9,Lmax_norm_D2U1);
  matlab_data->Set(no_data_entries+no_residual_entries+10,L2_norm_D3U1);
  matlab_data->Set(no_data_entries+no_residual_entries+11,Lmax_norm_D3U1);
  matlab_data->Set(no_data_entries+no_residual_entries+12,L2_norm_D1U2);
  matlab_data->Set(no_data_entries+no_residual_entries+13,Lmax_norm_D1U2);
  matlab_data->Set(no_data_entries+no_residual_entries+14,L2_norm_D2U2);
  matlab_data->Set(no_data_entries+no_residual_entries+15,Lmax_norm_D2U2);
  matlab_data->Set(no_data_entries+no_residual_entries+16,L2_norm_D3U2);
  matlab_data->Set(no_data_entries+no_residual_entries+17,Lmax_norm_D3U2);
  matlab_data->Set(no_data_entries+no_residual_entries+18,L2_norm_D1U3);
  matlab_data->Set(no_data_entries+no_residual_entries+19,Lmax_norm_D1U3);
  matlab_data->Set(no_data_entries+no_residual_entries+20,L2_norm_D2U3);
  matlab_data->Set(no_data_entries+no_residual_entries+21,Lmax_norm_D2U3);
  matlab_data->Set(no_data_entries+no_residual_entries+22,L2_norm_D3U3);
  matlab_data->Set(no_data_entries+no_residual_entries+23,Lmax_norm_D3U3);
  matlab_data->Save();

  return 0;
}
//----------------------------------------------------------------------------------------------------
int TranProblemNS::ComputeFunctionals( MV_ColMat<real>* U, MV_ColMat<real>* Uprev, real t, real dt )
{
  int nod,el,sd;

  Field Ufld_1(grid);
  Field Ufld_2(grid);
  Field Ufld_3(grid);
  Field Pfld(grid);
  
  Ufld_1.InitByNodes(U,0);
  Ufld_2.InitByNodes(U,1);
  Ufld_3.InitByNodes(U,2);
  Pfld.InitByNodes(U,3);

  MV_ColMat<real> Udot(no_nodes,nsd);
  for ( nod=0; nod < no_nodes; nod++ ){
    for ( sd=0; sd < nsd; sd++ ) Udot(nod,sd) = ( (*U)(nod,sd) - (*Uprev)(nod,sd) ) / dt; 
  }
  Field Udotfld_1(grid);
  Field Udotfld_2(grid);
  Field Udotfld_3(grid);
  Udotfld_1.InitByNodes(&Udot,0);
  Udotfld_2.InitByNodes(&Udot,1);
  Udotfld_3.InitByNodes(&Udot,2);
  
  Field Ffld_1(grid);
  Field Ffld_2(grid);
  Field Ffld_3(grid);
  real zero = 0.0;
  Ffld_1.InitByConstant(&zero);
  Ffld_2.InitByConstant(&zero);
  Ffld_3.InitByConstant(&zero);

  Field Psi_1(grid);
  Field Psi_2(grid);
  Field Psi_3(grid);
  Psi_1.InitByConstant(&zero);
  Psi_2.InitByConstant(&zero);
  Psi_3.InitByConstant(&zero);
    
  real drag = 0.0;
  for ( el=0; el < no_elms; el++ ){
    drag += ( Udotfld_1.Eval(el) + Ufld_1.Eval(el)*Ufld_1.EvalGradient(el,0) + Ufld_2.Eval(el)*Ufld_1.EvalGradient(el,1) 
				  + Ufld_3.Eval(el)*Ufld_1.EvalGradient(el,2) ) * Psi_1.Eval(el) 
		+ ( Udotfld_2.Eval(el) + Ufld_1.Eval(el)*Ufld_2.EvalGradient(el,0) + Ufld_2.Eval(el)*Ufld_2.EvalGradient(el,1) 
			 + Ufld_3.Eval(el)*Ufld_2.EvalGradient(el,2) ) * Psi_2.Eval(el) 
		+ ( Udotfld_3.Eval(el) + Ufld_1.Eval(el)*Ufld_3.EvalGradient(el,0) + Ufld_2.Eval(el)*Ufld_3.EvalGradient(el,1) 
			 + Ufld_3.Eval(el)*Ufld_3.EvalGradient(el,2) ) * Psi_3.Eval(el) 
		- Pfld.Eval(el) * (Psi_1.EvalGradient(el,0)+Psi_2.EvalGradient(el,1)+Psi_3.EvalGradient(el,2))
		+ (2.0/reynolds_number) * ( Ufld_1.EvalGradient(el,0)*Psi_1.EvalGradient(el,0)  
											 + 0.5*(Ufld_1.EvalGradient(el,1)+Ufld_2.EvalGradient(el,0))*0.5*(Psi_1.EvalGradient(el,1)+Psi_2.EvalGradient(el,0))
											 + 0.5*(Ufld_1.EvalGradient(el,2)+Ufld_3.EvalGradient(el,0))*0.5*(Psi_1.EvalGradient(el,2)+Psi_3.EvalGradient(el,0))
											 + 0.5*(Ufld_2.EvalGradient(el,0)+Ufld_1.EvalGradient(el,1))*0.5*(Psi_2.EvalGradient(el,0)+Psi_1.EvalGradient(el,1))
											 + Ufld_2.EvalGradient(el,1)*Psi_2.EvalGradient(el,1)
											 + 0.5*(Ufld_2.EvalGradient(el,2)+Ufld_3.EvalGradient(el,1))*0.5*(Psi_2.EvalGradient(el,2)+Psi_3.EvalGradient(el,1))
											 + 0.5*(Ufld_3.EvalGradient(el,0)+Ufld_1.EvalGradient(el,2))*0.5*(Psi_3.EvalGradient(el,0)+Psi_1.EvalGradient(el,0))
											 + 0.5*(Ufld_3.EvalGradient(el,1)+Ufld_2.EvalGradient(el,2))*0.5*(Psi_3.EvalGradient(el,1)+Psi_2.EvalGradient(el,0))
											 + Ufld_3.EvalGradient(el,2)*Psi_3.EvalGradient(el,2) )
		- ( Ffld_1.Eval(el)*Psi_1.Eval(el) + Ffld_2.Eval(el)*Psi_2.Eval(el) + Ffld_3.Eval(el)*Psi_3.Eval(el) );   
  }

  cout << "Drag force = " << drag << endl;
  

  // Write norms to matlab file
  /*
	 matlab_data->SetTime(t);

	 matlab_data->Set(no_data_entries+0,drag);
	 matlab_data->Save();
  */

  return 0;
}
//----------------------------------------------------------------------------------------------------
int TranProblemNS::SetTolerances()
{
  // Set tolerances for iterative methods
  NL_tol   = 1.0e-2;
  UP_tol   = 1.0e-1;
  stat_tol = 1.0e-10;

  NL_max_tol = 1000.0;
  UP_max_tol = 1000.0;

  max_no_UP_iter = 3;
  max_no_NL_iter = 100;  

  UP_conv = true;
  NL_conv = true;

  return 0;
}
//----------------------------------------------------------------------------------------------------
