// Original author: Johan Hoffman. Added to Kraftwerk 2002-01-14.
// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#ifndef TranProblemNS_h
#define TranProblemNS_h

#include "ScaleExtrapolation.hh"
#include "StatProblem.hh"
#include "TranProblem.hh"
//#include "EquationNS_cG1cG1.hh"
#include "EquationVel_cG1cG1.hh"
#include "EquationPre_cG1cG1.hh"
//#include "ElementTetLin.hh"
//#include "ElementHexLin.hh"
#include "StatProblemVelocity.hh"
#include "StatProblemPressure.hh"
//#include "StatProblemNS.hh"

enum initialDataNS { zero_initial_data, white_noise, read_u0_inp, restart_inp, poiseuille_flow, couette_flow };

///
class TranProblemNS: 
  ///
  public TranProblem {
  
public:
  
  /// Constructor
  TranProblemNS(Grid *grid, MV_Vector<GridRelation>* P);

  /// Destructor
  ~TranProblemNS();
  
  /// Solve transient problem
  int Solve();

private:
  
  /// Check if velocity (U) - pressure (P) iteration has converged 
  bool CheckUPConvergence();

  /// Check if non linear iteration has converged 
  bool CheckNonLinearConvergence();

  /// Set initial data
  int SetInitialData();

  /// Set tolerances for solution iterations
  int SetTolerances();

  /// Write pertubation data to file
  int WritePertubationData( MV_ColMat<real>* U, MV_ColMat<real>* Uprev, real t, real dt );

  /// Write solution data to file
  int WriteData( MV_ColMat<real>* U, MV_ColMat<real>* Uprev, real t, real dt );
  
  // Compute functionals 
  int ComputeFunctionals( MV_ColMat<real>* U, MV_ColMat<real>* Uprev, real t, real dt );
  
  /// Compute continuous residual 
  int CompContResidual( MV_ColMat<real>* U, MV_ColMat<real>* Uprev, real t, real dt, MV_Vector<real>& contResidual );

  /// Non linear problem using fixed point iteration
  int SolveNonLinearProblemUsingFixPoint();

  /// Solve non linear problem using Newton-GMRES
  int SolveNonLinearProblemUsingGMRES();

  /// Matrix vector multiplication for the GMRES iteration
  int MatrixMult_A( MV_ColMat<real>* vmat, int comp );

  /// Solve preconditioned problem 
  int Solve_Px_v( MV_ColMat<real>* vmat, int comp );

  /// Set parameters for subgrid modeling  
  int SetSubgridModelingUtilities();

  // Inner variables
  MV_ColMat<real> uk_prev_NL;
  MV_ColMat<real> print_uk;

  MV_ColMat<real> uk_fine;
  MV_ColMat<real> uk_coarse;

  StatProblemVelocity* sp_vel;
  StatProblemPressure* sp_pre;
  //StatProblemNS*       sp_NS;

  OpenDX *opendx;

  //NeighborFE* neighbor;
  
  real reynolds_number;

  int no_data_entries, no_residual_entries, no_pertubation_entries;

  bool compute_projections;
  bool compute_reynolds_stresses;
  bool write_reynolds_stresses;  
  bool subgrid_model_on_divergence_form;

  int N;
  MV_ColMat<real> tau;
  MV_ColMat<real> tau_print;
  ScaleExtrapolation scale_expol;
  
  bool pertubation,pert_poiseuille,pert_couette;

  initialDataNS initial_data;

  int noeq;
  int noeq_vel;
  int noeq_pre;

  real NL_tol,UP_tol,NL_max_tol,UP_max_tol;
  int max_no_UP_iter,max_no_NL_iter;
  bool UP_conv,NL_conv;

  int NL_it,UP_it;
  real UP_res_norm_2, UP_res_norm_inf, NL_res_norm_2, NL_res_norm_inf;
  real pre_res_norm_2, pre_res_norm_inf, vel_res_norm_2, vel_res_norm_inf;
  real maxnorm_du;
  
};

#endif
