#include "ProblemPoisson.hh"
#include "EquationPoisson.hh"
#include "EquationPoisson2d.hh"

//-----------------------------------------------------------------------------
const char *ProblemPoisson::Description()
{
  if ( space_dimension == 3 )
    return "Poisson's equation (3D)";
  else
	 return "Poisson's equation (2D)";
}
//-----------------------------------------------------------------------------
void ProblemPoisson::Solve()
{
  Equation *equation;
  SparseMatrix A;
  Vector b;
  Vector u;
  KrylovSolver solver;
  Output output("poisson.m",1,1);

  // Get the specified dimension
  if ( space_dimension == 3 )
	 equation = new EquationPoisson();
  else
	 equation = new EquationPoisson2d();

  // Set up the discretiser
  Discretiser discretiser(grid,equation);
  
  // Discretise Poisson's equation
  discretiser.Assemble(&A,&b);

  // Solve the linear system
  solver.Solve(&A,&u,&b);
    
  // Save the solution
  output.SetLabel(0,"Temperature");
  output.AddFrame(grid,&u,0.0);

  // Cleanup
  delete equation;
}
//-----------------------------------------------------------------------------
