#ifndef EquationPre2d_cG1cG1_h
#define EquationPre2d_cG1cG1_h

class EquationPre2d_cG1cG1: 
  public Equation {


private: 

  real theta; // theta=0.5 => cG(1)cG(1) 

  real delta1;
  real C1;
  
  real Re,dt,t;

  real umid1_1,umid1_2,umid1_3;
  real umid2_1,umid2_2,umid2_3;
  real umid1,umid2;

  real D1umid1,D2umid1,D1umid2,D2umid2;

  real divUmid;

  real Ii;
  real Iii;
  real Iij;

  MV_Vector<real> f1;
  MV_Vector<real> f2;

public:
    
    //-----------------------------------------
    //      User definitions of equations
    //-----------------------------------------
    EquationPre2d_cG1cG1(): Equation( 1,3) { 

      pi    = KW_PI;

      theta = 0.5;

      C1    = 1.0;

      Ii    = 0.25;
      Iii   = 0.1;
      Iij   = 0.05;

      f1.newsize(4);
      f2.newsize(4);
  }


    ~EquationPre2d_cG1cG1(){}
    


    //--------- Equation definition -----------
    inline
    void jacobian( int el, int gpt, int tst_bf, int trial_bf, int comp ) {
      
      BasisFcn& p = *trialFcns[0];
      BasisFcn& q  = *testFcns[0];
      
      real LAP  = p.x()*q.x() + p.y()*q.y()+ p.z()*q.z();

      work[0] = LAP;

    }
      
    
    //--------- Load definition --------------
    inline
    void residual( int el, int gpt, int tst_bf ){
      
      if ( tst_bf == 0 ){
	uprev_TS1(0) = COEFF(0) -> Eval( coord, el );    
	uprev_TS2(0) = COEFF(1) -> Eval( coord, el );    
	//pprev_TS(0) = COEFF(2) -> Eval( coord, el );    

	uprev_NL1(0) = COEFF(3) -> Eval( coord, el );    
	uprev_NL2(0) = COEFF(4) -> Eval( coord, el );    
	//pprev_NL(0) = COEFF(5) -> Eval( coord, el );    
	
	u1(0) = COEFF(6) -> Eval( coord, el );    
	u2(0) = COEFF(7) -> Eval( coord, el );    
	//p(0) = COEFF(8) -> Eval( coord, el );    

	/*
	f1(0) = (1.0/Re) * 32.0 * ( coord(1)*(1.0-coord(1)) + coord(2)*(1.0-coord(2)) );
	f2(0) = 0.001 * sin(10.0*pi*coord(0));
	*/
	f1(0) = 0.0;
	f2(0) = 0.0;

	umid1 = theta*uprev_TS1(0) + (1.0-theta)*uprev_NL1(0);
	umid2 = theta*uprev_TS2(0) + (1.0-theta)*uprev_NL2(0);

      }

      if ( (gpt==0) && (tst_bf==0) ){

	COEFF(0) -> EvalGradient(el,uprev_TS1Grad);
	COEFF(1) -> EvalGradient(el,uprev_TS2Grad);
	//COEFF(2) -> EvalGradient(el,pprev_TSGrad);

	COEFF(3) -> EvalGradient(el,uprev_NL1Grad);
	COEFF(4) -> EvalGradient(el,uprev_NL2Grad);
	//COEFF(5) -> EvalGradient(el,pprev_NLGrad);

	COEFF(6) -> EvalGradient(el,u1Grad);
	COEFF(7) -> EvalGradient(el,u2Grad);
	//COEFF(8) -> EvalGradient(el,pGrad);
	
	dt = COEFF(9) -> Eval();
	//t = COEFF(10) -> Eval();
	Re = COEFF(11) -> Eval();

	real normU = sqrt( sqr(uprev_NL1(0)) + sqr(uprev_NL2(0)) );
	if ( h * Re > 1.0 ){ 
	  //delta1 = 0.5 / sqrt( 1.0/sqr(dt) + sqr(normU/h) );
	  delta1 = C1 * h;
	}
	else{ 
	  delta1 = C1 * sqr(h);
	}
	
	D1umid1 = theta*uprev_TS1Grad(0) + (1.0-theta)*u1Grad(0);
	D2umid1 = theta*uprev_TS1Grad(1) + (1.0-theta)*u1Grad(1);
	
	D1umid2 = theta*uprev_TS2Grad(0) + (1.0-theta)*u2Grad(0);
	D2umid2 = theta*uprev_TS2Grad(1) + (1.0-theta)*u2Grad(1);

	divUmid = D1umid1 + D2umid2;
      }

      BasisFcn& q = *testFcns[0];
      
      real f_Dq = f1(0)*q.x() + f2(0)*q.y();

      real uDu_Dq = ( umid1*D1umid1 + umid2*D2umid1 )*q.x() 
	          + ( umid1*D1umid2 + umid2*D2umid2 )*q.y(); 

      work[0] = f_Dq - uDu_Dq - (1.0/delta1) * divUmid*q();

    }



    //--------- LHS definition for exact integration on tetrahedrons --------------
    inline
    void jacobianExactInt( MV_Vector<real>& jac, int& tst_bf, int& trial_bf, 
			   int& j_eq, int& el, Grid* grid ){ 
      
      FET3n2D* FE=NULL;
      FE = &(grid->FET3);
      if ( FE->getElmNo() != el ) FE->refill( el );

      if ( (tst_bf==0) && (trial_bf==0) ){
      h = 2.0*FE->circumRadius();
      }

      real qx,qy;
      if ( tst_bf == 0 ){ 
	qx = FE->dN1x();
	qy = FE->dN1y();
      }else if ( tst_bf == 1 ){
	qx = FE->dN2x();
	qy = FE->dN2y();
      }else if ( tst_bf == 2 ){
	qx = FE->dN3x();
	qy = FE->dN3y();
      }	

      real px,py;
      if ( trial_bf == 0 ){
	px = FE->dN1x();
	py = FE->dN1y();
      }else if ( trial_bf == 1 ){
	px = FE->dN2x();
	py = FE->dN2y();
      }else if ( trial_bf == 2 ){
	px = FE->dN3x();
	py = FE->dN3y();
      }	

      real pq;
      if ( tst_bf == trial_bf ){
	pq = Iii;
      } else{
	pq = Iij;
      }

      real MASS = pq;
      real alpha = 0.0;

      jac(0) = alpha*MASS + px*qx + py*qy;
            
    }


  
    //--------- Load definition for exact integration on tetrahedrons --------------
    inline
    void residualExactInt( MV_ColMat<real>& bres, int& bf, int& el, Grid* grid ){ 
      
      FET3n2D* FE=NULL;
      FE = &(grid->FET3);
      if ( FE->getElmNo() != el ) FE->refill( el );

      if ( bf == 0 ){
      h = 2.0*FE->circumRadius();

      x(0) = FE->x1();
      x(1) = FE->x2();
      x(2) = FE->x3();

      y(0) = FE->y1();
      y(1) = FE->y2();
      y(2) = FE->y3();

      uprev_TS1(0)     = COEFF(0) -> GetNodalValue( FE->n1() );    
      uprev_TS1(1)     = COEFF(0) -> GetNodalValue( FE->n2() );
      uprev_TS1(2)     = COEFF(0) -> GetNodalValue( FE->n3() );
      FE->evalGradientLoc(uprev_TS1,uprev_TS1Grad);
      uprev_TS2(0)     = COEFF(1) -> GetNodalValue( FE->n1() );
      uprev_TS2(1)     = COEFF(1) -> GetNodalValue( FE->n2() );
      uprev_TS2(2)     = COEFF(1) -> GetNodalValue( FE->n3() );
      FE->evalGradientLoc(uprev_TS2,uprev_TS2Grad);
      //pprev_TS(0)      = COEFF(2) -> GetNodalValue( FE->n1() );
      //pprev_TS(1)      = COEFF(2) -> GetNodalValue( FE->n2() );
      //pprev_TS(2)      = COEFF(2) -> GetNodalValue( FE->n3() );
      //FE->evalGradientLoc(pprev_TS,pprev_TSGrad);

      uprev_NL1(0)     = COEFF(3) -> GetNodalValue( FE->n1() );    
      uprev_NL1(1)     = COEFF(3) -> GetNodalValue( FE->n2() );
      uprev_NL1(2)     = COEFF(3) -> GetNodalValue( FE->n3() );
      FE->evalGradientLoc(uprev_NL1,uprev_NL1Grad);
      uprev_NL2(0)     = COEFF(4) -> GetNodalValue( FE->n1() );
      uprev_NL2(1)     = COEFF(4) -> GetNodalValue( FE->n2() );
      uprev_NL2(2)     = COEFF(4) -> GetNodalValue( FE->n3() );
      FE->evalGradientLoc(uprev_NL2,uprev_NL2Grad);
      //pprev_NL(0)      = COEFF(5) -> GetNodalValue( FE->n1() );
      //pprev_NL(1)      = COEFF(5) -> GetNodalValue( FE->n2() );
      //pprev_NL(2)      = COEFF(5) -> GetNodalValue( FE->n3() );
      //FE->evalGradientLoc(pprev_NL,pprev_NLGrad);

      u1(0)     = COEFF(6) -> GetNodalValue( FE->n1() );
      u1(1)     = COEFF(6) -> GetNodalValue( FE->n2() );
      u1(2)     = COEFF(6) -> GetNodalValue( FE->n3() );
      FE->evalGradientLoc(u1,u1Grad);
      u2(0)     = COEFF(7) -> GetNodalValue( FE->n1() );
      u2(1)     = COEFF(7) -> GetNodalValue( FE->n2() );
      u2(2)     = COEFF(7) -> GetNodalValue( FE->n3() );
      FE->evalGradientLoc(u2,u2Grad);
      p(0)      = COEFF(8) -> GetNodalValue( FE->n1() );      
      p(1)      = COEFF(8) -> GetNodalValue( FE->n2() );
      p(2)      = COEFF(8) -> GetNodalValue( FE->n3() );
      FE->evalGradientLoc(p,pGrad);

      dt           = COEFF(9) -> Eval();
      //t            = COEFF(10) -> Eval();
      Re           = COEFF(11) -> Eval();
      
      f1(0) = (1.0/Re) * 32.0 * ( y(0)*(1.0-y(0)) );
      f1(1) = (1.0/Re) * 32.0 * ( y(1)*(1.0-y(1)) );
      f1(2) = (1.0/Re) * 32.0 * ( y(2)*(1.0-y(2)) );
      f1 = 0.0;

      f2(0) = 0.001 * sin(10.0*pi*x(0));
      f2(1) = 0.001 * sin(10.0*pi*x(1));
      f2(2) = 0.001 * sin(10.0*pi*x(2));
      f2 = 0.0;

      real u1sum = u1(0) + u1(1) + u1(2) + u1(3);
      real u2sum = u2(0) + u2(1) + u2(2) + u2(3);
      real normU = sqrt( sqr(u1sum) + sqr(u2sum) )/4;
      
      if ( h * Re > 1.0 ){ 
	//delta1 = 0.5 / sqrt( 1.0/sqr(dt) + sqr(normU/h) );
	delta1 = C1 * h;
      }
      else{ 
	delta1 = C1 * sqr(h);
      }

      umid1_1 = theta*uprev_TS1(0) + (1.0-theta)*uprev_NL1(0);
      umid1_2 = theta*uprev_TS1(1) + (1.0-theta)*uprev_NL1(1);
      umid1_3 = theta*uprev_TS1(2) + (1.0-theta)*uprev_NL1(2);

      umid2_1 = theta*uprev_TS2(0) + (1.0-theta)*uprev_NL2(0);
      umid2_2 = theta*uprev_TS2(1) + (1.0-theta)*uprev_NL2(1);
      umid2_3 = theta*uprev_TS2(2) + (1.0-theta)*uprev_NL2(2);

      umid1 = umid1_1 + umid1_2 + umid1_3;
      umid2 = umid2_1 + umid2_2 + umid2_3;
      
      
      D1umid1 = theta*uprev_TS1Grad(0) + (1.0-theta)*u1Grad(0);
      D2umid1 = theta*uprev_TS1Grad(1) + (1.0-theta)*u1Grad(1);

      D1umid2 = theta*uprev_TS2Grad(0) + (1.0-theta)*u2Grad(0);
      D2umid2 = theta*uprev_TS2Grad(1) + (1.0-theta)*u2Grad(1);

      }

      real qx,qy;
      if ( bf == 0 ){ 
	qx = FE->dN1x();
	qy = FE->dN1y();
      }else if ( bf == 1 ){
	qx = FE->dN2x();
	qy = FE->dN2y();
      }else if ( bf == 2 ){
	qx = FE->dN3x();
	qy = FE->dN3y();
      }	

      real f_Dq = ( f1(0)+f1(1)+f1(2) )*Ii*qx
      	        + ( f2(0)+f2(1)+f2(2) )*Ii*qy;

      real uDu_Dq = ( umid1*D1umid1+umid2*D2umid1 )*Ii*qx
                  + ( umid1*D1umid2+umid2*D2umid2 )*Ii*qy;
                
      
      if ( bf == 0 ){
      divUmid = theta*uprev_TS1Grad(0) + (1.0-theta)*u1Grad(0)
	      + theta*uprev_TS2Grad(1) + (1.0-theta)*u2Grad(1);
      }      

      //bres(0,bf) = (1.0/delta1) * (divUmid*Ii);
      bres(0,bf) = f_Dq - uDu_Dq - (1.0/delta1)*(divUmid*Ii);
            
    }
  

  inline
  void MatrixVectorProductExactInt( MV_ColMat<real>& bres, int& bf, int& el, FET4n3D& FE ){}


  inline
  void NonLinearResidualExactInt( MV_ColMat<real>& bres, int& bf, int& el, FET4n3D& FE ){}

};


#endif

