#ifndef EquationVel2d_cG1cG1_h
#define EquationVel2d_cG1cG1_h

class EquationVel2d_cG1cG1: 
  public Equation {


private: 

  real theta; // theta=0.5 => cG(1)cG(1) 

  real delta1,delta2,r;
  real C1,C2,K;

  real Re,dt,t;

  real Cs,S_abs,nu,nu_T;

  real umid1_1,umid1_2,umid1_3;
  real umid2_1,umid2_2,umid2_3;
  real umid1,umid2;

  real u1sum,u2sum,psum;

  real u1u1,u1u2,u2u1,u2u2;
  real f1u1,f1u2,f2u1,f2u2;

  real normU,divU;

  real Ii;
  real Iii;
  real Iij;

  MV_Vector<real> f1;
  MV_Vector<real> f2;

  real vx,vy;
  real uv1,uv2,uv3;
  real D1u1,D2u1,D1u2,D2u2;
  real uv;
  real MASS1,MASS2,NL1,NL2,LAP1,LAP2;
  real uDu_uDv_1,uDu_uDv_2;
  real LSdiv1,LSdiv2;
  real p_u_Dv_1,p_u_Dv_2;
  real f_uDv_1,f_uDv_2;
  real SD1,SD2;
  real uprev_TS_v_1,uprev_TS_v_2;
  real f_v_1,f_v_2;
  real p_div_v_1,p_div_v_2;
  real u_Dv_1,u_Dv_2;
  real u_v_1,u_v_2;

public:
    
    //-----------------------------------------
    //      User definitions of equations
    //-----------------------------------------
    EquationVel2d_cG1cG1() : Equation(2,2) { 
      
      pi                = KW_PI;

      theta             = 0.5; 

      C1                = 1.0;
      C2                = 1.0;
      K                 = 20.0;

      Cs                = 0.1;
      
      Ii                = 0.25;
      Iii               = 0.1;
      Iij               = 0.05;

      f1.newsize(4);
      f2.newsize(4);

      settings->Get("variational multscale eddy viscosity",&variational_multiscale_eddy_viscosity);
      settings->Get("smagorinsky eddy viscosity",&variational_multiscale_eddy_viscosity);
    }


    ~EquationVel2d_cG1cG1(){}


    //--------- Equation definition -----------
    inline
    void jacobian( int el, int gpt, int tst_bf, int trial_bf, int comp ) {

      if ( (tst_bf==0) && (trial_bf==0) && (comp==0) ){
	uprev_TS1(0) = COEFF(0) -> Eval( coord, el );    
	uprev_TS2(0) = COEFF(1) -> Eval( coord, el );    
	//pprev_TS(0) = COEFF(2) -> Eval( coord, el );    

	uprev_NL1(0) = COEFF(3) -> Eval( coord, el );    
	uprev_NL2(0) = COEFF(4) -> Eval( coord, el );    
	//pprev_NL(0) = COEFF(5) -> Eval( coord, el );    

	//u1(0) = COEFF(6) -> Eval( coord, el );    
	//u2(0) = COEFF(7) -> Eval( coord, el );    
	//p(0) = COEFF(8) -> Eval( coord, el );    

	if ( variational_multiscale_eddy_viscosity ){
	  u1_coarse(0) = COEFF(12) -> Eval( coord, el );
	  u2_coarse(0) = COEFF(13) -> Eval( coord, el );
	  p_coarse(0) = COEFF(14) -> Eval( coord, el );
	  
	  u1_fine(0) = COEFF(15) -> Eval( coord, el );
	  u2_fine(0) = COEFF(16) -> Eval( coord, el );
	  p_fine(0) = COEFF(17) -> Eval( coord, el );
	}
	
	umid1 = theta*uprev_TS1(0) + (1.0-theta)*uprev_NL1(0);
	umid2 = theta*uprev_TS2(0) + (1.0-theta)*uprev_NL2(0);

      }

      if ( (gpt==0) && (tst_bf==0) && (trial_bf==0) && (comp==0) ){

	COEFF(0) -> EvalGradient(el,uprev_TS1Grad);
	COEFF(1) -> EvalGradient(el,uprev_TS2Grad);
	//COEFF(2) -> EvalGradient(el,pprev_TSGrad);
		
	//COEFF(3) -> EvalGradient(el,uprev_NL1Grad);
	//COEFF(4) -> EvalGradient(el,uprev_NL2Grad);
	//COEFF(5) -> EvalGradient(el,pprev_NLGrad);
	
	//COEFF(6) -> EvalGradient(el,u1Grad);
	//COEFF(7) -> EvalGradient(el,u2Grad);
	//COEFF(8) -> EvalGradient(el,pGrad);

	dt = COEFF(9) -> Eval();
	//t = COEFF(10) -> Eval();
	Re = COEFF(11) -> Eval();
      
	if ( variational_multiscale_eddy_viscosity ){
	  COEFF(12) -> EvalGradient(el,u1Grad_coarse);
	  COEFF(13) -> EvalGradient(el,u2Grad_coarse);
	  COEFF(14) -> EvalGradient(el,pGrad_coarse);
	  
	  COEFF(15) -> EvalGradient(el,u1Grad_fine);
	  COEFF(16) -> EvalGradient(el,u2Grad_fine);
	  COEFF(17) -> EvalGradient(el,pGrad_fine);
	}
      
	normU = sqrt( sqr(uprev_NL1(0)) + sqr(uprev_NL2(0)) );
	if ( h * Re > 1.0 ){ 
	  //delta1 = 0.5 / sqrt( 1.0/sqr(dt) + sqr(normU/h) );
	  delta1 = C1 * h;
	}
	else{ 
	  delta1 = C1 * sqr(h);
	}
	if ( h * Re > 1.0 ){ 
	  delta2 = C2 * h;
	}
	else{ 
	  delta2 = C2 * sqr(h);
	}
      
      }

      BasisFcn& u1 = *trialFcns[0];
      BasisFcn& u2 = *trialFcns[1];
      BasisFcn& v  = *testFcns[0];

      LSdiv1 = (1.0-theta) * (u1.x()+u2.y()) * v();
      LSdiv2 = LSdiv1;

      uDu_uDv_1 = (1.0-theta) * ( umid1*u1.x()+umid2*u1.y() ) * ( umid1*v.x()+umid2*v.y() );
      uDu_uDv_2 = (1.0-theta) * ( umid1*u2.x()+umid2*u2.y() ) * ( umid1*v.x()+umid2*v.y() );

      NL1 = (1.0-theta) * ( umid1*u1.x() + umid2*u1.y() ) * v(); 
      NL2 = (1.0-theta) * ( umid1*u2.x() + umid2*u2.y() ) * v(); 
      
      MASS1 = u1() * v();
      MASS2 = u2() * v();

      LAP1 = 1/Re * (1.0-theta) * (u1.x()*v.x() + u1.y()*v.y());
      LAP2 = 1/Re * (1.0-theta) * (u2.x()*v.x() + u2.y()*v.y());

      work[0] = MASS1/dt + LAP1 + NL1 + delta1 * uDu_uDv_1 + delta2 * LSdiv1;
      work[1] = MASS2/dt + LAP2 + NL2 + delta1 * uDu_uDv_2 + delta2 * LSdiv2;
    
    }
      
    
    //--------- Load definition --------------
    inline
    void residual( int el, int gpt, int tst_bf ){
      
      if ( tst_bf==0 ){

	uprev_TS1(0) = COEFF(0) -> Eval( coord, el );    
	uprev_TS2(0) = COEFF(1) -> Eval( coord, el );    
	//pprev_TS(0) = COEFF(2) -> Eval( coord, el );    

	uprev_NL1(0) = COEFF(3) -> Eval( coord, el );    
	uprev_NL2(0) = COEFF(4) -> Eval( coord, el );    
	//pprev_NL(0) = COEFF(5) -> Eval( coord, el );    
	
	//u1(0) = COEFF(6) -> Eval( coord, el );    
	//u2(0) = COEFF(7) -> Eval( coord, el );    
	p(0) = COEFF(8) -> Eval( coord, el );    

	if ( variational_multiscale_eddy_viscosity ){
	  u1_coarse(0) = COEFF(12) -> Eval( coord, el );
	  u2_coarse(0) = COEFF(13) -> Eval( coord, el );
	  p_coarse(0) = COEFF(14) -> Eval( coord, el );
	  
	  u1_fine(0) = COEFF(15) -> Eval( coord, el );
	  u2_fine(0) = COEFF(16) -> Eval( coord, el );
	  p_fine(0) = COEFF(17) -> Eval( coord, el );
	}
	
	umid1 = theta*uprev_TS1(0) + (1.0-theta)*uprev_NL1(0);
	umid2 = theta*uprev_TS2(0) + (1.0-theta)*uprev_NL2(0);

      }

      if ( (gpt==0) && (tst_bf==0) ){

	COEFF(0) -> EvalGradient(el,uprev_TS1Grad);
	COEFF(1) -> EvalGradient(el,uprev_TS2Grad);
	//COEFF(2) -> EvalGradient(el,pprev_TSGrad);
	
	//COEFF(3) -> EvalGradient(el,uprev_NL1Grad);
	//COEFF(4) -> EvalGradient(el,uprev_NL2Grad);
	//COEFF(5) -> EvalGradient(el,pprev_NLGrad);
	
	//COEFF(6) -> EvalGradient(el,u1Grad);
	//COEFF(7) -> EvalGradient(el,u2Grad);
	COEFF(8) -> EvalGradient(el,pGrad);

	dt = COEFF(9) -> Eval();
	//t = COEFF(10) -> Eval();
	Re = COEFF(11) -> Eval();
      
	if ( variational_multiscale_eddy_viscosity ){
	  COEFF(12) -> EvalGradient(el,u1Grad_coarse);
	  COEFF(13) -> EvalGradient(el,u2Grad_coarse);
	  COEFF(14) -> EvalGradient(el,pGrad_coarse);
	  
	  COEFF(15) -> EvalGradient(el,u1Grad_fine);
	  COEFF(16) -> EvalGradient(el,u2Grad_fine);
	  COEFF(17) -> EvalGradient(el,pGrad_fine);
	}
      
	normU = sqrt( sqr(uprev_NL1(0)) + sqr(uprev_NL2(0)) );
	if ( h * Re > 1.0 ){ 
	  //delta1 = 0.5 / sqrt( 1.0/sqr(dt) + sqr(normU/h) );
	  delta1 = C1 * h;
	}
	else{ 
	  delta1 = C1 * sqr(h);
	}
	if ( h * Re > 1.0 ){ 
	  delta2 = C2 * h;
	}
	else{ 
	  delta2 = C2 * sqr(h);
	}
	
	/*
	f1(0) = (1.0/Re) * 32.0 * ( coord(1)*(1.0-coord(1)) + coord(2)*(1.0-coord(2)) );
	f2(0) = 0.001 * sin(10.0*pi*coord(0));
	*/
	f1(0) = 0.0;
	f2(0) = 0.0;

      }
      
      BasisFcn& u1 = *trialFcns[0];
      BasisFcn& u2 = *trialFcns[1];
      BasisFcn& v  = *testFcns[0];
      
      LSdiv1 = theta * (uprev_TS1Grad(0)+uprev_TS2Grad(1)) * v();
      LSdiv2 = LSdiv1;


      uDu_uDv_1 = theta * ( umid1*u1.x()+umid2*u1.y() ) * ( umid1*v.x()+umid2*v.y() );
      uDu_uDv_2 = theta * ( umid1*u2.x()+umid2*u2.y() ) * ( umid1*v.x()+umid2*v.y() );

      p_u_Dv_1 = pGrad(0) * ( umid1*v.x()+umid2*v.y() );
      p_u_Dv_2 = pGrad(1) * ( umid1*v.x()+umid2*v.y() );

      f_uDv_1 = f1(0) * ( umid1*v.x()+umid2*v.y() );
      f_uDv_2 = f2(0) * ( umid1*v.x()+umid2*v.y() );

      SD1 = f_uDv_1 - uDu_uDv_1 - p_u_Dv_1;
      SD2 = f_uDv_2 - uDu_uDv_2 - p_u_Dv_2;

      NL1 = theta * ( umid1*uprev_TS1Grad(0) + umid2*uprev_TS1Grad(1) ) * v(); 
      NL2 = theta * ( umid1*uprev_TS2Grad(0) + umid2*uprev_TS2Grad(1) ) * v(); 
      
      MASS1 = uprev_TS1(0) * v();
      MASS2 = uprev_TS2(0) * v();

      LAP1 = 1/Re * theta * (uprev_TS1Grad(0)*v.x() + uprev_TS1Grad(1)*v.y());
      LAP2 = 1/Re * theta * (uprev_TS2Grad(0)*v.x() + uprev_TS2Grad(1)*v.y());

      work[0] = MASS1/dt + LAP1 + NL1 + p(0)*v.x() + f1(0)*v() + delta1 * SD1 + delta2 * LSdiv1;
      work[1] = MASS2/dt + LAP2 + NL2 + p(0)*v.y() + f2(0)*v() + delta1 * SD2 + delta2 * LSdiv2;

    }
  
  
    //--------- Load definition for exact integration on tetrahedrons --------------
    inline
    void jacobianExactInt( MV_Vector<real>& jac, int& tst_bf, int& trial_bf, 
			   int& j_eq, int& el, Grid* grid ){ 
       
      
      FET3n2D* FE=NULL;
      FE = &(grid->FET3);
      if ( FE->getElmNo() != el ) FE->refill( el );

      if ( (tst_bf==0) && (trial_bf==0) ){
	h = 2.0*FE->circumRadius();
	
	uprev_TS1(0)     = COEFF(0) -> GetNodalValue( FE->n1() );    
	uprev_TS1(1)     = COEFF(0) -> GetNodalValue( FE->n2() );
	uprev_TS1(2)     = COEFF(0) -> GetNodalValue( FE->n3() );
	FE->evalGradientLoc(uprev_TS1,uprev_TS1Grad);
	uprev_TS2(0)     = COEFF(1) -> GetNodalValue( FE->n1() );
	uprev_TS2(1)     = COEFF(1) -> GetNodalValue( FE->n2() );
	uprev_TS2(2)     = COEFF(1) -> GetNodalValue( FE->n3() );
	FE->evalGradientLoc(uprev_TS2,uprev_TS2Grad);
	//pprev_TS(0)      = COEFF(2) -> GetNodalValue( FE->n1() );
	//pprev_TS(1)      = COEFF(2) -> GetNodalValue( FE->n2() );
	//pprev_TS(2)      = COEFF(2) -> GetNodalValue( FE->n3() );
	//FE->evalGradientLoc(pprev_TS,pprev_TSGrad);
	
	uprev_NL1(0)     = COEFF(3) -> GetNodalValue( FE->n1() );    
	uprev_NL1(1)     = COEFF(3) -> GetNodalValue( FE->n2() );
	uprev_NL1(2)     = COEFF(3) -> GetNodalValue( FE->n3() );
	//FE->evalGradientLoc(uprev_NL1,uprev_NL1Grad);
	uprev_NL2(0)     = COEFF(4) -> GetNodalValue( FE->n1() );
	uprev_NL2(1)     = COEFF(4) -> GetNodalValue( FE->n2() );
	uprev_NL2(2)     = COEFF(4) -> GetNodalValue( FE->n3() );
	//FE->evalGradientLoc(uprev_NL2,uprev_NL2Grad);
	//pprev_NL(0)      = COEFF(5) -> GetNodalValue( FE->n1() );
	//pprev_NL(1)      = COEFF(5) -> GetNodalValue( FE->n2() );
	//pprev_NL(2)      = COEFF(5) -> GetNodalValue( FE->n3() );
	//FE->evalGradientLoc(pprev_NL,pprev_NLGrad);
	
	u1(0)     = COEFF(6)  -> GetNodalValue( FE->n1() );
	u1(1)     = COEFF(6)  -> GetNodalValue( FE->n2() );
	u1(2)     = COEFF(6)  -> GetNodalValue( FE->n3() );
	//FE->evalGradientLoc(u1,u1Grad);
	u2(0)     = COEFF(7) -> GetNodalValue( FE->n1() );
	u2(1)     = COEFF(7) -> GetNodalValue( FE->n2() );
	u2(2)     = COEFF(7) -> GetNodalValue( FE->n3() );
	//FE->evalGradientLoc(u2,u2Grad);
	//p(0)      = COEFF(8) -> GetNodalValue( FE->n1() );
	//p(1)      = COEFF(8) -> GetNodalValue( FE->n2() );
	//p(2)      = COEFF(8) -> GetNodalValue( FE->n3() );
	//FE->evalGradientLoc(p,pGrad);
	
	dt           = COEFF(9) -> Eval();
	//t            = COEFF(10) -> Eval();
	Re           = COEFF(11) -> Eval();
	
	if ( variational_multiscale_eddy_viscosity ){
	  u1_coarse(0)     = COEFF(12)  -> GetNodalValue( FE->n1() );
	  u1_coarse(1)     = COEFF(12)  -> GetNodalValue( FE->n2() );
	  u1_coarse(2)     = COEFF(12)  -> GetNodalValue( FE->n3() );
	  FE->evalGradientLoc(u1_coarse,u1Grad_coarse);
	  u2_coarse(0)     = COEFF(13) -> GetNodalValue( FE->n1() );
	  u2_coarse(1)     = COEFF(13) -> GetNodalValue( FE->n2() );
	  u2_coarse(2)     = COEFF(13) -> GetNodalValue( FE->n3() );
	  FE->evalGradientLoc(u2_coarse,u2Grad_coarse);
	  //p_coarse(0)      = COEFF(14) -> GetNodalValue( FE->n1() );
	  //p_coarse(1)      = COEFF(14) -> GetNodalValue( FE->n2() );
	  //p_coarse(2)      = COEFF(14) -> GetNodalValue( FE->n3() );
	  //FE->evalGradientLoc(p_coarse,pGrad_coarse);
	  
	  u1_fine(0)     = COEFF(15)  -> GetNodalValue( FE->n1() );
	  u1_fine(1)     = COEFF(15)  -> GetNodalValue( FE->n2() );
	  u1_fine(2)     = COEFF(15)  -> GetNodalValue( FE->n3() );
	  FE->evalGradientLoc(u1_fine,u1Grad_fine);
	  u2_fine(0)     = COEFF(16) -> GetNodalValue( FE->n1() );
	  u2_fine(1)     = COEFF(16) -> GetNodalValue( FE->n2() );
	  u2_fine(2)     = COEFF(16) -> GetNodalValue( FE->n3() );
	  FE->evalGradientLoc(u2_fine,u2Grad_fine);
	  //p_fine(0)      = COEFF(17) -> GetNodalValue( FE->n1() );
	  //p_fine(1)      = COEFF(17) -> GetNodalValue( FE->n2() );
	  //p_fine(2)      = COEFF(17) -> GetNodalValue( FE->n3() );
	  //FE->evalGradientLoc(p_fine,pGrad_fine);
	}

	u1sum = u1(0) + u1(1) + u1(2);
	u2sum = u2(0) + u2(1) + u2(2);
	normU = sqrt( sqr(u1sum) + sqr(u2sum) ) / 4;
	
	if ( h * Re > 1.0 ){ 
	  //delta1 = 0.5 / sqrt( 1.0/sqr(dt) + sqr(normU/h) );
	  delta1 = C1 * h;
	}
	else{ 
	  delta1 = C1 * sqr(h);
	}
	if ( h * Re > 1.0 ){ 
	  delta2 = C2 * h;
	}
	else{ 
	  delta2 = C2 * sqr(h);
	}
      }
      
      if ( tst_bf == 0 ){ 
	uv1 = Iii;
	uv2 = uv3 = Iij;
	vx = FE->dN1x();
	vy = FE->dN1y();
      }else if ( tst_bf == 1 ){
	uv2 = Iii;
	uv1 = uv3 = Iij;
	vx = FE->dN2x();
	vy = FE->dN2y();
      }else if ( tst_bf == 2 ){
	uv3 = Iii;
	uv1 = uv2 = Iij;
	vx = FE->dN3x();
	vy = FE->dN3y();
      }

      if ( trial_bf == 0 ){
	D1u1 = D1u2 = FE->dN1x();
	D2u1 = D2u2 = FE->dN1y();
	umid1 = theta*uprev_TS1(0) + (1.0-theta)*uprev_NL1(0);
	umid2 = theta*uprev_TS2(0) + (1.0-theta)*uprev_NL2(0);
      }else if ( trial_bf == 1 ){
	D1u1 = D1u2 = FE->dN2x();
	D2u1 = D2u2 = FE->dN2y();
	umid1 = theta*uprev_TS1(1) + (1.0-theta)*uprev_NL1(1);
	umid2 = theta*uprev_TS2(1) + (1.0-theta)*uprev_NL2(1);
      }else if ( trial_bf == 2 ){
	D1u1 = D1u2 = FE->dN3x();
	D2u1 = D2u2 = FE->dN3y();
	umid1 = theta*uprev_TS1(2) + (1.0-theta)*uprev_NL1(2);
	umid2 = theta*uprev_TS2(2) + (1.0-theta)*uprev_NL2(2);
      }

      if ( (tst_bf==0) && (trial_bf==0) ){
	umid1_1 = theta*uprev_TS1(0) + (1.0-theta)*uprev_NL1(0);
	umid1_2 = theta*uprev_TS1(1) + (1.0-theta)*uprev_NL1(1);
	umid1_3 = theta*uprev_TS1(2) + (1.0-theta)*uprev_NL1(2);

	umid2_1 = theta*uprev_TS2(0) + (1.0-theta)*uprev_NL2(0);
	umid2_2 = theta*uprev_TS2(1) + (1.0-theta)*uprev_NL2(1);
	umid2_3 = theta*uprev_TS2(2) + (1.0-theta)*uprev_NL2(2);

	u1u1 = umid1_1*umid1_1*Iii + umid1_1*umid1_2*Iij + umid1_1*umid1_3*Iij 
	     + umid1_2*umid1_1*Iij + umid1_2*umid1_2*Iii + umid1_2*umid1_3*Iij 
    	     + umid1_3*umid1_1*Iij + umid1_3*umid1_2*Iij + umid1_3*umid1_3*Iii; 
	u1u2 = umid1_1*umid2_1*Iii + umid1_1*umid2_2*Iij + umid1_1*umid2_3*Iij 
	     + umid1_2*umid2_1*Iij + umid1_2*umid2_2*Iii + umid1_2*umid2_3*Iij 
	     + umid1_3*umid2_1*Iij + umid1_3*umid2_2*Iij + umid1_3*umid2_3*Iii; 
        u2u1 = u1u2;
	u2u2 = umid2_1*umid2_1*Iii + umid2_1*umid2_2*Iij + umid2_1*umid2_3*Iij
	     + umid2_2*umid2_1*Iij + umid2_2*umid2_2*Iii + umid2_2*umid2_3*Iij
	     + umid2_3*umid2_1*Iij + umid2_3*umid2_2*Iij + umid2_3*umid2_3*Iii;
      }

      if ( tst_bf == trial_bf ){
	uv = Iii;
      } else{
        uv = Iij;
      }

      if ( j_eq == 0 ){

	uDu_uDv_1 = (1.0-theta) * ( D1u1 * (vx*u1u1 + vy*u1u2)  
				  + D2u1 * (vx*u2u1 + vy*u2u2) );
	uDu_uDv_2 = 0.0;

	NL1 = (1.0-theta) * ( (umid1_1*uv1 + umid1_2*uv2 + umid1_3*uv3) * D1u1
			    + (umid2_1*uv1 + umid2_2*uv2 + umid2_3*uv3) * D2u1 );
	NL2 = 0.0;

	MASS1 = uv; 
	MASS2 = 0.0;

	LSdiv1 = (1.0-theta) * D1u1 * vx;
	LSdiv2 = (1.0-theta) * D1u1 * vy;

      } else if ( j_eq == 1 ){

	uDu_uDv_2 = (1.0-theta) * ( D1u2 * (vx*u1u1 + vy*u1u2)  
				  + D2u2 * (vx*u2u1 + vy*u2u2) );
	uDu_uDv_1 = 0.0;

	NL2 = (1.0-theta) * ( (umid1_1*uv1 + umid1_2*uv2 + umid1_3*uv3) * D1u2
			    + (umid2_1*uv1 + umid2_2*uv2 + umid2_3*uv3) * D2u2 );
	NL1 = 0.0;

	MASS2 = uv; 
	MASS1 = 0.0;

	LSdiv1 = (1.0-theta) * D2u2 * vx;
	LSdiv2 = (1.0-theta) * D2u2 * vy;

      }	else{
	cout << "ERROR: Wrong number of equations" << endl;
	exit(1);
      }


      if ( smagorinsky_eddy_viscosity ){
	if ( (tst_bf==0) && (trial_bf==0) ){
	  S_abs = sqrt( sqr(uprev_TS1Grad(0)) + sqr(uprev_TS1Grad(1))
	              + sqr(uprev_TS2Grad(0)) + sqr(uprev_TS2Grad(1)) );
	  nu_T  = pow(Cs*h,2.0) * S_abs; 
	  nu = (1.0/Re) + nu_T;
	}
      } else{
	nu = (1.0/Re);
      }	

      LAP1 = LAP2 = 0.0;
      if ( j_eq == 0 ) LAP1 = (1.0-theta) * nu * ( D1u1*vx + D2u1*vy );
      if ( j_eq == 1 ) LAP2 = (1.0-theta) * nu * ( D1u2*vx + D2u2*vy );
      
      if ( variational_multiscale_eddy_viscosity ){
	if ( (tst_bf==0) && (trial_bf==0) ){
	  S_abs = sqrt( sqr(uprev_NL1Grad(0)) + sqr(uprev_NL1Grad(1)) 
		      + sqr(uprev_NL2Grad(0)) + sqr(uprev_NL2Grad(1)) );
	  nu_T  = pow(Cs*h,2.0) * S_abs; 
	}
	if ( j_eq == 0 ) 
	  LAP1 += nu_T * ( (D1u1-u1Grad_coarse(0))*vx + (D2u1-u1Grad_coarse(1))*vy );
	if ( j_eq == 1 ) 
	  LAP2 += nu_T * ( (D1u2-u2Grad_coarse(0))*vx + (D2u2-u2Grad_coarse(1))*vy );
      }
      
      jac(0) = MASS1/dt + LAP1 + NL1 + delta1 * uDu_uDv_1 + delta2 * LSdiv1;
      jac(1) = MASS2/dt + LAP2 + NL2 + delta1 * uDu_uDv_2 + delta2 * LSdiv2;
            
    }



    //--------- Load definition for exact integration on tetrahedrons --------------
    inline
    void residualExactInt( MV_ColMat<real>& bres, int& bf, int& el, Grid* grid ){ 

      FET3n2D* FE=NULL;
      FE = &(grid->FET3);
      if ( FE->getElmNo() != el ) FE->refill( el );

      if ( bf == 0 ){
	h = 2.0*FE->circumRadius();
	
	x(0) = FE->x1();
	x(1) = FE->x2();
	x(2) = FE->x3();
	
	y(0) = FE->y1();
	y(1) = FE->y2();
	y(2) = FE->y3();
	
	uprev_TS1(0)     = COEFF(0) -> GetNodalValue( FE->n1() );    
	uprev_TS1(1)     = COEFF(0) -> GetNodalValue( FE->n2() );
	uprev_TS1(2)     = COEFF(0) -> GetNodalValue( FE->n3() );
	FE->evalGradientLoc(uprev_TS1,uprev_TS1Grad);
	uprev_TS2(0)     = COEFF(1) -> GetNodalValue( FE->n1() );
	uprev_TS2(1)     = COEFF(1) -> GetNodalValue( FE->n2() );
	uprev_TS2(2)     = COEFF(1) -> GetNodalValue( FE->n3() );
	FE->evalGradientLoc(uprev_TS2,uprev_TS2Grad);
	//pprev_TS(0)      = COEFF(2) -> GetNodalValue( FE->n1() );
	//pprev_TS(1)      = COEFF(2) -> GetNodalValue( FE->n2() );
	//pprev_TS(2)      = COEFF(2) -> GetNodalValue( FE->n3() );
	//FE->evalGradientLoc(pprev_TS,pprev_TSGrad);
	
	uprev_NL1(0)     = COEFF(3) -> GetNodalValue( FE->n1() );    
	uprev_NL1(1)     = COEFF(3) -> GetNodalValue( FE->n2() );
	uprev_NL1(2)     = COEFF(3) -> GetNodalValue( FE->n3() );
	//FE->evalGradientLoc(uprev_NL1,uprev_NL1Grad);
	uprev_NL2(0)     = COEFF(4) -> GetNodalValue( FE->n1() );
	uprev_NL2(1)     = COEFF(4) -> GetNodalValue( FE->n2() );
	uprev_NL2(2)     = COEFF(4) -> GetNodalValue( FE->n3() );
	//FE->evalGradientLoc(uprev_NL2,uprev_NL2Grad);
	//pprev_NL(0)      = COEFF(5) -> GetNodalValue( FE->n1() );
	//pprev_NL(1)      = COEFF(5) -> GetNodalValue( FE->n2() );
	//pprev_NL(2)      = COEFF(5) -> GetNodalValue( FE->n3() );
	//FE->evalGradientLoc(pprev_NL,pprev_NLGrad);
	
	u1(0)     = COEFF(6)  -> GetNodalValue( FE->n1() );
	u1(1)     = COEFF(6)  -> GetNodalValue( FE->n2() );
	u1(2)     = COEFF(6)  -> GetNodalValue( FE->n3() );
	FE->evalGradientLoc(u1,u1Grad);
	u2(0)     = COEFF(7) -> GetNodalValue( FE->n1() );
	u2(1)     = COEFF(7) -> GetNodalValue( FE->n2() );
	u2(2)     = COEFF(7) -> GetNodalValue( FE->n3() );
	FE->evalGradientLoc(u2,u2Grad);
	p(0)      = COEFF(8) -> GetNodalValue( FE->n1() );
	p(1)      = COEFF(8) -> GetNodalValue( FE->n2() );
	p(2)      = COEFF(8) -> GetNodalValue( FE->n3() );
	FE->evalGradientLoc(p,pGrad);
	
	dt           = COEFF(9) -> Eval();
	//t            = COEFF(10) -> Eval();
	Re           = COEFF(11) -> Eval();
	
	f1(0) = (1.0/Re) * 32.0 * ( y(0)*(1.0-y(0)) );
	f1(1) = (1.0/Re) * 32.0 * ( y(1)*(1.0-y(1)) );
	f1(2) = (1.0/Re) * 32.0 * ( y(2)*(1.0-y(2)) );
	f1 = 0.0;

	f2(0) = 0.001 * sin(10.0*pi*x(0));
	f2(1) = 0.001 * sin(10.0*pi*x(1));
	f2(2) = 0.001 * sin(10.0*pi*x(2));
	f2 = 0.0;

	u1sum = u1(0) + u1(1) + u1(2);
	u2sum = u2(0) + u2(1) + u2(2);
	normU = sqrt( sqr(u1sum) + sqr(u2sum) ) / 4;
	
	divU = u1Grad(0) + u2Grad(1);
	
	if ( h * Re > 1.0 ){ 
	  //delta1 = 0.5 / sqrt( 1.0/sqr(dt) + sqr(normU/h) );
	  delta1 = C1 * h;
	}
	else{ 
	  delta1 = C1 * sqr(h);
	}
	if ( h * Re > 1.0 ){ 
	  delta2 = C2 * h;
	}
	else{ 
	  delta2 = C2 * sqr(h);
	}
	if ( divU > K ){ 
	  r = 0.5 * divU;
	}
	else{ 
	  r = 0.0;
	}
	r = 0.0;
      }
      
      if ( bf == 0 ){ 
	uv1 = Iii;
	uv2 = uv3 = Iij;
	vx = FE->dN1x();
	vy = FE->dN1y();
      }else if ( bf == 1 ){
	uv2 = Iii;
	uv1 = uv3 = Iij;
	vx = FE->dN2x();
	vy = FE->dN2y();
      }else if ( bf == 2 ){
	uv3 = Iii;
	uv1 = uv2 = Iij;
	vx = FE->dN3x();
	vy = FE->dN3y();
      }	

      if ( bf == 0 ){
	umid1_1 = theta*uprev_TS1(0) + (1.0-theta)*uprev_NL1(0);
	umid1_2 = theta*uprev_TS1(1) + (1.0-theta)*uprev_NL1(1);
	umid1_3 = theta*uprev_TS1(2) + (1.0-theta)*uprev_NL1(2);

	umid2_1 = theta*uprev_TS2(0) + (1.0-theta)*uprev_NL2(0);
	umid2_2 = theta*uprev_TS2(1) + (1.0-theta)*uprev_NL2(1);
	umid2_3 = theta*uprev_TS2(2) + (1.0-theta)*uprev_NL2(2);

	umid1 = umid1_1 + umid1_2 + umid1_3;
	umid2 = umid2_1 + umid2_2 + umid2_3;
	
	psum = p(0) + p(1) + p(2);

	u1u1 = umid1_1*umid1_1*Iii + umid1_1*umid1_2*Iij + umid1_1*umid1_3*Iij
	     + umid1_2*umid1_1*Iij + umid1_2*umid1_2*Iii + umid1_2*umid1_3*Iij
	     + umid1_3*umid1_1*Iij + umid1_3*umid1_2*Iij + umid1_3*umid1_3*Iii;
	u1u2 = umid1_1*umid2_1*Iii + umid1_1*umid2_2*Iij + umid1_1*umid2_3*Iij
	     + umid1_2*umid2_1*Iij + umid1_2*umid2_2*Iii + umid1_2*umid2_3*Iij
	     + umid1_3*umid2_1*Iij + umid1_3*umid2_2*Iij + umid1_3*umid2_3*Iii;
	u2u1 = u1u2;
	u2u2 = umid2_1*umid2_1*Iii + umid2_1*umid2_2*Iij + umid2_1*umid2_3*Iij
	     + umid2_2*umid2_1*Iij + umid2_2*umid2_2*Iii + umid2_2*umid2_3*Iij
       	     + umid2_3*umid2_1*Iij + umid2_3*umid2_2*Iij + umid2_3*umid2_3*Iii;
      }

      uDu_uDv_1 = theta * ( uprev_TS1Grad(0) * (vx*u1u1 + vy*u1u2)  
	                  + uprev_TS1Grad(1) * (vx*u2u1 + vy*u2u2) );
	                  
      uDu_uDv_2 = theta * ( uprev_TS2Grad(0) * (vx*u1u1 + vy*u1u2)  
	                  + uprev_TS2Grad(1) * (vx*u2u1 + vy*u2u2) );
	                  

      if ( bf == 0 ){
	f1u1 = f1(0)*umid1_1*Iii + f1(0)*umid1_2*Iij + f1(0)*umid1_3*Iij
	     + f1(1)*umid1_1*Iij + f1(1)*umid1_2*Iii + f1(1)*umid1_3*Iij
	     + f1(2)*umid1_1*Iij + f1(2)*umid1_2*Iij + f1(2)*umid1_3*Iii;
	f1u2 = f1(0)*umid2_1*Iii + f1(0)*umid2_2*Iij + f1(0)*umid2_3*Iij
	     + f1(1)*umid2_1*Iij + f1(1)*umid2_2*Iii + f1(1)*umid2_3*Iij
	     + f1(2)*umid2_1*Iij + f1(2)*umid2_2*Iij + f1(2)*umid2_3*Iii;
	f2u1 = f2(0)*umid1_1*Iii + f2(0)*umid1_2*Iij + f2(0)*umid1_3*Iij
	     + f2(1)*umid1_1*Iij + f2(1)*umid1_2*Iii + f2(1)*umid1_3*Iij
	     + f2(2)*umid1_1*Iij + f2(2)*umid1_2*Iij + f2(2)*umid1_3*Iii;
	f2u2 = f2(0)*umid2_1*Iii + f2(0)*umid2_2*Iij + f2(0)*umid2_3*Iij
	     + f2(1)*umid2_1*Iij + f2(1)*umid2_2*Iii + f2(1)*umid2_3*Iij
	     + f2(2)*umid2_1*Iij + f2(2)*umid2_2*Iij + f2(2)*umid2_3*Iii;
      }

      f_uDv_1 = vx*f1u1 + vy*f1u2;
      f_uDv_2 = vx*f2u1 + vy*f2u2;

      NL1   = theta * ( (umid1_1*uv1+umid1_2*uv2+umid1_3*uv3) * uprev_TS1Grad(0)
              	      + (umid2_1*uv1+umid2_2*uv2+umid2_3*uv3) * uprev_TS1Grad(1) );

      NL2   = theta * ( (umid1_1*uv1+umid1_2*uv2+umid1_3*uv3) * uprev_TS2Grad(0)
	              + (umid2_1*uv1+umid2_2*uv2+umid2_3*uv3) * uprev_TS2Grad(1) );


      if ( smagorinsky_eddy_viscosity ){
	if ( bf == 0 ){
	  S_abs = sqrt( sqr(uprev_TS1Grad(0)) + sqr(uprev_TS1Grad(1))
	              + sqr(uprev_TS2Grad(0)) + sqr(uprev_TS2Grad(1)) );
	  nu_T  = pow(Cs*h,2.0) * S_abs; 
	  nu = (1.0/Re) + nu_T;
	}
      } else{
	nu = (1.0/Re);
      }
      LAP1 = theta * nu * ( uprev_TS1Grad(0)*vx + uprev_TS1Grad(1)*vy );
      LAP2 = theta * nu * ( uprev_TS2Grad(0)*vx + uprev_TS2Grad(1)*vy );

      uprev_TS_v_1 = uprev_TS1(0)*uv1+uprev_TS1(1)*uv2+uprev_TS1(2)*uv3;
      uprev_TS_v_2 = uprev_TS2(0)*uv1+uprev_TS2(1)*uv2+uprev_TS2(2)*uv3;
      
      f_v_1 = f1(0)*uv1+f1(1)*uv2+f1(2)*uv3;
      f_v_2 = f2(0)*uv1+f2(1)*uv2+f2(2)*uv3;

      p_div_v_1 = psum*vx*Ii;
      p_div_v_2 = psum*vy*Ii;

      u_Dv_1 = ( umid1*vx + umid2*vy ) * Ii; 
      u_Dv_2 = u_Dv_1;

      SD1 = f_uDv_1 - uDu_uDv_1 - pGrad(0)*u_Dv_1;
      SD2 = f_uDv_2 - uDu_uDv_2 - pGrad(1)*u_Dv_2;

      LSdiv1 = theta * (uprev_TS1Grad(0)+uprev_TS2Grad(1)) * vx;
      LSdiv2 = theta * (uprev_TS1Grad(0)+uprev_TS2Grad(1)) * vy;

      u_v_1 = u1(0)*uv1+u1(1)*uv2+u1(2)*uv3;
      u_v_2 = u2(0)*uv1+u2(1)*uv2+u2(2)*uv3;

      bres(0,bf) = uprev_TS_v_1/dt - LAP1 - NL1 + p_div_v_1 + f_v_1 + delta1*SD1 - delta2*LSdiv1 - r*u_v_1;
      bres(1,bf) = uprev_TS_v_2/dt - LAP2 - NL2 + p_div_v_2 + f_v_2 + delta1*SD2 - delta2*LSdiv2 - r*u_v_2;

    }
  

  inline
  void MatrixVectorProductExactInt( MV_ColMat<real>& bres, int& bf, int& el, FET4n3D& FE ){}

  inline
  void NonLinearResidualExactInt( MV_ColMat<real>& bres, int& bf, int& el, FET4n3D& FE ){}

};


#endif

