// Original author: Johan Hoffman. Added to Kraftwerk 2002-01-14.
// Copyright (C) 2002 Johan Hoffman and Anders Logg.
// Licensed under the GNU GPL Version 2.

#ifndef __STAT_PROBLEM_PRESSURE2D_HH
#define __STAT_PROBLEM_PRESSURE2D_HH

#include "StatProblem.hh"

class StatProblemPressure2d: public StatProblem {
  
public:
  
  StatProblemPressure2d( MV_Vector<Grid>* g );
  
  ~StatProblemPressure2d(){}
  
  int Timestep( MV_ColMat<real> &val, real time, real deltat );
  
  int ComputeDiscreteResidual( real time, real deltat, real& norm_2, real& norm_inf );
  
private:
  
  int InitEquationData( real time, real deltat );

  int SetDomain();
  
  real ReNo;
  real bnd_x0,bnd_x1,bnd_y0,bnd_y1,bnd_z0,bnd_z1;
  
  Field u1_prev_TS,u2_prev_TS,p_prev_TS;
  Field u1_prev_NL,u2_prev_NL,p_prev_NL;
  Field u1,u2,p;
  Field dt,t,Re;
  
  MV_ColMat<real> bndx0;
  MV_ColMat<real> bndx1;
  MV_ColMat<real> bndy0;
  MV_ColMat<real> bndy1;
  
  bool periodic_bc_x;
  bool periodic_bc_y;

  bool outflow_bc_x0;
  bool outflow_bc_x1;
};

#endif
