#include <kraftwerk.h>

void SetParameters ();

int main(int argc, char **argv)
{
  // Set kraftwerk parameters
  SetParameters();
  
  // Initialize kraftwerk
  kw_init(argc,argv);
  
  // Save parameters
  kw_save_parameters();

  // Solve the problem
  kw_solve();

  // Finished
  kw_end();
  
  return 0;
}

void SetParameters()
{
  // Set problem
  kw_set_problem("navier-stokes");
  
  // Set parameters
  kw_set_parameter("problem description", "Time-dependent 3d Navier-Stokes, bluff body, Re = 1000");
  kw_set_parameter("grid file",           "../../../../data/grids/tetgrid_bluff_body_32_8_8.inp");
  kw_set_parameter("output file prefix",  "bluff_body");
  kw_set_parameter("display",             "terminal");
  kw_set_parameter("output samples",      10);
  kw_set_parameter("start time",          0.0 );
  kw_set_parameter("final time",          1.0 );
  kw_set_parameter("solve primal",        true);
  kw_set_parameter("solve dual",          false);
  kw_set_parameter("write data",          true);
  kw_set_parameter("write residuals",     false);
  kw_set_parameter("debug level",         0);

  // Special parameters
  kw_set_parameter("write couette pertubations",    false);
  kw_set_parameter("write poiseuille pertubations", false);
  
  kw_set_parameter("compute projections",  false);

  kw_set_parameter("reynolds number",      1000.0);

  kw_set_parameter("backward facing step", false);
  kw_set_parameter("bluff body",           true);
  kw_set_parameter("jet",                  false);
  kw_set_parameter("driven cavity",        false);

  kw_set_parameter("slip bc x0",           false);
  kw_set_parameter("slip bc x1",           false);
  kw_set_parameter("slip bc y0",           false);
  kw_set_parameter("slip bc y1",           false);
  kw_set_parameter("slip bc z0",           false);
  kw_set_parameter("slip bc z1",           false);

  kw_set_parameter("no slip bc x0",        false);
  kw_set_parameter("no slip bc x1",        false);
  kw_set_parameter("no slip bc y0",        true);
  kw_set_parameter("no slip bc y1",        true);
  kw_set_parameter("no slip bc z0",        true);
  kw_set_parameter("no slip bc z1",        true);
  
  kw_set_parameter("periodic bc x",        false);
  kw_set_parameter("periodic bc y",        false);
  kw_set_parameter("periodic bc z",        false);

  kw_set_parameter("inflow bc x0",         true);
  kw_set_parameter("outflow bc x0",        false);
  kw_set_parameter("outflow bc x1",        true);

  kw_set_parameter("couette bc y",         false);

  kw_set_parameter("driven cavity bc x0",  false);
  kw_set_parameter("driven cavity bc y1",  false);

  kw_set_parameter("dual drag bc",         false);

  kw_set_parameter("bnd x0",               0.0);
  kw_set_parameter("bnd x1",               4.0);
  kw_set_parameter("bnd y0",               0.0);
  kw_set_parameter("bnd y1",               1.0);
  kw_set_parameter("bnd z0",               0.0);
  kw_set_parameter("bnd z1",               1.0);
}
